/**

2002 ZORAN Corporation, All Rights Reserved THIS IS PROPRIETARY SOURCE CODE OF
ZORAN CORPORATION

%file

  AvLibRegister.c

%desc

%author

  Pavel Piarvou

%created

  March 4, 2008

*/

#include "AvOsStub.h"
#include <assert.h>

#include "Types.h"
#include "avcodec.h"
#include "avformat.h"
#include "av.h"
#include "AvLibRegister.h"
#include "AvzLoadTables.h"
#include "AvApi.h"
#include "InitAv.h"

/*

%func

  AvLibRegister

%desc

  initialize pointers to functions of AV and Avz libraries which should be used
  in external libraries

%param

    None.

%return

    None.

*/
VOID
AvLibRegister()
{
    g_pfnAvcodec_init                   = AvApi_avcodec_init;

    g_pfnAvcodec_open                   = AvApi_avcodec_open;
    g_pfnAvcodec_close                  = AvApi_avcodec_close;
    g_pfnRegister_avcodec               = AvApi_register_avcodec;
    g_pfnAvcodec_find_encoder           = AvApi_avcodec_find_encoder;

    // TODO: move to avf
    g_pfnAvcodec_find_encoder_by_name   = AvApi_avcodec_find_encoder_by_name;
    g_pfnAvcodec_find_decoder_by_name   = AvApi_avcodec_find_decoder_by_name;
    g_pfnAvcodec_find_decoder           = AvApi_avcodec_find_decoder;

    g_pfnAvcodec_alloc_context          = AvApi_avcodec_alloc_context;
    g_pfnAvcodec_alloc_frame            = AvApi_avcodec_alloc_frame;
    g_pfnAvcodec_encode_video           = AvApi_avcodec_encode_video;
    g_pfnAvcodec_decode_video           = AvApi_avcodec_decode_video;
    g_pfnAvcodec_encode_audio           = AvApi_avcodec_encode_audio;
    g_pfnAvcodec_decode_audio           = AvApi_avcodec_decode_audio;

    g_pfnAvpicture_fill                 = AvApi_avpicture_fill;
}


VOID
AvzLibRegister()
{
#ifdef AACZ_ENCODER
    g_pfnAvGetAaczEncoder               = AvGetCodecAaczEncoder;
#endif // AACZ_ENCODER

#ifdef AACZ_DECODER
    g_pfnAvGetAaczDecoder               = AvGetCodecAaczDecoder;
#endif //AACZ_DECODER

    g_pfnAvGetH264zDecoder              = AvGetCodecH264zDecoder;
    g_pfnAvGetRawzDecoder               = AvGetCodecRawzDecoder;
    g_pfnAvGetRawzEncoder               = AvGetCodecRawzEncoder;
    g_pfnAvGetJpegzDecoder              = AvGetCodecJpegzDecoder;

#ifndef BUILD_VERILOG
    g_pfnAvGetMp2Decoder                = AvGetCodecMp2Decoder;
    g_pfnAvGetMp2Encoder                = AvGetCodecMp2Encoder;
#endif

#if 0
    g_pfnAvGetMjpegDecoder      = AvGetCodecMjpegDecoder;
#endif

    g_pfnAvGetPcmH264zEncoder           = AvGetCodecPcmH264zEncoder;

    g_pfnAvGetAdpcmImaQtEncoder         = AvGetCodecAdpcmImaQtEncoder;
    g_pfnAvGetAdpcmImaQtDecoder         = AvGetCodecAdpcmImaQtDecoder;
    g_pfnAvGetAdpcmImaWavEncoder        = AvGetCodecAdpcmImaWavEncoder;
    g_pfnAvGetAdpcmImaWavDecoder        = AvGetCodecAdpcmImaWavDecoder;
    g_pfnAvGetAdpcmImaDk3Encoder        = AvGetCodecAdpcmImaDk3Encoder;
    g_pfnAvGetAdpcmImaDk3Decoder        = AvGetCodecAdpcmImaDk3Decoder;
    g_pfnAvGetAdpcmImaDk4Encoder        = AvGetCodecAdpcmImaDk4Encoder;
    g_pfnAvGetAdpcmImaDk4Decoder        = AvGetCodecAdpcmImaDk4Decoder;
    g_pfnAvGetAdpcmImaWsEncoder         = AvGetCodecAdpcmImaWsEncoder;
    g_pfnAvGetAdpcmImaWsDecoder         = AvGetCodecAdpcmImaWsDecoder;
    g_pfnAvGetAdpcmImaSmjpegEncoder     = AvGetCodecAdpcmImaSmjpegEncoder;
    g_pfnAvGetAdpcmImaSmjpegDecoder     = AvGetCodecAdpcmImaSmjpegDecoder;
    g_pfnAvGetAdpcmMsEncoder            = AvGetCodecAdpcmMsEncoder;
    g_pfnAvGetAdpcmMsDecoder            = AvGetCodecAdpcmMsDecoder;
    g_pfnAvGetAdpcm4xmEncoder           = AvGetCodecAdpcm4xmEncoder;
    g_pfnAvGetAdpcm4xmDecoder           = AvGetCodecAdpcm4xmDecoder;
    g_pfnAvGetAdpcmXaEncoder            = AvGetCodecAdpcmXaEncoder;
    g_pfnAvGetAdpcmXaDecoder            = AvGetCodecAdpcmXaDecoder;
    g_pfnAvGetAdpcmAdxEncoder           = AvGetCodecAdpcmAdxEncoder;
    g_pfnAvGetAdpcmAdxDecoder           = AvGetCodecAdpcmAdxDecoder;
    g_pfnAvGetAdpcmEaEncoder            = AvGetCodecAdpcmEaEncoder;
    g_pfnAvGetAdpcmEaDecoder            = AvGetCodecAdpcmEaDecoder;

    g_pfnAvGetPcmS16leEncoder           = AvGetCodecPcmS16leEncoder;
    g_pfnAvGetPcmS16leDecoder           = AvGetCodecPcmS16leDecoder;
    g_pfnAvGetPcmS16beEncoder           = AvGetCodecPcmS16beEncoder;
    g_pfnAvGetPcmS16beDecoder           = AvGetCodecPcmS16beDecoder;
    g_pfnAvGetPcmU16leEncoder           = AvGetCodecPcmU16leEncoder;
    g_pfnAvGetPcmU16leDecoder           = AvGetCodecPcmU16leDecoder;
    g_pfnAvGetPcmU16beEncoder		= AvGetCodecPcmU16beEncoder;
    g_pfnAvGetPcmU16beDecoder           = AvGetCodecPcmU16beDecoder;
    g_pfnAvGetPcmS8Encoder              = AvGetCodecPcmS8Encoder;
    g_pfnAvGetPcmS8Decoder              = AvGetCodecPcmS8Decoder;
    g_pfnAvGetPcmU8Encoder              = AvGetCodecPcmU8Encoder;
    g_pfnAvGetPcmU8Decoder              = AvGetCodecPcmU8Decoder;
    g_pfnAvGetPcmAlawEncoder            = AvGetCodecPcmAlawEncoder;
    g_pfnAvGetPcmAlawDecoder            = AvGetCodecPcmAlawDecoder;
    g_pfnAvGetPcmMulawEncoder           = AvGetCodecPcmMulawEncoder;
    g_pfnAvGetPcmMulawDecoder           = AvGetCodecPcmMulawDecoder;

    g_pfnAvcodec_get_context_defaults = AvApi_avcodec_get_context_defaults;

    g_pfnAvCodec_decoder_priorities_set = AvApi_avcodec_set_thread_priorities;
    g_pfnAvCodec_operation_mode_set = AvApi_avcodec_set_operation_mode;
//avz
    g_pfnAv_mep_set_code_buf            = NULL;



}

#if 0
#endif

VOID AvLibInitCall()
{
    AvCodecInit();
}

AVCODEC_INIT
AvGetCodecInitFn(void)
{
   if (NULL == g_pfnAvcodec_init)
   {
        fprintf(stderr, "function %s in file %s not initialized %d\n", (PCHAR)__FUNCTION__, (PCHAR)__FILE__, __LINE__);\
        OsSysHaltEx("function not initialized %d\n", (PCHAR)__FILE__, __LINE__);\
   }
   return g_pfnAvcodec_init;
}

/*

%func

  AvzSetTablesPermanency

%desc

  sets or clears permanency property of tables for codec

%param

    bMark  - TRUE means that permanency property shoul be set, overwise cleared
    uiCodecs2Load - IDs of codecs which tables should be permanently in memory

%return

    None.

*/
VOID AvzSetTablesPermanency(BOOL bMark, UINT uiCodecs2Load)
{
#if (FW_VERSION == FW_VERSION_C13)
    // This code is removed due to dynamic linkage errors with undefined
    // AvzMarkTablesPermanency & AvzUnMarkTablesPermanency
    assert(0);
#else    
    #if (STATIC_PARAM_TABLES == 0)
    if (TRUE == bMark)
    {
        AvzMarkTablesPermanency(uiCodecs2Load);
    }
    else
    {
        AvzUnMarkTablesPermanency(uiCodecs2Load);
    }
    #endif
#endif
}

/*

%func

  AvzLoadTablesCodec

%desc

  initiate loading tables for codecs uiCodecs2Load

%param

    uiCodecs2Load - IDs of codecs to load tables

%return

    None.

*/
VOID
AvzLoadTablesCodec(UINT uiCodecs2Load)
{
#if (FW_VERSION == FW_VERSION_C13)
    // This code is removed due to dynamic linkage errors with undefined AvzLoadTablesList4Codec
    assert(0);   
#else
    #if (STATIC_PARAM_TABLES == 0)
    ULONG ulErrorCode = 0;
    BOOL retVal = TRUE;
    retVal = AvzLoadTablesList4Codec(uiCodecs2Load, &ulErrorCode);
    if (FALSE == retVal )
    {
        fprintf(stderr, "Error during loading AVZ tables 0x%08X\n", ulErrorCode);
    }
    #endif //   #if (STATIC_PARAM_TABLES == 0)
#endif
}


/*

%func

  AvzFreeTablesOfCodec

%desc

  free memory for requested codec

%param

    uiCodecs2Load - IDs of codecs to free their tables

%return

    None.

*/
VOID
AvzFreeTablesOfCodec(UINT uiCodecs2Free)
{
#if (FW_VERSION == FW_VERSION_C13)
    // This code is removed due to dynamic linkage errors with undefined AvzFreeTablesListOfCodec
    assert(0);   
#else
    #if (STATIC_PARAM_TABLES == 0)
    AvzFreeTablesListOfCodec(uiCodecs2Free);
    #endif //   #if (STATIC_PARAM_TABLES == 0)
#endif
}

/*

%func

  AvzTablesLoad

%desc

  load all tables for all codecs

TODO: get list of names of files from resource file

%param

    None.

%return

    None.

*/
VOID AvzTablesLoad()
{
#if (FW_VERSION == FW_VERSION_C13)
    // This code is removed due to dynamic linkage errors with undefined AvzInitTablesList2Load
    assert(0);   
#else
    #if (STATIC_PARAM_TABLES == 0)

    AvzInitTablesList2Load(NULL, 0);

    #endif
#endif
}

/*

%func

  AvzSetReadResourceCallback

%desc

  set callback function to read resource files

%param

    pfReadResourceCallback - pointer to function which manage reading from file

%return

    None.

*/
VOID
AvzSetReadResourceCallback(PVOID pfReadResourceCallback)
{
#if (FW_VERSION == FW_VERSION_C13)
    // This code is removed due to dynamic linkage errors with undefined AvzReadResourceRegisterCallback
    assert(0);   
#else
    #if (STATIC_PARAM_TABLES == 0)
    AvzReadResourceRegisterCallback((PVOID)pfReadResourceCallback);
    #endif
#endif
}

