// This file is part of the camera os components sources.
#include "Audio.h"
#include "Allocator.h"
//#include "debug.h"
//#include "Component.h"
#include "Timing.h"

USHORT guwPlaybackHeaderCount;
IMA_ADPCM_CHAN_STATUS gsChannelStatus[2];



/*
EXTERN PBYTE pbyGlobalWavBuf;
EXTERN ULONG ulGWBIndex;
EXTERN PBYTE pbyGlobalWavWrkPtr;
*/

//STATIC ULONG ulBytesToHeader = 0;

#define CLAMP_TO_SHORT(value) \
if (value > 32767) \
    value = 32767; \
else if (value < -32768) \
    value = -32768; \

/* step_table[] and index_table[] are from the ADPCM reference source */
/* This is the index table: */
static const int index_table[16] = {
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
};

/** 
 * This is the step table. Note that many programs use slight deviations from
 * this table, but such deviations are negligible:
 */
static const int step_table[89] = {
    7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
    19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
    50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
    130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
    337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
    876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
    2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
    5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
    15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
};

static const short wMaxStepIndex = sizeof(step_table) / sizeof(int) - 1;

/*
STATIC INT
Clip(int a, int amin, int amax)
{
    if (a < amin)
        return amin;
    else if (a > amax)
        return amax;
    else
        return a;
}
*/

STATIC SHORT
IMAAdpcmExpandSample(IMA_ADPCM_CHAN_STATUS *c, char byCode)
{
    SHORT wStep = step_table[c->step_index];
    SHORT wDiff = wStep >> 3;
    LONG  lPrev_Sample = c->prev_sample;
    SHORT wStepIndex = c->step_index;

    if ( byCode & 1 )
    {
        wDiff += (wStep >> 2);
    }
    if ( byCode & 2 )
    {
        wDiff += (wStep >> 1);
    }
    if ( byCode & 4 )
    {
        wDiff += wStep;
    }
    if ( byCode & 8 )
    {
        wDiff = -wDiff;
    }

    lPrev_Sample += wDiff;
    CLAMP_TO_SHORT( lPrev_Sample );
    c->prev_sample = (SHORT)lPrev_Sample;
     
    wStepIndex += index_table[(UCHAR)byCode];    
    if ( wStepIndex < 0 )
    {
        wStepIndex = 0;
    }
    else if ( wStepIndex > wMaxStepIndex )
    {
        wStepIndex = wMaxStepIndex;
    }
    c->step_index = (USHORT)wStepIndex;
    
    return (SHORT)lPrev_Sample; 
}

STATIC CHAR 
IMAAdpcmCompressSample(IMA_ADPCM_CHAN_STATUS *c, short sample)
{
    CHAR  byCode = 0;
    USHORT uwStepIndex = c->step_index;    
    SHORT  wDelta = sample - c->prev_sample;
    USHORT uwStep = step_table[uwStepIndex];

    if ( wDelta < 0 )
    {
        byCode = 0x08;
        wDelta = -wDelta;
    }

    if ( wDelta >= uwStep )
    {
        byCode |= 0x04;
        wDelta -= uwStep;
    }
    uwStep >>= 1;

    if ( wDelta >= uwStep )
    {
        byCode |= 0x02;
        wDelta -= uwStep;
    }
    uwStep >>= 1;

    if ( wDelta >= uwStep )
    {
        byCode |= 0x01;
    }

    IMAAdpcmExpandSample( c, byCode );

    return byCode;
}


STATIC BOOL
Init()
{
    return TRUE;
}



STATIC BOOL
Config(PSAUDIOCMPPARAMS psCmpParams)
{
    return TRUE;
}

STATIC BOOL
Compress( PSAUDIOCONVINFO psConvInfo )
{
    IMA_ADPCM_CHAN_STATUS sChannelStatus[2];
    BOOL bLowSample = TRUE;
    BOOL bStereoMode = psConvInfo->uiInChannels == 2;
    UINT n = psConvInfo->ulInBufferSize; /* buffer's size in bytes */
    UINT counter = 0;
    UINT step = bStereoMode ? 4 : 2;
    // Source:      Uncompressed data: One sample per 16 bit.
    // Destination: Compressed data:   One sample per 4 bit

    PSHORT pwSamples       = (PSHORT)psConvInfo->pbyInBuffer;
    PBYTE  pbDestination = (PBYTE)psConvInfo->pbyOutBuffer;

    // Write out a Header packet.    
    sChannelStatus[0].prev_sample = *pwSamples++;
    sChannelStatus[0].step_index = gsChannelStatus[0].step_index;
    *pbDestination++ = ( sChannelStatus[0].prev_sample ) & 0xFF; // little endian
    *pbDestination++ = ( sChannelStatus[0].prev_sample >> 8 ) & 0xFF;
    *pbDestination++ = gsChannelStatus[0].step_index;
    *pbDestination++ = 0;
    n -= 2; /* sample's size is 2 bytes, so buffer's size decremented to 2 */

    if ( bStereoMode )
    {
        sChannelStatus[1].prev_sample = *pwSamples++;
        sChannelStatus[1].step_index = gsChannelStatus[1].step_index;
        *pbDestination++ = ( sChannelStatus[1].prev_sample ) & 0xFF; // little endian
        *pbDestination++ = ( sChannelStatus[1].prev_sample >> 8 ) & 0xFF;
        *pbDestination++ = gsChannelStatus[1].step_index;
        *pbDestination++ = 0;
        n -= 2; /* sample's size is 2 bytes, so buffer's size decremented to 2 */
    }

    // stereo: 4 bytes (8 samples) for left, 4 bytes for right, 4 bytes left, ...
    // Mono: 1 byte at a time. 2 samples at a time.
    for ( ; n > 0; n -= step )
    {
        if ( bLowSample )
        {
            pbDestination[0] = IMAAdpcmCompressSample(&sChannelStatus[0], *pwSamples++) & 0x0F;
            if ( bStereoMode )
            {
                pbDestination[4] = IMAAdpcmCompressSample(&sChannelStatus[1], *pwSamples++) & 0x0F;
            }
            bLowSample = FALSE;
        }
        else
        {
            pbDestination[0] |= (IMAAdpcmCompressSample(&sChannelStatus[0], *pwSamples++) << 4) & 0xF0;
            if ( bStereoMode )
            {
                pbDestination[4] |= (IMAAdpcmCompressSample(&sChannelStatus[1], *pwSamples++) << 4) & 0xF0;
                counter = (counter + 1) & 3; /* counter changes from 0 to 3 */
                if ( counter == 0 )
                {
                    pbDestination += 4;
                }
            }
            bLowSample = TRUE;
            ++pbDestination;
        }
    }
    
    gsChannelStatus[0].prev_sample = sChannelStatus[0].prev_sample;
    gsChannelStatus[0].step_index = sChannelStatus[0].step_index;
    if ( bStereoMode )
    {
        gsChannelStatus[1].prev_sample = sChannelStatus[1].prev_sample;
        gsChannelStatus[1].step_index = sChannelStatus[1].step_index;
    }

    return TRUE;
}

STATIC VOID
AdjustStepIndexValue(short * step_index)
{
    if (*step_index < 0)
    {
        *step_index = 0;
    }
    else if (*step_index > wMaxStepIndex)
    {
        *step_index = wMaxStepIndex;
    }
}

// Stereo mode
STATIC VOID
DecompressTwoInTwoOut(IMA_ADPCM_CHAN_STATUS * sChannelStatus, PBYTE  pbySource, UINT uiSourceSize, PSHORT pbyDecompBuff)
{
    for( ; uiSourceSize > 0; uiSourceSize -= 8, pbySource += 8 )
    {
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], pbySource[0] & 0x0F);        
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[1], pbySource[4] & 0x0F);        

        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], (pbySource[0] >> 4) & 0x0F);
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[1], (pbySource[4] >> 4) & 0x0F);

        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], pbySource[1] & 0x0F);        
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[1], pbySource[5] & 0x0F);        

        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], (pbySource[1] >> 4) & 0x0F);
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[1], (pbySource[5] >> 4) & 0x0F);

        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], pbySource[2] & 0x0F);        
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[1], pbySource[6] & 0x0F);        

        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], (pbySource[2] >> 4) & 0x0F);
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[1], (pbySource[6] >> 4) & 0x0F);

        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], pbySource[3] & 0x0F);        
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[1], pbySource[7] & 0x0F);        

        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], (pbySource[3] >> 4) & 0x0F);
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[1], (pbySource[7] >> 4) & 0x0F);
    }
}

// Mono mode
STATIC VOID
DecompressOneInOneOut(IMA_ADPCM_CHAN_STATUS * sChannelStatus, PBYTE  pbySource, UINT uiSourceSize, PSHORT pbyDecompBuff)
{
    for( ; uiSourceSize > 0; --uiSourceSize, ++pbySource )
    {
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], pbySource[0] & 0x0F);        
        *pbyDecompBuff++ = IMAAdpcmExpandSample(&sChannelStatus[0], (pbySource[0] >> 4) & 0x0F);
    }
}

STATIC BOOL
Decompress( PSAUDIOCONVINFO psConvInfo )
{
    // Source:      Compressed data:   One sample per 4 bits.
    // Destination: Uncompressed data: One sample per 16 bit.

    PSHORT pbyDecompBuff  = (PSHORT)psConvInfo->pbyOutBuffer;
    PBYTE  pbySource      = psConvInfo->pbyInBuffer;
    UINT uiSourceSize     = psConvInfo->ulInBufferSize - IMA_ADPCM_HEADER_OVERHEAD * psConvInfo->uiInChannels;

    BOOL bInTwoChannel = (psConvInfo->uiInChannels == 2);
    BOOL bOutTwoChannel = (psConvInfo->uiOutChannels == 2);
    IMA_ADPCM_CHAN_STATUS sChannelStatus[2];

    sChannelStatus[0].prev_sample = *pbySource++;
    sChannelStatus[0].prev_sample |= *pbySource++ << 8;
    sChannelStatus[0].step_index = *pbySource++;
    AdjustStepIndexValue(&sChannelStatus[0].step_index);
    ++pbySource; // skip fourth header's byte

    if ( bInTwoChannel )
    {
        sChannelStatus[1].prev_sample = *pbySource++;
        sChannelStatus[1].prev_sample |= *pbySource++ << 8;
        sChannelStatus[1].step_index = *pbySource++;
        AdjustStepIndexValue(&sChannelStatus[1].step_index);
        ++pbySource; // skip fourth header's byte
    }

    *pbyDecompBuff++ = sChannelStatus[0].prev_sample;

    if ( bOutTwoChannel )
    {
        *pbyDecompBuff++ = sChannelStatus[1].prev_sample;
    }

    /* There are supported Mono->Mono and Stereo->Stereo modes only.
       Mixed modes are not supported now. */
    if ( bInTwoChannel && bOutTwoChannel )
    {
        DecompressTwoInTwoOut(sChannelStatus, pbySource, uiSourceSize, pbyDecompBuff);
    }
    else if ( !bInTwoChannel && !bOutTwoChannel )
    {
        DecompressOneInOneOut(sChannelStatus, pbySource, uiSourceSize, pbyDecompBuff);
    }

    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
// Compression functions
///////////////////////////////////////////////////////////////////////////////

SAUDIOCOMPRESSPROC sIMAAdpcmProc =
{
    Init,
    Config,
    Compress,
    Decompress
};


