/**
                                                                                
2002 ZORAN Corporation, All Rights Reserved THIS IS PROPRIETARY SOURCE CODE OF  
ZORAN CORPORATION
                                                                                
*/

#ifndef AVCODEC_H264Z_H
#define AVCODEC_H264Z_H

#include "av.h"
#include "brcz.h"

///////////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////////

#define H264_MAX_PICTURE_COUNT       6

//////////////////////////////////////////////////////////////////////////////
// Entropy modes
//////////////////////////////////////////////////////////////////////////////

#define H264_ENTROPY_MODE_VLC         0
#define H264_ENTROPY_MODE_CABAC       1

//////////////////////////////////////////////////////////////////////////////
// Slice types
//////////////////////////////////////////////////////////////////////////////

#define H264_SLICE_TYPE_P             5
#define H264_SLICE_TYPE_B             6
#define H264_SLICE_TYPE_I             7
#define H264_SLICE_TYPE_SP            8
#define H264_SLICE_TYPE_SI            9

#define H264_SLICE_TYPE_NOT_DEFINED   10

//////////////////////////////////////////////////////////////////////////////
// NAL unit types
//////////////////////////////////////////////////////////////////////////////

#define H264_NAL_SLICE                1
#define H264_NAL_IDR_SLICE            5
#define H264_NAL_SEI                  6
#define H264_NAL_SPS                  7
#define H264_NAL_PPS                  8
#define H264_NAL_AUD                  9
#define H264_NAL_END_SEQUENCE        10
#define H264_NAL_END_STREAM          11

//////////////////////////////////////////////////////////////////////////////
// SEI Messages payload types
//////////////////////////////////////////////////////////////////////////////
#define H264_SEI_BUFFERING_PERIOD       0
#define H264_SEI_PICTURE_TIMING         1
#define H264_SEI_RECOVERY_POINT         6
// MEP decisions
//////////////////////////////////////////////////////////////////////////////

#define H264_MP_LUT_SIZE_INTER_THRESHOLDS    52
#define H264_MP_LUT_SIZE_ZERO_THRESHOLDS     52
#define H264_DC_VARIENCE_WEIGHTS             8
#define H264_INTRA_SAD_THRESHOLD_LUT         6

#if ( FW_VERSION == FW_VERSION_C12P )
#define H264_MAX_WIDTH  1920
#define H264_MAX_HEIGHT 1088
#else
#define H264_MAX_WIDTH  1280
#define H264_MAX_HEIGHT  720
#endif

// For C10 - we support only one SPS and one PPS (originally: 32)
#define MAX_SPS_COUNT 1
#define MAX_PPS_COUNT 1


/**
 * HRD
 */
typedef struct HRD{

	int cpb_cnt_minus1;
	int bit_rate_scale;
	int cpb_size_scale;

	// limit for C10: cpb_cnt_minus1 could be 0 and 1;
	int bit_rate_value_minus1[2];
	int cpb_size_value_minus1[2];
	int cbr_flag[2];

	uint8_t initial_cpb_removal_delay_length_minus1;
	uint8_t cpb_removal_delay_length_minus1;

	uint8_t dpb_output_delay_length_minus1;
	uint8_t time_offset_length;
}HRD;

/**
 * Video Usability Informations
 */
typedef struct VUI{
	int aspect_ratio_info_present_flag;
    int aspect_ratio_idc;
	int sar_width;
	int sar_height;

	uint8_t overscan_info_present_flag;
	uint8_t overscan_appropriate_flag;

	uint8_t video_signal_type_present_flag;
	uint8_t video_format;
	uint8_t video_full_range_flag;
	uint8_t colour_description_present_flag;
	uint8_t colour_primaries;
	uint8_t transfer_characteristics;
	uint8_t matrix_coefficients;

	uint8_t chroma_loc_info_present_flag;
	int chroma_sample_loc_type_top_field;
	int chroma_sample_loc_type_bottom_field;

	int timing_info_present_flag;

	int num_units_in_tick;
	int time_scale;
	int fixed_frame_rate_flag;

	int nal_hrd_parameters_present_flag;
	HRD nal_hrd_parameters;

	int vcl_hrd_parameters_present_flag;
	HRD vcl_hrd_parameters;

	uint8_t low_delay_hrd_flag;
	uint8_t pic_struct_present_flag;
	uint8_t bitstream_restriction_flag;
	uint8_t motion_vectors_over_pic_boundaries_flag;
	int max_bytes_per_pic_denom;

	int max_bits_per_mb_denom;
	int log2_max_mv_length_horizontal;
	int log2_max_mv_length_vertical;
	int num_reorder_frames;

	int max_dec_frame_buffering;
}VUI;

/**
 * Sequence parameter set
 */
typedef struct SPS{

    int profile_idc;
    int level_idc;
    int transform_bypass;              ///< qpprime_y_zero_transform_bypass_flag
    int log2_max_frame_num;            ///< log2_max_frame_num_minus4 + 4
    int poc_type;                      ///< pic_order_cnt_type
    int log2_max_poc_lsb;              ///< log2_max_pic_order_cnt_lsb_minus4
    int delta_pic_order_always_zero_flag;
    int offset_for_non_ref_pic;
    int offset_for_top_to_bottom_field;
    int poc_cycle_length;              ///< num_ref_frames_in_pic_order_cnt_cycle
    int ref_frame_count;               ///< num_ref_frames
    int gaps_in_frame_num_allowed_flag;
    int mb_width;                      ///< frame_width_in_mbs_minus1 + 1
    int mb_height;                     ///< frame_height_in_mbs_minus1 + 1
    int frame_mbs_only_flag;
    int mb_aff;                        ///<mb_adaptive_frame_field_flag
    int direct_8x8_inference_flag;
    int crop;                   ///< frame_cropping_flag
    int crop_left;              ///< frame_cropping_rect_left_offset
    int crop_right;             ///< frame_cropping_rect_right_offset
    int crop_top;               ///< frame_cropping_rect_top_offset
    int crop_bottom;            ///< frame_cropping_rect_bottom_offset
    int vui_parameters_present_flag;
//    AVRational sar;
    int timing_info_present_flag;
    uint32_t num_units_in_tick;
    uint32_t time_scale;
    int fixed_frame_rate_flag;
    short offset_for_ref_frame[256]; //FIXME dyn aloc?
    int bitstream_restriction_flag;
    int num_reorder_frames;
    int scaling_matrix_present;
    uint8_t scaling_matrix4[6][16];
    uint8_t scaling_matrix8[2][64];

	VUI vui_parameters;
}SPS;

/**
 * Picture parameter set
 */
typedef struct PPS{
    int sps_id;
    int cabac;                  ///< entropy_coding_mode_flag
    int pic_order_present;      ///< pic_order_present_flag
    int slice_group_count;      ///< num_slice_groups_minus1 + 1
    int mb_slice_group_map_type;
    int ref_count[2];           ///< num_ref_idx_l0/1_active_minus1 + 1
    int weighted_pred;          ///< weighted_pred_flag
    int weighted_bipred_idc;
    int init_qp;                ///< pic_init_qp_minus26 + 26
    int init_qs;                ///< pic_init_qs_minus26 + 26
    int chroma_qp_index_offset;
    int deblocking_filter_parameters_present; ///< deblocking_filter_parameters_present_flag
    int constrained_intra_pred; ///< constrained_intra_pred_flag
    int redundant_pic_cnt_present; ///< redundant_pic_cnt_present_flag
    int transform_8x8_mode;     ///< transform_8x8_mode_flag
    uint8_t scaling_matrix4[6][16];
    uint8_t scaling_matrix8[2][64];
}PPS;

/**
 * H264Context
 */
typedef struct H264Context{
//    MpegEncContext s;
    int nal_ref_idc;
    int nal_unit_type;
    uint8_t *rbsp_buffer;
    unsigned int rbsp_buffer_size;

    /**
      * Used to parse AVC variant of h264
      */
    int is_avc; ///< this flag is != 0 if codec is avc1
    int got_avcC; ///< flag used to parse avcC data only once
    int nal_length_size; ///< Number of bytes used for nal length (1, 2 or 4)

    int chroma_qp; //QPc

    int prev_mb_skipped;
    int next_mb_skipped;

    //prediction stuff
    int chroma_pred_mode;
    int intra16x16_pred_mode;

    int top_mb_xy;
    int left_mb_xy[2];

    int8_t intra4x4_pred_mode_cache[5*8];
    int8_t (*intra4x4_pred_mode)[8];
    void (*pred4x4  [9+3])(uint8_t *src, uint8_t *topright, int stride);//FIXME move to dsp?
    void (*pred8x8l [9+3])(uint8_t *src, int topleft, int topright, int stride);
    void (*pred8x8  [4+3])(uint8_t *src, int stride);
    void (*pred16x16[4+3])(uint8_t *src, int stride);
    unsigned int topleft_samples_available;
    unsigned int top_samples_available;
    unsigned int topright_samples_available;
    unsigned int left_samples_available;
    uint8_t (*top_borders[2])[16+2*8];
    uint8_t left_border[2*(17+2*9)];

    /**
     * non zero coeff count cache.
     * is 64 if not available.
     */
//    DECLARE_ALIGNED_8(uint8_t, non_zero_count_cache[6*8]);
    uint8_t (*non_zero_count)[16];

    /**
     * Motion vector cache.
     */
//    DECLARE_ALIGNED_8(int16_t, mv_cache[2][5*8][2]);
//    DECLARE_ALIGNED_8(int8_t, ref_cache[2][5*8]);
//#define LIST_NOT_USED -1 //FIXME rename?
//#define PART_NOT_AVAILABLE -2

    /**
     * is 1 if the specific list MV&references are set to 0,0,-2.
     */
    int mv_cache_clean[2];

    /**
     * number of neighbors (top and/or left) that used 8x8 dct
     */
    int neighbor_transform_size;

    /**
     * block_offset[ 0..23] for frame macroblocks
     * block_offset[24..47] for field macroblocks
     */
    int block_offset[2*(16+8)];

    uint32_t *mb2b_xy; //FIXME are these 4 a good idea?
    uint32_t *mb2b8_xy;
    int b_stride; //FIXME use s->b4_stride
    int b8_stride;

    int mb_linesize;   ///< may be equal to s->linesize or s->linesize*2, for mbaff
    int mb_uvlinesize;

    int emu_edge_width;
    int emu_edge_height;

    int halfpel_flag;
    int thirdpel_flag;

    int unknown_svq3_flag;
    int next_slice_index;

    SPS sps_buffer[MAX_SPS_COUNT];
    SPS sps; ///< current sps

    PPS pps_buffer[MAX_PPS_COUNT];
    /**
     * current pps
     */
    PPS pps; //FIXME move to Picture perhaps? (->no) do we need that?

    uint32_t dequant4_buffer[6][52][16];
    uint32_t dequant8_buffer[2][52][64];
    uint32_t (*dequant4_coeff[6])[16];
    uint32_t (*dequant8_coeff[2])[64];
    int dequant_coeff_pps;     ///< reinit tables when pps changes

    int slice_num;
    uint8_t *slice_table_base;
    uint8_t *slice_table;      ///< slice_table_base + 2*mb_stride + 1
    int slice_type;
    int slice_type_fixed;

    //interlacing specific flags
    int mb_aff_frame;
    int mb_field_decoding_flag;
    int mb_mbaff;              ///< mb_aff_frame && mb_field_decoding_flag

    int sub_mb_type[4];

    //POC stuff
    int poc_lsb;
    int poc_msb;
    int delta_poc_bottom;
    int delta_poc[2];
    int frame_num;
    int prev_poc_msb;             ///< poc_msb of the last reference pic for POC type 0
    int prev_poc_lsb;             ///< poc_lsb of the last reference pic for POC type 0
    int frame_num_offset;         ///< for POC type 2
    int prev_frame_num_offset;    ///< for POC type 2
    int prev_frame_num;           ///< frame_num of the last pic for POC type 1/2

    /**
     * frame_num for frames or 2*frame_num for field pics.
     */
    int curr_pic_num;

    /**
     * max_frame_num or 2*max_frame_num for field pics.
     */
    int max_pic_num;

    //Weighted pred stuff
    int use_weight;
    int use_weight_chroma;
    int luma_log2_weight_denom;
    int chroma_log2_weight_denom;
    int luma_weight[2][48];
    int luma_offset[2][48];
    int chroma_weight[2][48][2];
    int chroma_offset[2][48][2];
    int implicit_weight[48][48];

    //deblock
    int deblocking_filter;         ///< disable_deblocking_filter_idc with 1<->0
    int slice_alpha_c0_offset;
    int slice_beta_offset;

    int redundant_pic_count;

    int direct_spatial_mv_pred;
    int dist_scale_factor[16];
    int dist_scale_factor_field[32];
    int map_col_to_list0[2][16];
    int map_col_to_list0_field[2][32];

    /**
     * num_ref_idx_l0/1_active_minus1 + 1
     */
    int ref_count[2];            ///< counts frames or fields, depending on current mb mode
    Picture *short_ref[32];
    Picture *long_ref[32];
    Picture default_ref_list[2][32];
    Picture ref_list[2][48];     ///< 0..15: frame refs, 16..47: mbaff field refs
    Picture *delayed_pic[16]; //FIXME size?
    Picture *delayed_output_pic;

    /**
     * memory management control operations buffer.
     */
//    MMCO mmco[MAX_MMCO_COUNT];
//    int mmco_index;

    int long_ref_count;  ///< number of actual long term references
    int short_ref_count; ///< number of actual short term references

    //data partitioning
    GetBitContext intra_gb;
    GetBitContext inter_gb;
    GetBitContext *intra_gb_ptr;
    GetBitContext *inter_gb_ptr;

//    DECLARE_ALIGNED_8(DCTELEM, mb[16*24]);

    /**
     * Cabac
     */
//    CABACContext cabac;
//    uint8_t      cabac_state[460];
    int          cabac_init_idc;

    /* 0x100 -> non null luma_dc, 0x80/0x40 -> non null chroma_dc (cb/cr), 0x?0 -> chroma_cbp(0,1,2), 0x0? luma_cbp */
    uint16_t     *cbp_table;
    int top_cbp;
    int left_cbp;
    /* chroma_pred_mode for i4x4 or i16x16, else 0 */
    uint8_t     *chroma_pred_mode_table;
    int         last_qscale_diff;
    int16_t     (*mvd_table[2])[2];
//    DECLARE_ALIGNED_8(int16_t, mvd_cache[2][5*8][2]);
    uint8_t     *direct_table;
    uint8_t     direct_cache[5*8];

    uint8_t zigzag_scan[16];
    uint8_t zigzag_scan8x8[64];
    uint8_t zigzag_scan8x8_cavlc[64];
    uint8_t field_scan[16];
    uint8_t field_scan8x8[64];
    uint8_t field_scan8x8_cavlc[64];
    const uint8_t *zigzag_scan_q0;
    const uint8_t *zigzag_scan8x8_q0;
    const uint8_t *zigzag_scan8x8_cavlc_q0;
    const uint8_t *field_scan_q0;
    const uint8_t *field_scan8x8_q0;
    const uint8_t *field_scan8x8_cavlc_q0;

    int x264_build;
}H264Context;

///////////////////////////////////////////////////////////////////////////////
// The following structure contains an information for frame encoding/decoding
// by h264z codec.
// This structure is used as AVCODECCONTEXT private data for h264z codec.
///////////////////////////////////////////////////////////////////////////////

typedef struct H264zEncContext{

    struct AVCodecContext *avctx;

    int codec_id;

    // If 1 - encoding mode
    // If 0 - decoding mode    
    int encoding; 

    // If 1 - VLC
    // If 0 - CABAC
    int entropy_mode; 
    
    // Whether headers (sequence parameters set and pictrue parameters set)
    // have already been written for this clip
    int headers_written;

    // Whether headers (sequence parameters set and pictrue parameters set)
    // have already been read for this clip
    int headers_read;
    
    // gaps_in_frame_num_value_allowed_flag - gap between consecutive frames numbers
    int gaps_in_frame_num_value_allowed_flag;

    // Maximum frame_num in an IDR
    int max_frame_num;

    // Current frame number
    int frame_num;

    int real_frame_num;

    // 
    int max_pic_order_cnt_lsb;

    //
    int gop_start;

    //
    int flags; 

    // Max number of b-frames for encoding
    int max_b_frames;
                            
    // Decoder bitstream context
    GetBitContext gb;

    ParseContext parse_context;

    // Picture buffer
    AVFrame picture[H264_MAX_PICTURE_COUNT];
                     
    // Pointer to forward reference frame
    AVFrame *last_picture_ptr;

    // Pointer to backward reference frame
    AVFrame *next_picture_ptr;

    // Pointer to currently encoded/decoded frame. 
    AVFrame *current_picture_ptr;

    // Pointer to just-finished encoded/decoded frame. 
    AVFrame *ready_picture_ptr;

    // Pointer to just-finished encoded/decoded frame's forward reference. 
    AVFrame *buffered_picture_ptr[2];

    // Copy of picture, which is currently encoded/decoded.
    // This structure is an application interface, to reveal information,
    // about currently coded picture. (Like picture type)
    AVFrame coded_picture;

    // Picture size, must be a multiple of 16
    int physical_width;
    int physical_height;

    int logical_width;
    int logical_height;

    // Picture initialization quantization paramter
    // decoder: TBD
    int pic_init_qp;

    // Slice delta in quantization parameter
    int slice_qp_delta;

    // Number of macro blocks in frame.
    int mb_in_frame; 
 
    // Currently coded picture type: I_TYPE, P_TYPE, B_TYPE
    int pict_type; 

    // Currently coded slice type: H264_SLICE_TYPE_I, H264_SLICE_TYPE_P, H264_SLICE_TYPE_B
    int slice_type;

    // MV resolution.
    int f_code;

    // Maximum motion estimation search range in pixel units. The search size
    // is for single side of macro block.
    int me_search_x;            
    int me_search_y;     

    // Bit Rate Control data
    SBRCCONTEXT brc_ctx;

    // Indicates for decoder to skip frames.
    int hurry_up;

    // Deblocking filter
    int deblocking_filter;         ///< disable_deblocking_filter_idc with 1<->0
    int slice_alpha_c0_offset;
    int slice_beta_offset;

    // Cabac
    int cabac_init_idc;

    // no reordering needed / has no b-frames
    int low_delay; 

    int bitstream_buffer_size;

    // Notifies for decoder if codec was already initialized.
    int context_initialized;       

    // ============================
    int nal_unit_type;
    int nal_ref_idc;
    uint8_t *rbsp_buffer;
    unsigned int rbsp_buffer_size;

    int slice_header_offset;
    int slice_header_bits;

    // Current sps
    SPS sps_buffer[MAX_SPS_COUNT];
    SPS sps; 

    // Current pps
    PPS pps_buffer[MAX_PPS_COUNT];
    PPS pps; 

    // Quantizer
    int qscale;
    int chroma_qp;

    int decoder_init_done;

	int gop_i_frame;

    // Used by decoder for deciding frame_b_type 
    int pic_order_cnt;
    int prev_pic_order_cnt;

    // Used by encoder for deciding frame_b_type 
    int prev_display_picture_number;

	uint8_t *saved_recy_buffer;
	uint8_t *saved_recc_buffer;
	uint8_t *saved_dec_buffer;


} H264zEncContext;

int
h264z_encode_slice_header_2vc(H264zEncContext *s, AVFrame* pic, int slice_type, BOOL bFirstTime );


void
h264z_encode_custom_nals(H264zEncContext *s );

void
h264_encode_skip_frame(H264zEncContext *s, AVFrame* picture);

int
h264z_get_slice_type(int pict_type);

void
h264z_add_stuffing_bits(H264zEncContext *s, PSAVFRAME pic_ptr, UINT uiCodeSize );

int
h264z_encode_preconfig(AVCodecContext *avctx,
                     unsigned char *buf,
                     int buf_size,
                     AVFrame *data,
                     BOOL bFirstTime );

void
h264_write_sei_messages(AVCodecContext *avctx,
                       AVFrame *data,
                       BOOL bFirstTime );


#endif // AVCODEC_H264Z_H

