/*
 * Generic DCT based hybrid video encoder
 * Copyright (c) 2000, 2001, 2002 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file mpegvideo.h
 * mpegvideo header.
 */

#include "AvCodec.h"
 
#ifndef AVCODEC_MPEGVIDEO_H
#define AVCODEC_MPEGVIDEO_H

#define FRAME_SKIPED 100 ///< return value for header parsers if frame is not coded

#define I_TYPE          FF_I_TYPE  ///< Intra
#define P_TYPE          FF_P_TYPE  ///< Predicted
#define B_TYPE          FF_B_TYPE  ///< Bi-dir predicted
#define P_TYPE_DUMMY    ( FF_P_TYPE | 0x10 )

/**
 * Picture.
 */
typedef struct Picture
{
    FF_COMMON_FRAME;

	// just to be at the same size as AVFrame
	struct Picture* dummy_ptr;

} Picture;

typedef struct ParseContext
{
    uint8_t *buffer;
    int index;
    int last_index;
    int buffer_size;
    uint32_t state;             ///< contains the last few bytes in MSB order
    int frame_start_found;
    int overread;               ///< the number of bytes which where irreversibly read from the next frame
    int overread_index;         ///< the index into ParseContext.buffer of the overreaded bytes
} ParseContext;

/**
 * JpegEncContext.
 */
typedef struct JpegEncContext 
{

    struct AVCodecContext *avctx;

} JpegEncContext;

#endif /* AVCODEC_MPEGVIDEO_H */
