// This file is part of the camera os components sources.
#include "Audio.h"
//#include "Component.h"
#include "Allocator.h"
#include "debug.h"

/* These are for MS-ADPCM */
static const int AdaptationTable[] = {
        230, 230, 230, 230, 307, 409, 512, 614,
        768, 614, 512, 409, 307, 230, 230, 230
};

static const int AdaptCoeff1[] = {
        256, 512, 0, 192, 240, 460, 392
};

static const int AdaptCoeff2[] = {
        0, -256, 0, 64, 0, -208, -232
};

#define CLAMP_TO_SHORT(value) \
if (value > 32767) \
    value = 32767; \
else if (value < -32768) \
    value = -32768; \
    
typedef struct MS_ADPCM_CHAN_STATUS {
    int predictor;
    short int step_index;
    int step;
    /* for encoding */
    int prev_sample;

    /* MS version */
    short sample1;
    short sample2;
    int coeff1;
    int coeff2;
    int idelta;
} MS_ADPCM_CHAN_STATUS;

/**
                                                                                
%function                                                                       
                                                                                
  Clip
                                                                                
%desc
                                                                                
*/
STATIC INT
Clip(int a, int amin, int amax)
{
    if (a < amin)
        return amin;
    else if (a > amax)
        return amax;
    else
        return a;
}

/**
                                                                                
%function                                                                       
                                                                                
  MsAdpcmCompressSample
                                                                                
%desc
                                                                                
*/
STATIC CHAR 
MsAdpcmCompressSample(MS_ADPCM_CHAN_STATUS *c, short sample)
{
    int predictor, nibble, bias;

    predictor = (((c->sample1) * (c->coeff1)) + ((c->sample2) * (c->coeff2))) / 256;
    
    nibble= sample - predictor;
    
    if(nibble>=0) 
    {
        bias= c->idelta/2;
    }
    else          
    {
        bias=-c->idelta/2;
    }
        
    nibble= (nibble + bias) / c->idelta;
    nibble= Clip(nibble, -8, 7)&0x0F;
    
    predictor += (signed)((nibble & 0x08)?(nibble - 0x10):(nibble)) * c->idelta;
    CLAMP_TO_SHORT(predictor);

    c->sample2 = c->sample1;
    c->sample1 = predictor;

    c->idelta = (AdaptationTable[(int)nibble] * c->idelta) >> 8;
    
    if (c->idelta < 16)
    {
        c->idelta = 16;
    }

    return nibble;
}

/**
                                                                                
%function                                                                       
                                                                                
  MsAdpcmExpandNibble
                                                                                
%desc
                                                                                
*/
STATIC SHORT
MsAdpcmExpandNibble(MS_ADPCM_CHAN_STATUS *c, char nibble)
{
    int predictor=0;

    predictor = (((c->sample1) * (c->coeff1)) + ((c->sample2) * (c->coeff2))) / 256;
    predictor += (signed)((nibble & 0x08)?(nibble - 0x10):(nibble)) * c->idelta;
    CLAMP_TO_SHORT(predictor);

    c->sample2 = c->sample1;
    c->sample1 = predictor;
    c->idelta = (AdaptationTable[(int)nibble] * c->idelta) >> 8;
    
    if (c->idelta < 16) 
    {
        c->idelta = 16;
    }
    
    return (short)predictor;
}


/**
                                                                                
%function                                                                       
                                                                                
  Init
                                                                                
%desc
                                                                                
*/

STATIC BOOL
Init()
{
    return TRUE;
}

/**
                                                                                
%function                                                                       
                                                                                
  Config

*/

STATIC BOOL
Config(PSAUDIOCMPPARAMS psCmpParams)
{
    return TRUE;
}

/**
                                                                                
%function                                                                       
                                                                                
  Compress
                                                                                
*/

STATIC BOOL
Compress( PSAUDIOCONVINFO psConvInfo )
{
    UINT i;   

    // Source:      Uncompressed data: One sample per 16 bit.
    // Destination: Compressed data:   One sample per 4 bit

    PSHORT samples       = (PSHORT) psConvInfo->pbyInBuffer;
    PBYTE  pbDestination = (PBYTE)psConvInfo->pbyOutBuffer;

    INT st= psConvInfo->uiInChannels == 2;

    MS_ADPCM_CHAN_STATUS sChannelStatus[2];

    for(i=0; i<psConvInfo->uiOutChannels; i++)
    {
        int predictor=0;
        *pbDestination++ = (BYTE)predictor;
        sChannelStatus[i].coeff1 = AdaptCoeff1[predictor];
        sChannelStatus[i].coeff2 = AdaptCoeff2[predictor];
    }
     
    for(i=0; i<psConvInfo->uiOutChannels; i++)
    {
        sChannelStatus[i].idelta = 16;

        *pbDestination++ = sChannelStatus[i].idelta & 0xFF;
        *pbDestination++ = sChannelStatus[i].idelta >> 8;
    }
    
    for(i=0; i<psConvInfo->uiOutChannels; i++)
    {
        sChannelStatus[i].sample1= *samples++;

        *pbDestination++ = sChannelStatus[i].sample1 & 0xFF;
        *pbDestination++ = sChannelStatus[i].sample1 >> 8;
    }
    
    for(i=0; i<psConvInfo->uiOutChannels; i++)
    {
        sChannelStatus[i].sample2= *samples++;

        *pbDestination++ = sChannelStatus[i].sample2 & 0xFF;
        *pbDestination++ = sChannelStatus[i].sample2 >> 8;
    }
    
    for(i=7*psConvInfo->uiOutChannels; i<1024; i++) 
    {
        int nibble;
        nibble = MsAdpcmCompressSample(&sChannelStatus[0], *samples++)<<4;
        nibble|= MsAdpcmCompressSample(&sChannelStatus[st], *samples++);
        *pbDestination++ = nibble;
    }
    
    return TRUE;
}

/**
                                                                                
%function                                                                       
                                                                                
  Decompress
                                                                                
*/

STATIC BOOL
Decompress( PSAUDIOCONVINFO psConvInfo )
{
    // Source:      Compressed data:   One sample per 4 bits.
    // Destination: Uncompressed data: One sample per 16 bit.

    PSHORT samples = (PSHORT) psConvInfo->pbyOutBuffer;
    PBYTE  pbySource     = psConvInfo->pbyInBuffer;
    INT iSourceSize  = psConvInfo->ulInBufferSize - 7*psConvInfo->uiInChannels;

    BOOL bInTwoChannel = (psConvInfo->uiInChannels == 2);
    BOOL bOutTwoChannel = (psConvInfo->uiOutChannels == 2);
    MS_ADPCM_CHAN_STATUS sChannelStatus[2];
    INT block_predictor[2];

    if (iSourceSize < 0)
    {
        return FALSE;
    }
    
    block_predictor[0] = Clip(*pbySource++, 0, 6);
    block_predictor[1] = 0; 
           
    if (bInTwoChannel)
    {
        block_predictor[1] = Clip(*pbySource++, 0, 6);
    }
    
    sChannelStatus[0].idelta = (SHORT)((*pbySource & 0xFF) | ((pbySource[1] << 8) & 0xFF00));
    pbySource += 2;        
    
    if (bInTwoChannel)
    {
        sChannelStatus[1].idelta = (SHORT)((*pbySource & 0xFF) | ((pbySource[1] << 8) & 0xFF00));
        pbySource+=2;
    }
    
    sChannelStatus[0].coeff1 = AdaptCoeff1[block_predictor[0]];
    sChannelStatus[0].coeff2 = AdaptCoeff2[block_predictor[0]];
    sChannelStatus[1].coeff1 = AdaptCoeff1[block_predictor[1]];
    sChannelStatus[1].coeff2 = AdaptCoeff2[block_predictor[1]];
    
    sChannelStatus[0].sample1 = ((*pbySource & 0xFF) | ((pbySource[1] << 8) & 0xFF00));
    pbySource+=2;
    
    if (bInTwoChannel) 
    {
        sChannelStatus[1].sample1 = ((*pbySource & 0xFF) | ((pbySource[1] << 8) & 0xFF00));
        pbySource+=2;
    }
    
    sChannelStatus[0].sample2 = ((*pbySource & 0xFF) | ((pbySource[1] << 8) & 0xFF00));
    pbySource+=2;
    
    if (bInTwoChannel) 
    {
        sChannelStatus[1].sample2 = ((*pbySource & 0xFF) | ((pbySource[1] << 8) & 0xFF00));
        pbySource+=2;
    }
    
    *samples++ = sChannelStatus[0].sample1;
    
    if (bOutTwoChannel)
    {
        if (bInTwoChannel) 
        {            
            *samples++ = sChannelStatus[1].sample1;
        }
        else
        {
            *samples++ = sChannelStatus[0].sample1;
        }
    }
    
    *samples++ = sChannelStatus[0].sample2;
    
    if (bOutTwoChannel)
    {
        if (bInTwoChannel) 
        {
            *samples++ = sChannelStatus[1].sample2;
        }
        else
        {
            *samples++ = sChannelStatus[0].sample2;
        }
    }
    
    for( ; iSourceSize>0; iSourceSize-- )
    {
        SHORT   wSample = 0;
        
        wSample = MsAdpcmExpandNibble(&sChannelStatus[0], (pbySource[0] >> 4) & 0x0F);
        *samples++ = wSample;
        
        if (bOutTwoChannel)
        {
            if (bInTwoChannel)
            {
                *samples++ = MsAdpcmExpandNibble(&sChannelStatus[1], pbySource[0] & 0x0F);
            }
            else
            {
                //copy left channel to right channel
                *samples++ = wSample;
                wSample = MsAdpcmExpandNibble(&sChannelStatus[0], (pbySource[0] ) & 0x0F);
                *samples++ = wSample;//left channel
                *samples++ = wSample;//right channel
            }
        }
        else
        {
            if (!bInTwoChannel)
            {
                *samples++ = MsAdpcmExpandNibble(&sChannelStatus[0], pbySource[0] & 0x0F);
            }
        }
        
        pbySource ++;
    }
    
    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
// Compression functions
///////////////////////////////////////////////////////////////////////////////

SAUDIOCOMPRESSPROC sMsAdpcmProc =
{
    Init,
    Config,
    Compress,
    Decompress
};

