#include <ctype.h>
#include "common.h"
#include "Av.h"
#include "avcodec.h"
#include "mpegvideo.h"
#include "h264z.h"
#include "golomb.h"

#include "timing.h"
#include "zav.h"
#include "avz.h"

#undef NDEBUG
#include <assert.h>

//#define H264Z_DEBUG_PRINT

// If this define is enabled, the AUD will be written after PPS and SPS,
// and this will result in warning in VEGA ANALYZER, but good playability
// in QuickTime, if the DEFINE is not defined, the AUD will be written
// before SPS & PPS - no warning in Vega, but file not playable in QuickTime
#define AUD_POSITION_AFTER_PPS



//////////////////////////////////////////////////////////////////////////////
// Aspect ration values.
//////////////////////////////////////////////////////////////////////////////

#if 0 /* Unused */
static const uint16_t pixel_aspect[14][2]={
 {0, 1},
 {1, 1},
 {12, 11},
 {10, 11},
 {16, 11},
 {40, 33},
 {24, 11},
 {20, 11},
 {32, 11},
 {80, 33},
 {18, 11},
 {15, 11},
 {64, 33},
 {160,99},
};
#endif /* 0 */

int ue_length[] =
{ 1,
  3, 3,
  5, 5, 5, 5,
  7, 7, 7, 7, 7, 7, 7, 7,
  9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
  11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
  11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
  13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
  13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
  13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
  13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
};


/*************************************************************************/
/*                                                                       */
/*                            H264 Encoder                               */
/*                                                                       */
/*************************************************************************/

//////////////////////////////////////////////////////////////////////////////
// Encoder image buffers allocation/release functions
//////////////////////////////////////////////////////////////////////////////

/**

%func

    encoder_get_buffer

%desc

    The function allocates all needed buffer for frame encoding.

*/

static int
encoder_get_buffer(AVCodecContext *avctx,
                   AVFrame        *pic)
{
    if ( pic->skip_frame == 0 )
    {
    	pic->ref_context = CodecBuffersGet( avctx, NULL, pic->pict_type );
    }

    return 0;
}


void
h264z_add_stuffing_bits(H264zEncContext *s, PSAVFRAME pic_ptr, UINT uiCodeSize )
{
	PBYTE stuf_buf;

	 		// Add stuffing bits if needed
			if ( s->brc_ctx.iStuffingBits > 0 )
			{
                TimingMarkF( TIMING_GROUP_VIDEO_BRC_HRD, "Encoding UNDERFLOW - Add stuffing bits. Frame %d TYPE %s",
                             pic_ptr->coded_picture_number,
                             pic_ptr->pict_type == I_TYPE ? "I" :
                             pic_ptr->pict_type == P_TYPE ? "P" :
                             pic_ptr->pict_type == B_TYPE ? "B" : "ERROR" );

				// we can do such increment cause codesize includes header
				stuf_buf = pic_ptr->code_buf + uiCodeSize;

				// Start stuffing code
                // putbits( 0x01, 24 );
				// putbits( 0x00, 1 );       /* forbidden_zero_bit */
				// putbits( 0x00, 2 );       /* nal_ref_idc */
				// putbits( 12 , 5 );  /* nal_unit_type */
				// 0000 0000 0000 0000 0000 0001 0000 1100 = 0x00 00 01 0C
				*stuf_buf++ = 0x00;
				*stuf_buf++ = 0x00;
				*stuf_buf++ = 0x01;
				*stuf_buf++ = 0x0C;

				*stuf_buf++ = 0xFF;
				*stuf_buf++ = 0xFF;
				*stuf_buf++ = 0xFF;
				*stuf_buf++ = 0xFF;

				// We write stuffing start code & then we write first four stuffing bytes,
				// so we can be sure that all adrresses before (stuf_buf)/4)*4 do not contain
				// garbage, the address for memset should be dword aligned, so now we can
				// safely use (((UINT)stuf_buf/4)*4)) to get dword aligned address

				// should memset get dword aligned address? = to check
				memset( (PBYTE) (((UINT)stuf_buf/4)*4), 0xFF, s->brc_ctx.iStuffingBits / 8);


			}
}
/**

%func

    encoder_release_buffer

%desc

    The function releases all buffers of the given frame and marks
    the entry as free.

*/

static void
encoder_release_buffer(AVCodecContext *avctx,
                       AVFrame        *pic)
{
    // Mark the entry as free.
    pic->buf_allocated = AV_BUF_FREE;

        CodecBuffersRelease( avctx, pic->ref_context );
}

/**

%func

    encoder_release_external_buffer

%desc

    The function releases all external buffers of the given frame.
*/

static void
encoder_release_external_buffer(AVFrame *pic)
{
    int i;

    // Release all references memories.
    for ( i = AV_BUF_EXTERNAL_START; i <= AV_BUF_EXTERNAL_END; i++ )
    {
        PoolFree( NULL, pic->data[i] );
        pic->data[i] = NULL;
    }
}


/**

%func

    decoder_release_external_buffer

%desc

    The function releases all external buffers of the given frame.
*/

static void
decoder_release_external_buffer(AVFrame *pic)
{
    PoolFree( NULL, pic->code_buf );
    pic->code_buf = NULL;
}


static void
decoder_release_picture(AVCodecContext *avctx,
                        AVFrame        *pic)
{

    if ( pic->buf_allocated & AV_BUF_REFERENCE_USED )
    {
        avctx->release_buffer(avctx, pic);
    }

    // Mark the entry as free.
    pic->data[AV_BUF_YUV_IMAGE]   = NULL;
    pic->data[AV_BUF_YUV_IMAGE_C] = NULL;
    pic->buf_allocated   = AV_BUF_FREE;
}

//////////////////////////////////////////////////////////////////////////////
// Decoder & Encoder common functions.
//////////////////////////////////////////////////////////////////////////////

/**

%func

   find_unused_picture

%desc

    The function looks for a free frame in the encoder/decoder frames
    database.

*/

static int
find_unused_picture(H264zEncContext *s, int shared)
{
    int i;

    for ( i = 0; i < H264_MAX_PICTURE_COUNT; i++ )
    {
        //The frame is considered free if there is no YUV buffer
        if ( s->picture[i].buf_allocated == AV_BUF_FREE )
        {
            s->picture[i].buf_allocated = AV_BUF_REFERENCE_USED | AV_BUF_PICTURE_USED;
            break;
        }
    }

    assert( i<H264_MAX_PICTURE_COUNT );

    return i;
}


/**

%func

    alloc_picture

%desc

    The function allocates all buffer needed for frame
    encoding/decoding
*/

static int
alloc_picture(H264zEncContext *s, AVFrame *pic, int shared)
{
    int r;

    // Callback function from user
    r = s->avctx->get_buffer( s->avctx, pic );

    // In encoding mode, the yuv422 buffer is not part of the picture
    // :ZZZ: Check how to handle failure of get_buffer
    if( r < 0 )
    {
        fprintf(stderr, "get_buffer() failed %d\n", r);
        return -1;
    }

    return 0;
}

/**

%func

    free_picture

%desc

    deallocates a picture
*/

static void
free_picture(H264zEncContext *s, AVFrame *pic)
{
    if ( pic->buf_allocated != AV_BUF_FREE )
    {
        s->avctx->release_buffer(s->avctx, pic);
        decoder_release_picture( s->avctx, pic );
    }
}

/**

%func

    h264z_common_end

%desc

    The function ends encoding/decoding sequence, release all codec's memories.
    The function called after last frame was encoded/decoded.
*/

void
h264z_common_end(H264zEncContext *s)
{
    int i;

    for( i = 0; i < H264_MAX_PICTURE_COUNT; i++ )
    {
        free_picture(s, &s->picture[i]);
    }
}

/**

%func

    h264z_common_init

%desc

    The function initialize common structure for both encoder and decoder
    The function is called before first frame is given to encoder/decoder

*/
int
h264z_common_init(H264zEncContext *s)
{
    s->flags = s->avctx->flags;

    // convert fourcc to upper case */
    s->avctx->codec_tag=   toupper( s->avctx->codec_tag     &0xFF)
                        + (toupper((s->avctx->codec_tag>>8 )&0xFF)<<8 )
                        + (toupper((s->avctx->codec_tag>>16)&0xFF)<<16)
                        + (toupper((s->avctx->codec_tag>>24)&0xFF)<<24);

    // Initialize the pointer to the currently coded picture.
    s->avctx->coded_frame= &s->coded_picture;

    s->parse_context.state = (uint32_t)-1;

	//RL: to update
	memset(&s->sps.vui_parameters, 0, sizeof(s->sps.vui_parameters));

    return 0;
}

/*************************************************************
*
* Common functions
*
*************************************************************/

/*

    ff_h264_stuffing

*/

void
ff_h264_stuffing(PutBitContext *pbc)
{
    int length;

    put_bits(pbc, 1, 1);

    length = (-get_bit_count(pbc)) & 0x7;
    if(length)
    {
        put_bits(pbc, length, 0);
    }
}

/*

    ffff_h264_stuffing

*/

void
ffff_h264_stuffing(PutBitContext *pbc)
{
    int length;

    put_bits(pbc, 1, 1);

    length = (-get_bit_count(pbc)) & 0xF;
    if(length)
    {
        put_bits(pbc, length, 0);
    }
}

/*

    cabac_alignment

*/

void
cabac_alignment(PutBitContext *pbc)
{
    int length;
    length = (-get_bit_count(pbc)) & 7;

    if(length)
    {
        put_bits(pbc, length, (1<<length)-1);
    }
}

int get_ue_value_length(int value)
{
    unsigned int n1 = value + 1;
    int bins;
    int final_bits;

    for( bins = 0; n1 != 0; bins++, n1 >>= 1 )
      ;

    final_bits = bins * 2 - 1 ;

    return final_bits;
}


/*

    put_ue_bits

*/

void
put_ue_bits(PutBitContext *s, int value)
{
    int ue_value = value + 1;
    int ue_value_length;

	if(value > (int)(sizeof(ue_length)/sizeof(ue_length[0])))
	{
		// exception: out of ue_length table range values we are looking
		//            through get_ue_value_length() with an exception list
		ue_value_length = get_ue_value_length(value);

		if(ue_value_length > 32){
			put_bits( s, ue_value_length - 32, 0);
			ue_value_length = 32;
		}

	}
	else
	{
		ue_value_length = ue_length[value];
	}

    // Put ue value in bitstream
    put_bits( s, ue_value_length, ue_value );

}

/*

    put_se_bits

*/

void
put_se_bits(PutBitContext *s, int value)
{
    int se_value = value*2;
    int se_value_length;

    if (value <= 0)
    {
        se_value = 0 - se_value + 1;
    }

    se_value_length = ue_length[se_value-1];

    // Put ue value in bitstream
    put_bits( s, se_value_length, se_value );
}

int
h264z_encode_get_gop_size(H264zEncContext *s)
{
	return s->avctx->gop_size;
}


/*

    get_b_type

*/

static int
decide_frame_b_type(H264zEncContext *s)
{
    int diff   = 0;
    int b_type = 0;

    if ( s->encoding == 0 )
    {   // In decoding, we use picture_order_count to decide on B_Type 
        diff = s->prev_pic_order_cnt - s->pic_order_cnt;
        s->prev_pic_order_cnt = s->pic_order_cnt;

        // in interlacd mode, poc increments by 2 each frame
        if ( s->flags & CODEC_FLAG_INTERLACED_MODE )
        {
            diff /= 2;
    }
    }
    else
    {   // In encoding, we use display_picture_number to decide on B_Type
        diff = s->prev_display_picture_number - s->current_picture_ptr->display_picture_number;
        s->prev_display_picture_number = s->current_picture_ptr->display_picture_number;
    }

    if( s->pict_type == B_TYPE )
    {
        switch ( diff )
        {
            case 1: // IPB ==> 0 2 1
                b_type = AV_FRAME_TYPE_B_SINGLE;   
                break;
            case 2: // IPBB ==> 0 3 1 2
                b_type = AV_FRAME_TYPE_B_DOUBLE_1;
                break;
            default: //
                b_type = AV_FRAME_TYPE_B_DOUBLE_2;
        }
    }

    return b_type;
}


/*

%func

    h264z_frame_start

%desc

    Generic function for encode/decode. The function is called
    before each frame is coded/decoded

*/

int
h264z_frame_start(H264zEncContext *s, AVCodecContext *avctx)
{
    AVFrame *pic;
    int i;

	int GOP_size = h264z_encode_get_gop_size(s);

    // Mark and release old frames.
    // ( When current frame is reference frame (I or P) and there are already
    // 2 reference frame, we can release the oldest reference frame)
    if ( ( s->pict_type != B_TYPE ) &&
         (( s->current_picture_ptr != NULL ) && ( s->current_picture_ptr->skip_frame == 0 )) &&
         ( s->last_picture_ptr != NULL ) )
    {
        avctx->release_buffer( avctx, s->last_picture_ptr );

        /* release forgotten pictures */
        /* if(mpeg124/h263) */
        // :ZZZ: Check what this case means.
        //if ( s->encoding == 0 )
        //{
        //    for( i=0; i<H264_MAX_PICTURE_COUNT; i++ )
        //    {
        //        if( ( s->picture[i].buf_allocated != AV_BUF_FREE ) &&
        //            &s->picture[i] != s->next_picture_ptr &&
        //            s->picture[i].reference )
        //        {
        //            fprintf(stderr, "releasing zombie picture\n");
        //            avctx->release_buffer(avctx, &s->picture[i]);
        //        }
        //    }
        //}
    }

alloc:

    // If decoding mode.
    if ( s->encoding == 0 )
    {

        // Find a free picture
        i = find_unused_picture(s, 0);

        pic = &s->picture[i];

        // Reset reference context pointer
        pic->ref_context = NULL;

        pic->reference = ( s->pict_type != B_TYPE ) ? 3 : 0;
        pic->pict_type = s->pict_type;

        if( s->current_picture_ptr != NULL )
        {
            // Set picture number in decoding order, according to the previously
            // decoded frame number.
            pic->coded_picture_number = s->current_picture_ptr->coded_picture_number + 1;
        }

        s->current_picture_ptr = pic;
        s->ready_picture_ptr = pic;

        alloc_picture(s, s->current_picture_ptr, 0);

    }
    else
    {
        // allocate reference context for current picture
        alloc_picture(s, s->current_picture_ptr, 0 );
    }

    // Update Frame B type
    s->current_picture_ptr->frame_b_type = decide_frame_b_type(s);

    // We update this parameter for application interface to be compatible
    // with other codecs.
    s->current_picture_ptr->key_frame = (s->pict_type == I_TYPE) &&
                                        (s->current_picture_ptr->frame_in_gop % GOP_size == 0);

    // Copy the currently encoded picture to the application interface
    // structure.
    s->coded_picture = *s->current_picture_ptr;

    // Skipped frame requires special treatment
    if ( ( s->pict_type == P_TYPE ) && ( s->current_picture_ptr->skip_frame == 1 ) )
    {
        // This update is very important for direct mode of B frame.
        s->next_picture_ptr->display_picture_number = s->current_picture_ptr->display_picture_number;
        return 0;
    }

    // Handle references.
    // last_picture_ptr is forward reference.
    // next_picture_ptr if backward reference.
    if ( s->pict_type != B_TYPE )
    {
        s->last_picture_ptr = s->next_picture_ptr;
        s->next_picture_ptr = s->current_picture_ptr;
    }

    // Part of error recovery
    if ( ( s->pict_type != I_TYPE ) &&
         ( ( s->last_picture_ptr == NULL ) ||
           ( s->last_picture_ptr->buf_allocated == AV_BUF_FREE ) ) )
    {
        fprintf(stderr, "error: first frame is no keyframe\n");
        return -1;

        //these should have been dropped if we dont have a reference
        assert(s->pict_type != B_TYPE);
        goto alloc;
    }

    s->hurry_up= s->avctx->hurry_up;

    return 0;
}



/*

%func

    h264z_frame_end

%desc

    Generic function for encode/decode. The function called
    after each frame has been coded/decoded

*/

void
h264z_frame_end(H264zEncContext *s)
{
    // Only if encoding mode.
    if ( s->encoding == 1 )
    {
        // All external buffers can be release at this point (YUV)
        encoder_release_external_buffer( s->ready_picture_ptr );

        // :ZZZ: I changed this code.
        // Release frame if it is not used as reference
        if ( ( s->ready_picture_ptr->pict_type == B_TYPE ) ||
             ( s->ready_picture_ptr->skip_frame == 1 ) )
        {
            s->avctx->release_buffer(s->avctx, s->ready_picture_ptr);
        }
    }
    else
    {
        // Release external buffer (bitstrem)
        decoder_release_external_buffer( s->ready_picture_ptr );

        // Release non reference frames
        if ( s->ready_picture_ptr->pict_type == B_TYPE )
        {
            s->avctx->release_buffer( s->avctx, s->ready_picture_ptr );
            decoder_release_picture( s->avctx, s->ready_picture_ptr );

        }
        // Release reference frames
        else
        {
            // If we already have two reference frames, we can release the picture
            if ( ( s->buffered_picture_ptr[0] != NULL ) &&
                 ( s->buffered_picture_ptr[1] != NULL ) )
            {
                decoder_release_picture( s->avctx, s->buffered_picture_ptr[1] );
            }

            // Update buffered picture
            s->buffered_picture_ptr[1] = s->buffered_picture_ptr[0];
            s->buffered_picture_ptr[0] = s->ready_picture_ptr;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
// Encoder only
//////////////////////////////////////////////////////////////////////////////

/**

%func

    load_input_picture

%desc

    The function finds a free entry in the frame's database and occupies it
    by the received frame. The function allocates all memory buffers
    needed for the frame's encoding, and insert the frame to the "input frames
    queue".

*/

static int
load_input_picture(H264zEncContext *s, AVFrame *pic_arg)
{
    AVFrame *pic = NULL;
    int i;

    if ( pic_arg != NULL )
    {
        // Look for free picture.
        i = find_unused_picture(s, 0);

        pic = &s->picture[i];

		// Take TF table from AvCtx
		memcpy( &pic->tf_table, &s->avctx->tf_table, sizeof(pic->tf_table) );
        memcpy( &pic->hp_table, &s->avctx->hp_table, sizeof(pic->hp_table) );
        memcpy( &pic->magnet_table, &s->avctx->magnet_table, sizeof(pic->magnet_table) );

        // Reset reference context pointer
        pic->ref_context = NULL;

        //:FIXME: can't we use a memcpy here so when we add things
        //AVFrame they get copied too?
        // Set the frame parameters.
        pic->pict_type           = pic_arg->pict_type;
        pic->scene_cut           = pic_arg->scene_cut;
        pic->last_in_gop         = pic_arg->last_in_gop;
        pic->pts                 = pic_arg->pts;
        pic->frame_width         = pic_arg->frame_width;
        pic->frame_height        = pic_arg->frame_height;
        pic->left                = pic_arg->left;
        pic->top                 = pic_arg->top;
        pic->width               = pic_arg->width;
        pic->height              = pic_arg->height;
        pic->skip_frame          = pic_arg->skip_frame;
        pic->start_of_last_gop   = pic_arg->start_of_last_gop; 
        pic->last_in_encoded_gop = pic_arg->last_in_encoded_gop;
        pic->snapshot            = pic_arg->snapshot;        

        // load chroma frame parameters
        pic->chroma_frame_width  = pic_arg->chroma_frame_width;
        pic->chroma_frame_height = pic_arg->chroma_frame_height;
        pic->chroma_left         = pic_arg->chroma_left;
        pic->chroma_top          = pic_arg->chroma_top;
        pic->chroma_width        = pic_arg->chroma_width;
        pic->chroma_height       = pic_arg->chroma_height;

        pic->use_gmv             = pic_arg->use_gmv;
        pic->gmv_x               = pic_arg->gmv_x;
        pic->gmv_y               = pic_arg->gmv_y;
        pic->gmv_fwd_x           = pic_arg->gmv_fwd_x;
        pic->gmv_fwd_y           = pic_arg->gmv_fwd_y;
        pic->gmv_bwd_x           = pic_arg->gmv_bwd_x;
        pic->gmv_bwd_y           = pic_arg->gmv_bwd_y;
        pic->gmv_confidence      = pic_arg->gmv_confidence;
        pic->gmv_skip_cost_ratio = pic_arg->gmv_skip_cost_ratio;

        pic->sFccData = pic_arg->sFccData;

        pic->dummy_gop           = pic_arg->dummy_gop;

        assert( pic->chroma_width == pic->width );

        pic->close_frames_sad = pic_arg->close_frames_sad;

        // :FIXME: in the future we will receive search area size per
        // each frame.
        pic->me_search_x = pic_arg->me_search_x;
        pic->me_search_y = pic_arg->me_search_y;

        pic->display_picture_number = pic_arg->display_picture_number;
        pic->coded_picture_number   = pic_arg->coded_picture_number;
        pic->frame_in_gop           = pic_arg->frame_in_gop;

        // The following data is supplied by user (not allocated by codec)
        pic->data[AV_BUF_YUV_IMAGE]     = pic_arg->data[AV_BUF_YUV_IMAGE];
        pic->data[AV_BUF_YUV_IMAGE_C]   = pic_arg->data[AV_BUF_YUV_IMAGE_C];

        // From now on, the buffers are at our responsibility.
        pic_arg->data[AV_BUF_YUV_IMAGE] = NULL;
        pic_arg->data[AV_BUF_YUV_IMAGE_C] = NULL;
	}

	if ( pic != NULL )
    {
        pic->replace_by_skip = 0;

    // In case of Skip Frame with 2 video codecs, we are handling the skip frame in
    // a lower level (we encode the frame as normal, and replace the encoded frame by
    // a skip frame 
        if ( s->avctx->num_of_codecs == 2 ) 
        {
    		pic->replace_by_skip = pic->skip_frame;
            pic->skip_frame = 0;
        }
    }

    s->current_picture_ptr = pic;
    s->ready_picture_ptr = pic;

    return 0;
}

/*

    h264z_encode_nal

*/

int
h264z_encode_nal( PutBitContext* pb, int nal_ref_idc, int nal_unit_type )
{
    // Put each NAL start code 4-bytes length to allow fast NAL update for QT's mov file format
    // zero_byte
    put_bits(pb, 8, 0 );

    // start_code_prefix_one_3bytes
    put_bits(pb, 24, 0x000001 );

    // forbidden_zero_bit
    put_bits(pb, 1, 0 );

    // nal_ref_idc
    put_bits(pb, 2, nal_ref_idc);

    // nal_unit_type
    put_bits(pb, 5, nal_unit_type);

    return 0;
}


/*

    h264z_encode_aud ( access unit delimiter)

*/

int
h264z_encode_aud(PutBitContext* pb, int pict_type)
{
    h264z_encode_nal(pb, 0, H264_NAL_AUD);

    // primary_pic_type
    put_bits(pb, 3, pict_type - I_TYPE);

	// rbsp_trailing_bits - byte alignment
    ff_h264_stuffing(pb);

    return 1;
}


/*

    h264z_encode_hrd

*/

int
h264z_encode_hrd(H264zEncContext *s, HRD *hrd, PutBitContext* pb)
{
	hrd->cpb_cnt_minus1 = 0;
    put_ue_bits(pb, hrd->cpb_cnt_minus1);

    hrd->bit_rate_scale = s->brc_ctx.iHRDBrScale;
    hrd->cpb_size_scale = s->brc_ctx.iHRDBsScale;
    hrd->bit_rate_value_minus1[0] = s->brc_ctx.iHRDBr;
    hrd->cpb_size_value_minus1[0] = s->brc_ctx.iHRDBs;

    put_bits(pb, 4, hrd->bit_rate_scale);
    put_bits(pb, 4, hrd->cpb_size_scale);


	put_ue_bits(pb, hrd->bit_rate_value_minus1[0]);
	put_ue_bits(pb, hrd->cpb_size_value_minus1[0]);

	hrd->cbr_flag[0] = s->avctx->cbr_flag;
	put_bits(pb, 1, hrd->cbr_flag[0]);


	hrd->initial_cpb_removal_delay_length_minus1 = 23;
	put_bits(pb, 5, hrd->initial_cpb_removal_delay_length_minus1);

	hrd->cpb_removal_delay_length_minus1 = 15;
	put_bits(pb, 5, hrd->cpb_removal_delay_length_minus1);

	hrd->dpb_output_delay_length_minus1 = 5;
	put_bits(pb, 5, hrd->dpb_output_delay_length_minus1);

	hrd->time_offset_length = 24;// error with start code prevention
	put_bits(pb, 5, hrd->time_offset_length);

	return 1;
}

/*

    h264z_encode_vui

*/

int
h264z_encode_vui(H264zEncContext *s, VUI *vui, PutBitContext* pb)
{
    vui->aspect_ratio_info_present_flag = 1;
    put_bits(pb, 1, vui->aspect_ratio_info_present_flag);

    vui->aspect_ratio_idc = 1;
    put_bits(pb, 8, vui->aspect_ratio_idc);

	vui->overscan_info_present_flag = 0;
	put_bits(pb, 1, vui->overscan_info_present_flag);

    vui->video_signal_type_present_flag = !!(s->avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_VUI_VST);
    put_bits(pb, 1, vui->video_signal_type_present_flag);
    if (vui->video_signal_type_present_flag)
    {
		vui->colour_description_present_flag = (s->avctx->puwColrAtom != NULL);
        put_bits(pb, 3, 5);
        put_bits(pb, 1, 1);
        put_bits(pb, 1, vui->colour_description_present_flag);
        if (vui->colour_description_present_flag)
        {
            put_bits(pb, 8, s->avctx->puwColrAtom[0]);
            put_bits(pb, 8, s->avctx->puwColrAtom[1]);
            put_bits(pb, 8, s->avctx->puwColrAtom[2]);
        }
    }

	vui->chroma_loc_info_present_flag = 0;
    put_bits(pb, 1, vui->chroma_loc_info_present_flag);

	vui->timing_info_present_flag = 1;
    put_bits(pb, 1, vui->timing_info_present_flag);

    if (s->avctx->frame_rate / s->avctx->frame_rate_base > 30)
    {
        vui->num_units_in_tick = 2002; 
    }
    else
    {
        vui->num_units_in_tick = 4004;
    }

    put_bits(pb, 32, vui->num_units_in_tick);

	vui->time_scale = 240000;
    put_bits(pb, 32, vui->time_scale);

	vui->fixed_frame_rate_flag = 1;
    put_bits(pb, 1, vui->fixed_frame_rate_flag);

    if ( s->avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_NAL_HDR )
    {
        vui->nal_hrd_parameters_present_flag = 1;
        put_bits(pb, 1, vui->nal_hrd_parameters_present_flag);

        // HRD Parameters:
        h264z_encode_hrd(s, &vui->nal_hrd_parameters, pb);
    }
    else
    {
        vui->nal_hrd_parameters_present_flag = 0;
        put_bits(pb, 1, vui->nal_hrd_parameters_present_flag);
        // time_offset_length default is 24 (if it is not defined)
        vui->nal_hrd_parameters.time_offset_length = 24;
    }

    if ( s->avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_VCL_HDR )
    {
        vui->vcl_hrd_parameters_present_flag = 1;
        put_bits(pb, 1, vui->vcl_hrd_parameters_present_flag);

            // VCL HRD Parameters:
    	h264z_encode_hrd(s, &vui->vcl_hrd_parameters, pb);
    }
    else
    {
        vui->vcl_hrd_parameters_present_flag = 0;
        put_bits(pb, 1, vui->vcl_hrd_parameters_present_flag);
        // time_offset_length default is 24 (if it is not defined)
        vui->vcl_hrd_parameters.time_offset_length = 24;
    }

    if ( s->avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_NAL_HDR ||
         s->avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_VCL_HDR )
    {
        vui->low_delay_hrd_flag = 0;
        put_bits(pb, 1, vui->low_delay_hrd_flag);
    }

    if ( s->flags & CODEC_FLAG_INTERLACED_MODE )
    {
        vui->pic_struct_present_flag = 1;
    }
    else
    {
        vui->pic_struct_present_flag = 0;
    }
	if ( s->flags & CODEC_FLAG_VUI_PRESENT_FLAG)
    {
        vui->pic_struct_present_flag = 1;
    }
    else
    {
        vui->pic_struct_present_flag = 0;
    }
    put_bits(pb, 1, vui->pic_struct_present_flag);

	vui->bitstream_restriction_flag = 1;
    put_bits(pb, 1, vui->bitstream_restriction_flag);

	vui->motion_vectors_over_pic_boundaries_flag = 1;
    put_bits(pb, 1, vui->motion_vectors_over_pic_boundaries_flag);

	// max_bytes_per_pic_denom
	if ( s->physical_width <= 640 )
	{
		put_ue_bits(pb, 2);
	}
	else
	{
		put_ue_bits(pb, 0);
	}


	// max_bits_per_mb_denom
    put_ue_bits(pb, 0);

	vui->log2_max_mv_length_horizontal = 16;
    put_ue_bits(pb, vui->log2_max_mv_length_horizontal);

	vui->log2_max_mv_length_vertical = 16;
    put_ue_bits(pb, vui->log2_max_mv_length_vertical);

	vui->num_reorder_frames = 1;
    put_ue_bits(pb, vui->num_reorder_frames);

	// max_dec_frame_buffering
    if ( s->physical_width <= 640 )
	{
		put_ue_bits(pb, 3);
	}
	else if (s->physical_width <= 1280)
    {
        put_ue_bits(pb, 6);
    }
    else
	{
		put_ue_bits(pb, 4);
	}

    return 1;
}

/*

    h264z_encode_sps

*/

int
h264z_encode_sps(H264zEncContext *s, PutBitContext* pb)
{
	PutBitContext pbSwitch;

	uint8_t SPS_temp_buf[128];

	int data_size = 0;
	int i;
	int count = 0;
	BOOL bIntraOnly = s->avctx->is_intra_only;

	/***************************************************************************************/
	/* COMMON NOTE: Since we started to write HRD buffer Size and Input Rate to the header */
	/* we can encounter Start Code Prevention problem in  any  place  of  the  HRD  header */
	/* We decided to save SPS in the separate buffer, and then check it for if we need  to */
	/* add start code prevention. SPS is the minimal byte-aligned  block,  which  includes */
	/* hrd_header. It start and ends byte-aligned.                                         */
	/* Dec 22nd, Boris Bolshem                                                             */
	/***************************************************************************************/

	// Save Current PBContext
	pbSwitch = *pb;

	// Init s->pb to the temporal buffer, so that all new code will be wriiten to it
	// so we can do start code prevention later
    init_put_bits(pb, SPS_temp_buf, 128, NULL, NULL);

    h264z_encode_nal(pb, 1, H264_NAL_SPS);

    // profile_idc: Baseline/Main
    put_bits(pb, 8, s->sps.profile_idc);

    // constraint_set0_flag
    put_bits(pb, 1, 0);

    // constraint_set1_flag: 1 for Main
	if ( s->physical_width <= 640 )
	{
		put_bits(pb, 1, 1);
	}
	else
	{
		put_bits(pb, 1, 0);
	}

    // constraint_set2_flag
    put_bits(pb, 1, 0);

    // constraint_set3_flag
    put_bits(pb, 1, 0);

    // reserved_zero_4bits
    put_bits(pb, 4, 0);

    // level_idc
    put_bits(pb, 8, s->sps.level_idc );

	// seq_parameter_set_id
    put_ue_bits(pb, 0);

	// log2_max_frame_num_minus4
    put_ue_bits(pb, s->sps.log2_max_frame_num - 4);

	// pic_order_cnt_type: 0 or 1 - currently only 0 is implemented
	put_ue_bits(pb, 0);

    // log2_max_pic_order_cnt_lsb_minus4:
    put_ue_bits(pb, s->sps.log2_max_poc_lsb - 4);

	// num_ref_frames: 0, 1, or 2
    if (bIntraOnly)
    {
        put_ue_bits(pb, 0);
    }
    else if ( s->max_b_frames > 0 )
    {
        put_ue_bits(pb, 2);
    }
    else
    {
        put_ue_bits(pb, 1);
    }

	// gaps_in_frame_num_value_allowed_flag
    put_bits(pb, 1, s->gaps_in_frame_num_value_allowed_flag);

	// pic_width_in_mbs_minus1
    put_ue_bits(pb, s->physical_width / MB_WIDTH_IN_PIXELS - 1);

	// pic_height_in_map_units_minus1
    put_ue_bits(pb, s->physical_height / MB_WIDTH_IN_PIXELS - 1);

    // frame_mbs_only_flag: 1
    put_bits(pb, 1, 1);

	// direct_8x8_inference_flag: 1
    put_bits(pb, 1, 1);

    // if needed (width or height not multiple of macroblock),
    // add cropping of video frame
    if ( s->physical_height != s->logical_height ||
         s->physical_width != s->logical_width )
    {
        // cropping needed
        put_bits(pb, 1, 1);

        if ( s->avctx->main_rotation_angle == 0 && s->avctx->display_rotation_angle == 0 )
        {
            // left crop
        put_ue_bits(pb, 0);
            // right crop (in units of SubWidthC=2, since we encode 4:2:0)
        put_ue_bits(pb, (s->physical_width - s->logical_width)/2);
            // top crop
        put_ue_bits(pb, 0);
            // bottom crop (in units of SubHeightC=2, since we encode 4:2:0)
        put_ue_bits(pb, (s->physical_height - s->logical_height)/2);
        }
        else if ( ( s->avctx->main_rotation_angle == 0 && s->avctx->display_rotation_angle == 90 ) ||
                  ( s->avctx->main_rotation_angle == 270 && s->avctx->display_rotation_angle == 0 ) ||
                  ( s->avctx->main_rotation_angle == 0 && s->avctx->display_rotation_angle == 180 ) ||
                  ( s->avctx->main_rotation_angle == 180 && s->avctx->display_rotation_angle == 0 ) )
        {
            // cropping should be top instead of bottom

            // left crop
            put_ue_bits(pb, 0);
            // right crop (in units of SubWidthC=2, since we encode 4:2:0)
            put_ue_bits(pb, (s->physical_width - s->logical_width)/2);
            // top crop (in units of SubHeightC=2, since we encode 4:2:0)
            put_ue_bits(pb, (s->physical_height - s->logical_height)/2);
            // bottom crop
            put_ue_bits(pb, 0);
        }
        else if ( ( s->avctx->main_rotation_angle == 0 && s->avctx->display_rotation_angle == 270 ) ||
                  ( s->avctx->main_rotation_angle == 90 && s->avctx->display_rotation_angle == 0 ) )
        {
            // cropping should be left instead of right

            // left crop (in units of SubWidthC=2, since we encode 4:2:0)
            put_ue_bits(pb, (s->physical_width - s->logical_width)/2);
            // right crop
            put_ue_bits(pb, 0);
            // top crop
            put_ue_bits(pb, 0);
            // bottom crop (in units of SubHeightC=2, since we encode 4:2:0)
            put_ue_bits(pb, (s->physical_height - s->logical_height)/2);
        }
    }
    else
    {
        // frame_cropping_flag: 0
        put_bits(pb, 1, 0);
    }

    if(s->avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_VUI)
    {
	    // s->sps.vui_parameters_present_flag: 1
	    put_bits(pb, 1, 1);

	    h264z_encode_vui(s, &s->sps.vui_parameters, pb);
    }
    else
    {
	    // s->sps.vui_parameters_present_flag: 0
	    put_bits(pb, 1, 0);
    }

	// rbsp_trailing_bits - byte alignment
    ff_h264_stuffing(pb);

	data_size = (put_bits_count(pb) + 7)>>3;

    flush_put_bits(pb);

	// Restore original buffer
	*pb = pbSwitch ;

	// put start code to the original buffer
	put_bits(pb, 8, SPS_temp_buf[0]);
	put_bits(pb, 8, SPS_temp_buf[1]);
	put_bits(pb, 8, SPS_temp_buf[2]);
	put_bits(pb, 8, SPS_temp_buf[3]);

	// Start Code Prevention code:
	for(i = 4; i < data_size; i++)
	{
		if(count == 2 && !(SPS_temp_buf[i] & 0xFC)) // 2 indicates the number of zero bytes in the short start-code prefix
		{
			put_bits(pb, 8, 0x03);
			count = 0;
		}

		put_bits(pb, 8, SPS_temp_buf[i]);
		if(SPS_temp_buf[i] == 0x00)
			count++;
		else
			count = 0;
	}

    return 1;
}

/*

    h264z_encode_pps

*/

int
h264z_encode_pps(H264zEncContext *s, PutBitContext* pb)
{
    h264z_encode_nal(pb, 1, H264_NAL_PPS);

	// pic_parameter_set_id
    put_ue_bits(pb, 0);

	// seq_parameter_set_id
	put_ue_bits(pb, 0);

	// entropy_coding_mode_flag: 0 (CAVLC) or 1 (CABAC)
    put_bits(pb, 1, s->entropy_mode);

    if ( s->flags & CODEC_FLAG_INTERLACED_MODE )
    {
        // pic_order_present_flag = 1
        put_bits(pb, 1, 1);
    }
    else
    {
        // pic_order_present_flag = 0
    put_bits(pb, 1, 0);
    }

	// num_slice_groups_minus1: 0 - using a single group per picture
	put_ue_bits(pb, 0);

	// num_ref_idx_l0_active_minus1
	put_ue_bits(pb, 0);

	// num_ref_idx_l1_active_minus1
	put_ue_bits(pb, 0);

	// weighted_pred_flag: 0 - using default
  	put_bits(pb, 1, 0);

	// weighted_bipred_idc: 0 - using default
  	put_bits(pb, 2, 0);

	// pic_init_qp_minus26: relative to 26
   	put_se_bits(pb, s->pic_init_qp - 26);

	// pic_init_qs_minus26: not used, set to 0
   	put_se_bits(pb, 0);

	// chroma_qp_index_offset
  	put_se_bits(pb, 0);

	// deblocking_filter_control_present_flag
    //put_bits(&s->pb, 1, 1);
	if (s->deblocking_filter == 1)
	{
		put_bits(pb, 1, 0);
	}
	else
	{
		put_bits(pb, 1, 1);
	}

	// constrained_intra_pred_flag: 0
    put_bits(pb, 1, 0);

	// redundant_pic_cnt_present_flag: 0
    put_bits(pb, 1, 0);

	// rbsp_trailing_bits: byte alignment
    ff_h264_stuffing(pb);

    return 1;
}


/*

    encode Buffering Period SEI message

*/

int
h264z_encode_sei_buff_period(H264zEncContext *s, PutBitContext* pb, AVFrame *pic)
{
	int PayloadSize = 8;

	int initial_cpb_removal_delay, initial_cpb_removal_delay_offset, fullness, bitrate, size, buf_level, maxdelay;


    size = s->brc_ctx.iHRDBufferSize;

    // Calculate hrd/cpb state   
    if ( pic->coded_picture_number == 0 )
    {
        fullness = 0;
    }
    else
    {
        fullness = s->brc_ctx.iHRDBufferFullness > 0 ? s->brc_ctx.iHRDBufferFullness : 0;
    }

    if ( s->brc_ctx.iHRDInputRate > 0 )
    {
        bitrate = s->brc_ctx.iHRDInputRate;
    }
    else
    {
        bitrate = s->brc_ctx.iBitRate;
    }


	buf_level = size - fullness; // + s->brc_ctx.iHRDInputRateInBitsPerFrame;

	buf_level = MIN (size, buf_level );

    initial_cpb_removal_delay = (ULONG)(((UINT64)90000 * buf_level ) / bitrate);

    //initial_cpb_removal_delay = MIN( initial_cpb_removal_delay, 90000 );

     maxdelay = (ULONG)((UINT64)90000 * size / bitrate ); // can be calculated once per clip

    initial_cpb_removal_delay_offset = maxdelay - initial_cpb_removal_delay;


	// calculate payload size:
	if(s->sps.vui_parameters.nal_hrd_parameters_present_flag){
		PayloadSize += (s->sps.vui_parameters.nal_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1) * 2;
	}

	if(s->sps.vui_parameters.vcl_hrd_parameters_present_flag){
		PayloadSize += (s->sps.vui_parameters.vcl_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1) * 2;
	}

	PayloadSize = (PayloadSize + 7) >> 3;

	//last_payload_size_byte
	put_bits(pb, 8, PayloadSize);

	//seq_parameter_set_id
	put_ue_bits(pb, 0);

	if(s->sps.vui_parameters.nal_hrd_parameters_present_flag){
		//initial_cpb_removal_delay
		put_bits(pb,
				s->sps.vui_parameters.nal_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1,
				initial_cpb_removal_delay);

		//initial_cpb_removal_delay_offset
		put_bits(pb,
				s->sps.vui_parameters.nal_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1,
				initial_cpb_removal_delay_offset);
	}

	if(s->sps.vui_parameters.vcl_hrd_parameters_present_flag){
		//initial_cpb_removal_delay
		put_bits(pb,
				s->sps.vui_parameters.vcl_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1,
				initial_cpb_removal_delay);

		//initial_cpb_removal_delay_offset
		put_bits(pb,
				s->sps.vui_parameters.vcl_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1,
				initial_cpb_removal_delay_offset);
	}

    return 1;
}


/*
    encode Picture Timing SEI message
*/

int
h264z_encode_sei_picture_timing(H264zEncContext *s, PutBitContext* pb, AVFrame* pic, BOOL bFirstTime )
{
	int PayloadSize = 0;

	int GOP_size = h264z_encode_get_gop_size(s);

	// calculate payload size:
	if(s->sps.vui_parameters.nal_hrd_parameters_present_flag || s->sps.vui_parameters.vcl_hrd_parameters_present_flag){
		PayloadSize +=  s->sps.vui_parameters.nal_hrd_parameters.cpb_removal_delay_length_minus1 +
						s->sps.vui_parameters.nal_hrd_parameters.dpb_output_delay_length_minus1 + 2;
	}

	if(s->sps.vui_parameters.pic_struct_present_flag){
        // time_offset_length is always defined, since it has default of 24 (not zero).
        // We set it also when HRD is not activated.
        PayloadSize += 25 + s->sps.vui_parameters.nal_hrd_parameters.time_offset_length;
	}

#ifdef H264Z_DEBUG_PRINT
	if(PayloadSize == 0)
	{
		printf("Error: H264 encoder Picture timing message trying to encode but nothing to do, check VUI!\n");
	}
#endif

	PayloadSize = (PayloadSize + 7) >> 3;


	//last_payload_size_byte
	put_bits(pb, 8, PayloadSize);

	//seq_parameter_set_id
	if(s->sps.vui_parameters.nal_hrd_parameters_present_flag || s->sps.vui_parameters.vcl_hrd_parameters_present_flag){
		int dpb_output_delay = 0;
		int cpb_removal_delay;

        if ( bFirstTime == TRUE )
        {
            cpb_removal_delay = (s->real_frame_num % GOP_size) * 2;

			if(s->real_frame_num == 0)
		{
			cpb_removal_delay = 0;
		}
			else if( s->real_frame_num % GOP_size == 0 &&
				 pic->frame_in_gop % GOP_size == 0)
		{
			//assert(s->idr_frame == 1);
			cpb_removal_delay = GOP_size * 2;
		}
            pic->cpb_removal_delay = cpb_removal_delay;
        }
        else
        {
     		// on the second time, rewrite the value we already calculated at the first time
            cpb_removal_delay = pic->cpb_removal_delay;
        }

		if(pic->pict_type == I_TYPE){
			dpb_output_delay = 2;
		}else if(pic->pict_type == B_TYPE){
			dpb_output_delay = 0;
		}else{
			// P:
			dpb_output_delay = 2 + 2*s->max_b_frames;
		}

		put_bits(pb, s->sps.vui_parameters.nal_hrd_parameters.cpb_removal_delay_length_minus1 + 1, cpb_removal_delay);

		put_bits(pb, s->sps.vui_parameters.nal_hrd_parameters.dpb_output_delay_length_minus1 + 1, dpb_output_delay);

	}


	if(s->sps.vui_parameters.pic_struct_present_flag){
		put_bits(pb, 4, 0); //pic_struct

		//NumClockTS = 1;
		put_bits(pb, 1, 1); // clock_timestamp_flag
        if ( s->flags & CODEC_FLAG_INTERLACED_MODE )
        {
            put_bits(pb, 2, 1); // ct_type - interlaced
        }
        else
        {
            put_bits(pb, 2, 0); // ct_type - progressive
        }
        
		put_bits(pb, 1, 1); // nuit_field_based_flag
		put_bits(pb, 5, 3); // counting_type
		put_bits(pb, 1, 0); // full_timestamp_flag
		put_bits(pb, 1, 1); // discontinuity_flag
		put_bits(pb, 1, 0); // cnt_dropped_flag
		put_bits(pb, 8, 0); // n_frames

		put_bits(pb, 1, 0); // seconds_flag

        // time_offset_length is initialized anyway, since its default is 24 (not zero)
        if(s->sps.vui_parameters.nal_hrd_parameters.time_offset_length > 0)
        {
            put_bits(pb, s->sps.vui_parameters.nal_hrd_parameters.time_offset_length, 0); // time_offset
        }
	}

    return 1;
}


/*
    encode recovery point SEI message
*/

int
h264z_encode_sei_recovery_point(H264zEncContext *s, PutBitContext* pb)
{
	int PayloadSize = 6;

	// calculate payload size:
	PayloadSize = (PayloadSize + 7) >> 3;

	//last_payload_size_byte
	put_bits(pb, 8, PayloadSize);

	put_ue_bits(pb, 0); // recovery_frame_cnt

	put_bits(pb, 1, 1); // exact_match_flag
	put_bits(pb, 1, 0); // broken_link_flag
	put_bits(pb, 2, 0); // changing_slice_group_idc

    return 1;
}


/*

  encore required SEI message, apply start code prevention code for it
  and put it into general PutBit context

*/

int
h264z_encode_sei_message(H264zEncContext *s, int PayloadType, AVFrame* pic, BOOL bFirstTime)
{
	//
	uint8_t SEI_temp_buf[64];

	int data_size = 0;
	int i;
	int count = 0;
    int prevention_count = 0;

	PutBitContext pb;

	init_put_bits(&pb, SEI_temp_buf, 64, NULL, NULL);

    // zero_byte
    put_bits(&pb, 8, 0 );

    // start_code_prefix_one_3bytes
    put_bits(&pb, 24, 0x000001 );

    // forbidden_zero_bit
    put_bits(&pb, 1, 0 );

    // nal_ref_idc
    put_bits(&pb, 2, 0);

    // nal_unit_type
    put_bits(&pb, 5, H264_NAL_SEI);

	// last_payload_type_byte
	put_bits(&pb, 8, PayloadType);

	// call apropriate SEI Message encoder:
	switch(PayloadType)
	{
		case H264_SEI_BUFFERING_PERIOD:
			h264z_encode_sei_buff_period(s, &pb, pic);
			break;
		case H264_SEI_PICTURE_TIMING:
			h264z_encode_sei_picture_timing(s, &pb, pic, bFirstTime);
			break;
		case H264_SEI_RECOVERY_POINT:
			h264z_encode_sei_recovery_point(s, &pb);
			break;
		default:
			printf("Error: Uknnown SEI Payload type (%d) encoding request\n", PayloadType);
			return -1;

	}

	// rbsp_trailing_bits: byte alignment
    ff_h264_stuffing(&pb);

	//finalize SEI NAL
	put_bits(&pb, 8, 0x80);

	// Necessary for correct playback when the container format is .mov
	// For this file format all the sizes must be 2 bytes aligned.
	// Skip it for MP4
	if ( (!s->flags) & CODEC_FLAG_MP4_SUPPORT)
    {
        ffff_h264_stuffing(&pb);
    }

	flush_put_bits(&pb);

	data_size = (put_bits_count(&pb) + 7)>>3;


	// put start code
	put_bits(&pic->pb, 8, SEI_temp_buf[0]);
	put_bits(&pic->pb, 8, SEI_temp_buf[1]);
	put_bits(&pic->pb, 8, SEI_temp_buf[2]);
	put_bits(&pic->pb, 8, SEI_temp_buf[3]);

	// Start Code Prevention code:

	for(i = 4; i < data_size; i++)
	{
		if(count == 2 && !(SEI_temp_buf[i] & 0xFC)) // 2 indicates the number of zero bytes in the short start-code prefix
		{
            prevention_count++;
			put_bits(&pic->pb, 8, 0x03);
			count = 0;
		}

		put_bits(&pic->pb, 8, SEI_temp_buf[i]);
		if(SEI_temp_buf[i] == 0x00)
			count++;
		else
			count = 0;
	}

	for (i = 0; i< (0 - data_size); i+=3 )
	{
		put_bits(&pic->pb, 8, 0x0); // CABAC zero word
		put_bits(&pic->pb, 8, 0x0);
		put_bits(&pic->pb, 8, 0x03);
	}

    if ( PayloadType == H264_SEI_BUFFERING_PERIOD )
    {
        if ( bFirstTime == TRUE )
        {
            // Adding 5 "extra" 0x0 bytes here (allowed to add 0x0 bytes before start code)
            // In the second time, we can override some of them with needed 0x3 of start code prevention
            // if needed by values of the seond time we write this message
            pic->prevention_count = prevention_count;
            count = 5;
        }
        else
        {
            count = 5 - ( prevention_count - pic->prevention_count );
        }

        for ( i = 0; i < count; i++ )
        {
            put_bits(&pic->pb, 8, 0x0);
        }
    }

	return 0;
}


/*

    h264z_get_slice_type

*/

int
h264z_get_slice_type(int pict_type)
{
    int slice_type = H264_SLICE_TYPE_NOT_DEFINED;

    switch ( pict_type )
    {
        case I_TYPE:
        {
            slice_type = H264_SLICE_TYPE_I;
            break;
        }
        case P_TYPE:
        {
            slice_type = H264_SLICE_TYPE_P;
            break;
        }
        case B_TYPE:
        {
            slice_type = H264_SLICE_TYPE_B;
            break;
        }
        default:
        {
            printf ("h264z_get_slice_type: wrong picture type %d\n", pict_type );
            break;
        }
    }

    return slice_type;
}

/**

%func

    h264z_encode_slice_header

%decs*
*
* BOOL bFirstTime - wheather this is the first time we encode
* slice header for this frame

*/

int
h264z_encode_slice_header(H264zEncContext *s, AVFrame* pic, int slice_type, BOOL bFirstTime )
{
    int pic_order_cnt_lsb;
    int nal_unit_type;

	int GOP_size = h264z_encode_get_gop_size(s);
    BOOL bIntraOnly = s->avctx->is_intra_only;

	// Count
	int number_of_ref_frames_in_gop = ( ( GOP_size - 1 + s->max_b_frames ) / ( s->max_b_frames + 1 ) ) + 1;

    if ( pic->idr_frame == 1 )
    {
        nal_unit_type = H264_NAL_IDR_SLICE;
    }
    else
    {
        nal_unit_type = H264_NAL_SLICE;
    }

    h264z_encode_nal(&pic->pb, pic->nal_ref_idc, nal_unit_type);

    // slice offset in bytes, in DWORD alignment.
	// We need to send to CAVLC/CABAB HW a buffer right after slice NAL unit, for start code
	// prevention emulaiton handling. 
    s->slice_header_offset = (INT)pbBufPtr(&pic->pb);

	// first_mb_in_slice
    put_ue_bits(&pic->pb, 0);

    // slice_type
    put_ue_bits(&pic->pb, slice_type);

	// pic_parameter_set_id
    put_ue_bits(&pic->pb, 0);

    // Calculate current frame_num
    if (pic->idr_frame == 1)
    {
        // For GOP I frames frame_num = 0;
        if ( bFirstTime == TRUE )
        {
            s->frame_num = 0;
            s->real_frame_num = 0;
            pic->frame_num = 0;
        }              
        put_bits(&pic->pb, s->sps.log2_max_frame_num, pic->frame_num );
    }
    else
    {
        put_bits(&pic->pb, s->sps.log2_max_frame_num, ( (pic->frame_num + s->brc_ctx.uiSkipedGopIFrames * number_of_ref_frames_in_gop ) % s->max_frame_num ) );
    }

    // Once the frame number has been saved s->frame_num contains the number
    // of the next frame in the sequence
    if ( bFirstTime == TRUE && pic->pict_type != B_TYPE )
    {
        if (s->gaps_in_frame_num_value_allowed_flag == 0)
        {
            s->frame_num = ( s->frame_num + 1 );
        }
        else // Gaps in frame_num are allowed
        {
            s->frame_num = ( s->frame_num + s->gaps_in_frame_num_value_allowed_flag );
        }

        //s->frame_num =  s->frame_num % s->max_frame_num;
    }

    if ( bFirstTime == TRUE )
    {
        if (s->gaps_in_frame_num_value_allowed_flag == 0)
        {
            s->real_frame_num++;
        }
        else // Gaps in frame_num are allowed
        {
            s->real_frame_num += s->gaps_in_frame_num_value_allowed_flag;
        }

        s->real_frame_num %= (GOP_size + 1);
    } 



    // Calculate the picture order inside the GOP
    // In case we had skipped GOP-I frames, gop start is actually different (smaller by GOP size * number of skipped I frames)

    if ( s->flags & CODEC_FLAG_INTERLACED_MODE )
    {
        // In interlaced mode, pic_order_cnt_lsb increments by 2 every frame
        pic_order_cnt_lsb = 2 * (pic->display_picture_number - pic->gop_start + s->brc_ctx.uiSkipedGopIFrames * GOP_size ) % s->max_pic_order_cnt_lsb;
    }
    else
    {
        pic_order_cnt_lsb = (pic->display_picture_number - pic->gop_start + s->brc_ctx.uiSkipedGopIFrames * GOP_size ) % s->max_pic_order_cnt_lsb;
    }

	if( nal_unit_type == H264_NAL_IDR_SLICE ) // An I slice in I picture
    {
        // idr_pic_id
        put_ue_bits(&pic->pb, bIntraOnly ? (pic->display_picture_number&1) : 0);
    }

	// pic_order_cnt_lsb
    put_bits(&pic->pb, s->sps.log2_max_poc_lsb, pic_order_cnt_lsb);


    if ( s->flags & CODEC_FLAG_INTERLACED_MODE )
    {
        // delta_pic_order_cnt_bottom = 1
        put_se_bits(&pic->pb, 1 );
    }

	if( slice_type == H264_SLICE_TYPE_B )
    {
		// direct_spatial_mv_pred_flag
        put_bits(&pic->pb, 1, 1);
    }

	if( slice_type == H264_SLICE_TYPE_P || slice_type == H264_SLICE_TYPE_B )
    {
		// num_ref_idx_active_override_flag
        put_bits(&pic->pb, 1, 0);
    }

    // ref_pic_list_reordering

	if( slice_type != H264_SLICE_TYPE_I )
    {
		// ref_pic_list_reordering_flag_l0
        put_bits(&pic->pb, 1, 0);
    }
	if( slice_type == H264_SLICE_TYPE_B )
    {
        // ref_pic_list_reordering_flag_l1
        put_bits(&pic->pb, 1, 0);
    }

	if( pic->nal_ref_idc != 0 ) // if it's a reference frame
    {
		// dec_ref_pic_marking
    	if( nal_unit_type == H264_NAL_IDR_SLICE ) // an I slice in an I picture
    	{
		    // no_output_of_prior_pics_flag
            put_bits(&pic->pb, 1, 0);

		    // long_term_reference_flag
            put_bits(&pic->pb, 1, 0);
    	}
    	else
        {
    		// adaptive_ref_pic_marking_mode_flag
            put_bits(&pic->pb, 1, 0);
        }
    }

	if( s->entropy_mode == H264_ENTROPY_MODE_CABAC && slice_type != H264_SLICE_TYPE_I )
    {
		// cabac_init_idc  - TBD
        put_ue_bits(&pic->pb, 0);
    }

	// slice_qp_delta
    put_se_bits(&pic->pb, s->slice_qp_delta);

	if( s->deblocking_filter == 0 )
	{
		// disable_deblocking_filter_idc
        put_ue_bits(&pic->pb, 1);
	}
//	else
//	{
//		// disable_deblocking_filter_idc
//        put_ue_bits(&s->pb, 0);
//
//	    // slice_alpha_c0_offset
//        put_se_bits(&s->pb, (s->slice_alpha_c0_offset >> 1));
//
//	    // slice_beta_offset
//        put_se_bits(&s->pb, (s->slice_beta_offset >> 1));
//	}

    return 0;
}


void
h264z_encode_custom_nals(H264zEncContext *s)
{
    // TODO: Insert writing SEI messages and AUD here
}

int
h264z_encode_slice_header_2vc(H264zEncContext *s, AVFrame* pic, int slice_type, BOOL bFirstTime )
{
	h264z_encode_slice_header(s, pic, slice_type, bFirstTime );

	// CABAC alignment
	if ( s->entropy_mode == H264_ENTROPY_MODE_CABAC )
	{
		cabac_alignment(&pic->pb);
	}

	return 0;

}

/**

%func

    h264_encode_skip_frame

%decs

*/

/////////////////////////////////////////////////////////
// P_skip frame code
/////////////////////////////////////////////////////////

BYTE bySkipFramePCode160x128[] =
{
    0xE4, 0x40, 0x0C, 0xE8
};

BYTE bySkipFramePCode176x144[] =
{
    0xE4, 0x40, 0x07, 0xCC
};

BYTE bySkipFramePCode320x112[] =
{
    0xE4, 0x40, 0x02, 0xCA
};

BYTE bySkipFramePCode320x176[] =
{
    0xE4, 0x40, 0x00, 0x5B, 0x40
};

BYTE bySkipFramePCode320x180[] =
{
    0xE4, 0x40, 0x00, 0x36, 0x60
};

BYTE bySkipFramePCode320x240[] =
{
    0xE4, 0x40, 0x00, 0x0B, 0xB8
};

BYTE bySkipFramePCode352x288[] =
{
    0xE4, 0x40, 0x00, 0x00, 0xF9, 0x80
};

BYTE bySkipFramePCode640x360[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x18, 0x10
};
BYTE bySkipFramePCode640x480[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x04,
    0x9C
};

BYTE bySkipFramePCode720x480[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x19, 0x50
};

BYTE bySkipFramePCode720x576[] =
{
    0xe4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x49, 0xc0
};

BYTE bySkipFramePCode800x448[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x06, 0xF4
};

BYTE bySkipFramePCode800x592[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x04, 0x5C
};

BYTE bySkipFramePCode800x608[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x01, 0x37
};

BYTE bySkipFramePCode848x480[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x0D, 0x98
};

BYTE bySkipFramePCode960x544[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x08, 0x78
};

BYTE bySkipFramePCode864x480[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 
    0x03, 0x00, 0x00, 0x03, 0x00, 0x06, 0x54
};

BYTE bySkipFramePCode960x720[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 
    0x00, 0x06, 0x54
};

BYTE bySkipFramePCode1024x576[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x02, 0x82
};

BYTE bySkipFramePCode1280x720[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x02, 0x82
};

BYTE bySkipFramePCode1440x1080[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x0F, 0x98
};


BYTE bySkipFramePCode1920x1080[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x01, 0x5B
};

BYTE bySkipFramePCode816x1080[] =
{
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x49,
    0xC0, 0x00, 0x00, 0x00, 0x01, 0x01, 0x9E, 0x43, 0x1A, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x18, 0x10,
};

/////////////////////////////////////////////////////////
// B_skip frame code
/////////////////////////////////////////////////////////

BYTE bySkipFrameBCode160x128[] =
{
    0x04, 0x1C
};

BYTE bySkipFrameBCode176x144[] =
{
    0x02, 0x92
};

BYTE bySkipFrameBCode320x176[] =
{
    0x00, 0x1D, 0x30
};

BYTE bySkipFrameBCode720x480[] =
{
    0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00,
    0x07, 0xF8
};


BYTE bySkipFrameBCode720x576[] =
{
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x18,
    0x10
};

BYTE bySkipFrameBCode800x448[] =
{
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x02, 0x3E
};

BYTE bySkipFrameBCode800x608[] =
{
    0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03,
    0x00, 0x65, 0x40
};

BYTE bySkipFrameBCode960x544[] =
{
    0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03,
    0x00, 0x02, 0xCA
};

BYTE bySkipFrameBCode320x112[] =
{
   0x00, 0xE3, 0x80
};

BYTE bySkipFrameBCode320x180[] =
{
   0x00, 0x11, 0x70
};

BYTE bySkipFrameBCode320x240[] =
{
   0x00, 0x03, 0xBA
};

BYTE bySkipFrameBCode352x288[] =
{
   0x00, 0x00, 0x52, 0x40
};

BYTE bySkipFrameBCode640x360[] =
{
	0x00, 0x00, 0x03, 0x00, 0x00, 0x07, 0xA4
};


BYTE bySkipFrameBCode640x480[] =
{
	0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x01, 0x81
};



BYTE bySkipFrameBCode800x592[] =
{
    0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x01, 0x6D
};

BYTE bySkipFrameBCode848x480[] = 
{
    0x00, 0x00, 0x03,
    0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 
    0x00, 0x04, 0x5C
};

BYTE bySkipFrameBCode864x480[] =
{
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x01, 0xFE
};

BYTE bySkipFrameBCode960x720[] =
{
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 
    0x03, 0x00, 0x00, 0x03, 0x01, 0xFE
};

BYTE bySkipFrameBCode1024x576[] =
{
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 
    0x03, 0x00, 0xCF, 0x80
};

BYTE bySkipFrameBCode1280x720[] =
{
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0xCF, 0x80
};

BYTE bySkipFrameBCode1440x1080[] =
{
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x05, 0x24
};

BYTE bySkipFrameBCode1920x1080[] =
{
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x6F, 0x40
};

BYTE bySkipFrameBCode816x1080[] =
{
    0x00, 0x00, 0x30, 0x00, 0x00, 0x30, 0x00, 0x00, 0x30, 0x00, 0x00, 0x30, 0x00,
    0x00, 0x30, 0x00, 0x00, 0x30, 0x00, 0x00, 0x30, 0x00, 0x00, 0x18, 0x10
};


void
h264_encode_skip_frame(H264zEncContext *s, AVFrame* picture)
{
    INT i;
    PBYTE pbySkipFrameCode = NULL;

    INT iCodeSize = 0;
    ULONG uiImageSize = MAKE_LONG(s->physical_width, s->physical_height);

    if ( s->entropy_mode == H264_ENTROPY_MODE_CABAC )
    {
    	if ( picture->pict_type == P_TYPE)
    	{
    		switch ( uiImageSize )
    		{
                case MAKE_LONG(160, 128):
                case MAKE_LONG(128, 160):
    			{
    				pbySkipFrameCode = bySkipFramePCode160x128;
    				iCodeSize = sizeof(bySkipFramePCode160x128);
    				break;
    			}

                case MAKE_LONG(176, 144):
                case MAKE_LONG(144, 176):
                {
                    pbySkipFrameCode = bySkipFramePCode176x144;
                    iCodeSize = sizeof(bySkipFramePCode176x144);
                    break;
                }

                case MAKE_LONG(320, 112):
                case MAKE_LONG(112, 320):
                {
                    pbySkipFrameCode = bySkipFramePCode320x112;
                    iCodeSize = sizeof(bySkipFramePCode320x112);
                    break;
                }

                case MAKE_LONG(320, 176):
                case MAKE_LONG(176, 320):
                {
                    pbySkipFrameCode = bySkipFramePCode320x176;
                    iCodeSize = sizeof(bySkipFramePCode320x176);
                    break;
                }

                case MAKE_LONG(320, 192):
                case MAKE_LONG(192, 320):
    			{
    				pbySkipFrameCode = bySkipFramePCode320x180;
    				iCodeSize = 5;
    				break;
    			}

                case MAKE_LONG(320, 240):
                case MAKE_LONG(240, 320):
    			{
    				pbySkipFrameCode = bySkipFramePCode320x240;
    				iCodeSize = sizeof(bySkipFramePCode320x240);
    				break;
    			}

                case MAKE_LONG(352, 288):
                case MAKE_LONG(288, 352):
                {
                    pbySkipFrameCode = bySkipFramePCode352x288;
                    iCodeSize = sizeof(bySkipFramePCode352x288);
                    break;
                }

                case MAKE_LONG(640, 368):
                case MAKE_LONG(368, 640):
                {
                    pbySkipFrameCode = bySkipFramePCode640x360;
                    iCodeSize = sizeof(bySkipFramePCode640x360);
                    break;
                }

    			case MAKE_LONG(640, 480):
    			case MAKE_LONG(480, 640):
    			{
    				pbySkipFrameCode = bySkipFramePCode640x480;
    				iCodeSize = 11;
    				break;
    			}

    			case MAKE_LONG(720, 480):
    			case MAKE_LONG(480, 720):
    			{
    				pbySkipFrameCode = bySkipFramePCode720x480;
    				iCodeSize = sizeof(bySkipFramePCode720x480);
    				break;
    			}

                case MAKE_LONG(720, 576):
                case MAKE_LONG(576, 720):
    			{
    				pbySkipFrameCode = bySkipFramePCode720x576;
    				iCodeSize = sizeof(bySkipFramePCode720x576);
    				break;
    			}

                case MAKE_LONG(800, 448):
                case MAKE_LONG(448, 800):
                {
                    pbySkipFrameCode = bySkipFramePCode800x448;
                    iCodeSize = sizeof(bySkipFramePCode800x448);
                    break;
                }

    			case MAKE_LONG(800, 592):
    			case MAKE_LONG(592, 800):
    			{
    				pbySkipFrameCode = bySkipFramePCode800x592;
    				iCodeSize = sizeof(bySkipFramePCode800x592);
    				break;
    			}

    			case MAKE_LONG(800, 608):
    			case MAKE_LONG(608, 800):
    			{
    				pbySkipFrameCode = bySkipFramePCode800x608;
    				iCodeSize = sizeof(bySkipFramePCode800x608);
    				break;
    			}
                case MAKE_LONG(848, 480):
                case MAKE_LONG(480, 848):
                {
                    pbySkipFrameCode = bySkipFramePCode848x480;
                    iCodeSize = sizeof(bySkipFramePCode848x480);                        
                    break;
                }
                case MAKE_LONG(960, 544):
                {
                    pbySkipFrameCode = bySkipFramePCode960x544;
                    iCodeSize = sizeof(bySkipFramePCode960x544);
                    break;
                }

                case MAKE_LONG(864, 480):
                case MAKE_LONG(480, 864):
                {
                    pbySkipFrameCode = bySkipFramePCode864x480;
                    iCodeSize = sizeof(bySkipFramePCode864x480);
                    break;
                }

                case MAKE_LONG(960, 720):
                case MAKE_LONG(720, 960):
                {
                    pbySkipFrameCode = bySkipFramePCode960x720;
                    iCodeSize = sizeof(bySkipFramePCode960x720);
                    break;
                }

                case MAKE_LONG(1024, 576):
                case MAKE_LONG(576, 1024):
                {
                    pbySkipFrameCode = bySkipFramePCode1024x576;
                    iCodeSize = sizeof(bySkipFramePCode1024x576);
                    break;
                }

    			case MAKE_LONG(1280, 720):
    			case MAKE_LONG(720, 1280):
    			{
    				pbySkipFrameCode = bySkipFramePCode1280x720;
    				iCodeSize = sizeof(bySkipFramePCode1280x720);
    				break;
    			}

                case MAKE_LONG(1440, 1088):
                case MAKE_LONG(1088, 1440):
                {
                    pbySkipFrameCode = bySkipFramePCode1440x1080;
                    iCodeSize = sizeof(bySkipFramePCode1440x1080);
                    break;
                }

                case MAKE_LONG(1920, 1088):
                case MAKE_LONG(1088, 1920):
                {
                    pbySkipFrameCode = bySkipFramePCode1920x1080;
                    iCodeSize = sizeof(bySkipFramePCode1920x1080);
                    break;
                }

                case MAKE_LONG(816, 1088):
                case MAKE_LONG(1088, 816):
                {
                    pbySkipFrameCode = bySkipFramePCode816x1080;
                    iCodeSize = sizeof(bySkipFramePCode816x1080);
                    break;
                }

                default:
    			{
    				printf("%dx%d size is not supported for skip\n", s->physical_width, s->physical_height );
    				assert(0);
    			}
    		}
    	}
    	else // Frame B
    	{
    		switch ( uiImageSize )
    		{
                case MAKE_LONG(160, 128):
                case MAKE_LONG(128, 160):
    			{
    				pbySkipFrameCode = bySkipFrameBCode160x128;
    				iCodeSize = sizeof(bySkipFrameBCode160x128);
    				break;
    			}
                case MAKE_LONG(176, 144):
                case MAKE_LONG(144, 176):
                {
                    pbySkipFrameCode = bySkipFrameBCode176x144;
                    iCodeSize = sizeof(bySkipFrameBCode176x144);
                    break;
                }
                case MAKE_LONG(320, 112):
                case MAKE_LONG(112, 320):
                {
                    pbySkipFrameCode = bySkipFrameBCode320x112;
                    iCodeSize = sizeof(bySkipFrameBCode320x112);
                    break;
                }
                case MAKE_LONG(320, 176):
                case MAKE_LONG(176, 320):
                {
                    pbySkipFrameCode = bySkipFrameBCode320x176;
                    iCodeSize = sizeof(bySkipFrameBCode320x176);
                    break;
                }
                case MAKE_LONG(320, 192):
                case MAKE_LONG(192, 320):
    			{
    				pbySkipFrameCode = bySkipFrameBCode320x180;
    				iCodeSize = 3;
    				break;
    			}
                case MAKE_LONG(320, 240):
                case MAKE_LONG(240, 320):
    			{
    				pbySkipFrameCode = bySkipFrameBCode320x240;
    				iCodeSize = sizeof(bySkipFrameBCode320x240);
    				break;
    			}
                case MAKE_LONG(352, 288):
                case MAKE_LONG(288, 352):
                {
                    pbySkipFrameCode = bySkipFrameBCode352x288;
                    iCodeSize = sizeof(bySkipFrameBCode352x288);
                    break;
                }
                case MAKE_LONG(640, 368):
                case MAKE_LONG(368, 640):
                {
                    pbySkipFrameCode = bySkipFrameBCode640x360;
                    iCodeSize = sizeof(bySkipFrameBCode640x360);
                    break;
                }
    			case MAKE_LONG(640, 480):
    			case MAKE_LONG(480, 640):
    			{
    				pbySkipFrameCode = bySkipFrameBCode640x480;
    				iCodeSize = sizeof(bySkipFrameBCode640x480);
    				break;
    			}
    			case MAKE_LONG(720, 480):
    			case MAKE_LONG(480, 720):
    			{
    				pbySkipFrameCode = bySkipFrameBCode720x480;
    				iCodeSize = sizeof(bySkipFrameBCode720x480);
    				break;
    			}
                case MAKE_LONG(720, 576):
                case MAKE_LONG(576, 720):
    			{
    				pbySkipFrameCode = bySkipFrameBCode720x576;
    				iCodeSize = sizeof(bySkipFrameBCode720x576);
    				break;
    			}
                case MAKE_LONG(800, 448):
                case MAKE_LONG(448, 800):
                {
                    pbySkipFrameCode = bySkipFrameBCode800x448;
                    iCodeSize = sizeof(bySkipFrameBCode800x448);
                    break;
                }
    			case MAKE_LONG(800, 592):
    			case MAKE_LONG(592, 800):
    			{
    				pbySkipFrameCode = bySkipFrameBCode800x592;
    				iCodeSize = sizeof(bySkipFrameBCode800x592);
    				break;
    			}

    			case MAKE_LONG(800, 608):
    			case MAKE_LONG(608, 800):
    			{
    				pbySkipFrameCode = bySkipFrameBCode800x608;
    				iCodeSize = sizeof(bySkipFrameBCode800x608);
    				break;
    			}

                case MAKE_LONG(848, 480):
                case MAKE_LONG(480, 848):
                {
                    pbySkipFrameCode = bySkipFrameBCode848x480;
                    iCodeSize = sizeof(bySkipFrameBCode848x480);                        
                    break;
                }

                case MAKE_LONG(960, 544):
                {
                    pbySkipFrameCode = bySkipFrameBCode960x544;
                    iCodeSize = sizeof(bySkipFrameBCode960x544);                        
                    break;
                }

                case MAKE_LONG(864, 480):
                case MAKE_LONG(480, 864):
                {
                    pbySkipFrameCode = bySkipFrameBCode864x480;
                    iCodeSize = sizeof(bySkipFrameBCode864x480);
                    break;
                }

                case MAKE_LONG(960, 720):
                case MAKE_LONG(720, 960):
                {
                    pbySkipFrameCode = bySkipFrameBCode960x720;
                    iCodeSize = sizeof(bySkipFrameBCode960x720);
                    break;
                }

                case MAKE_LONG(1024, 576):
                case MAKE_LONG(576, 1024):
                {
                    pbySkipFrameCode = bySkipFrameBCode1024x576;
                    iCodeSize = sizeof(bySkipFrameBCode1024x576);
                    break;
                }
                
    			case MAKE_LONG(1280, 720):
    			case MAKE_LONG(720, 1280):
    			{
    				pbySkipFrameCode = bySkipFrameBCode1280x720;
    				iCodeSize = sizeof(bySkipFrameBCode1280x720);
    				break;
    			}

                case MAKE_LONG(1440, 1088):
                case MAKE_LONG(1088, 1440):
                {
                    pbySkipFrameCode = bySkipFrameBCode1440x1080;
                    iCodeSize = sizeof(bySkipFrameBCode1440x1080);
                    break;
                }
                case MAKE_LONG(1920, 1088):
                case MAKE_LONG(1088, 1920):
                {
                    pbySkipFrameCode = bySkipFrameBCode1920x1080;
                    iCodeSize = sizeof(bySkipFrameBCode1920x1080);
                    break;
                }
                case MAKE_LONG(816, 1088):
                case MAKE_LONG(1088, 816):
                {
                    pbySkipFrameCode = bySkipFrameBCode816x1080;
                    iCodeSize = sizeof(bySkipFrameBCode816x1080);
                    break;
                }

                default:
    			{
    				printf("%dx%d size is not supported for skip\n", s->physical_width, s->physical_height );
    				assert(0);
    			}
    		}
        }

        for ( i = 0; i < iCodeSize; i++ )
        {
            put_bits(&picture->pb, 8, pbySkipFrameCode[i] );    

        }
    }
    else /* CAVLC */
    {
        // encode mb_skip_run of all MBs
        put_ue_bits(&picture->pb, s->mb_in_frame );

        // rbsp_trailing_bits - byte alignment
        ff_h264_stuffing(&picture->pb);
    }

    flush_put_bits(&picture->pb);
}



void
h264_write_sei_messages(AVCodecContext *avctx,
                       AVFrame *data,
                       BOOL bFirstTime )
{
    H264zEncContext *s = avctx->priv_data;
	AVFrame* pic   = data;
    BOOL bWriteMessage = FALSE;

#ifndef AUD_POSITION_AFTER_PPS
    // encode access unit delimiter
    if( avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_AUD)
    {
            h264z_encode_aud(&pic->pb, pic->pict_type);
    }
#endif

    // Is it first frame in real clip?
    if  ( ( s->headers_written == 0 ) && 
          ( pic->original_pict_type == AV_FRAME_TYPE_I ) )   
          /*bFirstFrameInClip == TRUE )*/
    {
        if ( pic->dummy_gop == false )
        {
            s->headers_written = 1;
        }

        // Already have SPS/PPS
        if ( avctx->pic_init_qp != 0 )
        {
            s->pic_init_qp = avctx->pic_init_qp;
        }
        else        
        {
            //s->pic_init_qp = (INT)s->brc_ctx.uiQscaleCurrent;
            s->pic_init_qp = 26;
        }

        h264z_encode_sps( s, &pic->pb );
        h264z_encode_pps( s, &pic->pb );

        //inform of size of header - pps/sps headers are byte aligned
        pic->extra_header_length =  get_bit_count(&pic->pb)/8;

        // Set frame_num to zero
        s->frame_num = 0;
        s->real_frame_num = 0;

        // Set first frame to be IDR frame
        pic->idr_frame = 1;
    }
    else
    {
        //inform of size of header
        pic->extra_header_length =  0;
    }

    s->slice_qp_delta = (INT)s->brc_ctx.uiQscaleCurrent - s->pic_init_qp;

#ifdef AUD_POSITION_AFTER_PPS
    // encode access unit delimiter
    if(avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_AUD)
    {
            h264z_encode_aud(&pic->pb, pic->pict_type);
    }
#endif
    // In the first time, check if we need to write SEI message
    if ( bFirstTime == TRUE && pic != NULL )
    {
        // should be absent before first frame
        if ( (s->pict_type == I_TYPE) && ((s->real_frame_num % h264z_encode_get_gop_size(s)) == 0))
        {
            // mark this frame as one that needs sei message
            pic->idr_sei_message_written = 1;
            bWriteMessage = TRUE;
        }
        else
        {
            pic->idr_sei_message_written = 0;
        }
    }
    else
    {
        bWriteMessage = pic->idr_sei_message_written;
    }

    if ( bWriteMessage == TRUE )
    {
        // insert all  Buf period and recovery point SEI in start of GOP only;
        if (avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_BP_SEI)
        {
            h264z_encode_sei_message(s, H264_SEI_BUFFERING_PERIOD, pic, bFirstTime);
        }
        if (avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_PT_SEI)
        {
            h264z_encode_sei_message(s, H264_SEI_PICTURE_TIMING, pic, bFirstTime);
        }
        if (avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_RP_SEI)
        {
            h264z_encode_sei_message(s, H264_SEI_RECOVERY_POINT, pic, bFirstTime);
        }
    }
    else
    {
        if (avctx->syntax_elements_flag & MPEG4_SYNTAX_ELEMENT_PT_SEI)
        {
            h264z_encode_sei_message(s, H264_SEI_PICTURE_TIMING, pic, bFirstTime);
        }
    }

}
/**

%func

    h264z_encode_preconfig

%decs

*/

int
h264z_encode_preconfig(AVCodecContext *avctx,
                     unsigned char *buf,
                     int buf_size,
                     AVFrame *data,
                     BOOL bFirstTime )
{
    H264zEncContext *s = avctx->priv_data;
    AVFrame* pic   = data;


    init_put_bits(&pic->pb, buf, buf_size, NULL, NULL);

    // Do we have a picture to encode?
    if ( pic != NULL )
    {
        if ( pic->scene_cut == 1 )
        {
            s->frame_num = 0;
        }

        h264_write_sei_messages(avctx, pic, bFirstTime );

        pic->frame_num = s->frame_num;

    }
    return 0;
}
/**

%func

    h264z_encode_picture

%decs

*/

int
h264z_encode_picture(AVCodecContext *avctx,
                     unsigned char *buf,
                     int buf_size,
                     void *data)
{
    H264zEncContext *s = avctx->priv_data;
	AVFrame* pic_arg   = data;
    AVFrame* original_pic_arg   = data;
    int slice_type;
    UINT uiCodeSize = 0;
	BOOL bFirstFrameInClip = FALSE;
	BOOL bQuarterFrame = FALSE;
    int GOP_size = h264z_encode_get_gop_size(s);
	BOOL bIntraOnly = s->avctx->is_intra_only;

    if ( pic_arg->pict_type == AV_FRAME_TYPE_NULL ) 
    {
        pic_arg = NULL;
    }

    load_input_picture(s, pic_arg);

    // Do we have a picture to encode?
    if ( s->current_picture_ptr != NULL )
    {
        s->current_picture_ptr->code_buf = buf;

		s->current_picture_ptr->pic_arg = pic_arg;

        s->current_picture_ptr->code_buf_size = buf_size;

        //
        // Check skip mode of BRC

        // In B frame save state as before
        // In I or P start or end the skip sequence
        // if an I was skipped ,I insertion should be done at the
        // end of skip sequence.

		if ( ( s->current_picture_ptr->pict_type == I_TYPE ) ||
			 ( s->current_picture_ptr->pict_type == P_TYPE ) )
		{
			s->brc_ctx.uiCPBSkipModeFlag = 0;
		}

		if ( ( s->current_picture_ptr->pict_type == B_TYPE ) && ( s->brc_ctx.uiCPBSkipModeFlag == 1 ) )
		{
            // In case of Skip Frame with 2 video codecs, we are handling the skip frame in
            // a lower level (we encode the frame as normal, and replace the encoded frame by
            // a skip frame 
            if ( avctx->num_of_codecs == 2 ) 
            {
                s->current_picture_ptr->replace_by_skip = 1;
            }
            else
            {
                s->current_picture_ptr->skip_frame = 1;
            }
		}

#ifdef ZAV
		// In case of two video codecs, we should call GetTargetRatio here only for the FIRST frame
		// For other frames it will be called later, somewhere in  H264Encoder.c
		if ( ( s->headers_written == 0 ) &&
     		 ( s->current_picture_ptr->pict_type == AV_FRAME_TYPE_I )
				)
		{
			bFirstFrameInClip = TRUE;
		}

		if ( s->current_picture_ptr->pict_type >= AV_FRAME_TYPE_SAMPLE_I4_0 &&
			 s->current_picture_ptr->pict_type <= AV_FRAME_TYPE_SAMPLE_P )
		{
			bQuarterFrame = TRUE;
		}

		if  ( ( bFirstFrameInClip == TRUE ) ||
			  ( avctx->num_of_codecs == 1 ) ||
			  ( bQuarterFrame == TRUE ) 
			)
		{
			// We get Target Ratio - exception is first I frame
			s->current_picture_ptr->target_ratio_called = 1;
			BrcGetTargetRatio( &s->brc_ctx, s->current_picture_ptr );
		}
		else
		{
			s->current_picture_ptr->target_ratio_called = 0;
		}
#endif

        s->flags = avctx->flags;

        // Set current picture type.
        s->pict_type = s->current_picture_ptr->pict_type;

        s->physical_width        = s->current_picture_ptr->width;
        s->physical_height       = s->current_picture_ptr->height;

        // YUV buffer should be set according to external values.
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].frame_width  = s->current_picture_ptr->frame_width;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].frame_height = s->current_picture_ptr->frame_height;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].left         = s->current_picture_ptr->left;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].top          = s->current_picture_ptr->top;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].width        = s->current_picture_ptr->width;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].height       = s->current_picture_ptr->height;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].bpp          = 16;

        if ( s->avctx->pix_fmt == PIX_FMT_YUV420Z )
        {
            avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].bpp = 8;
        }

        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE_C].frame_width  = s->current_picture_ptr->chroma_frame_width;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE_C].frame_height = s->current_picture_ptr->chroma_frame_height;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE_C].left         = s->current_picture_ptr->chroma_left;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE_C].top          = s->current_picture_ptr->chroma_top;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE_C].width        = s->current_picture_ptr->chroma_width;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE_C].height       = s->current_picture_ptr->chroma_height;
        avctx->block_buf[AV_BLOCK_BUFFER_IMAGE_C].bpp          = 8;

        if ( s->current_picture_ptr->skip_frame == 1 )
        {
            // For skip frame code is for quantizer = 20
            s->brc_ctx.uiQscaleCurrent = 20;
        }

        if ( h264z_frame_start(s, avctx) < 0 )
        {
            printf( "h264z_frame_start failed, returning\n" );
            return -1;
        }

        slice_type = h264z_get_slice_type (s->pict_type);

		// Set IDR frame if needed
        if ( s->current_picture_ptr->display_picture_number == 0 )
        {
            s->current_picture_ptr->idr_frame = 1;
        }
        else
        {
            s->current_picture_ptr->idr_frame = 0;
        }

		// we check gop_frame_count cause frame_num may be greater than GOP_size (if I frame was skipped) (Boris B.)
		// i.e. frame_num 20 will be actually 4 (20-16), cause 16 is the maximal value for frame_num
		if ( (s->pict_type == I_TYPE) &&
		     (bIntraOnly || ((s->current_picture_ptr->frame_in_gop % GOP_size) == 0)))
		{
			s->current_picture_ptr->idr_frame = 1;
            s->gop_start = s->current_picture_ptr->display_picture_number;
		}
		// We save gop start per frame cause when we use 2 VC, we can work on previous number
		s->current_picture_ptr->gop_start = s->gop_start;

		if ( avctx->num_of_codecs == 1 ) 
		{
			h264z_encode_preconfig(avctx,
                  	   		   	   buf,
                     		       buf_size,
                     		       s->current_picture_ptr,
							       TRUE );
		}

        if ( slice_type == H264_SLICE_TYPE_I || slice_type == H264_SLICE_TYPE_P )
        {
            s->current_picture_ptr->nal_ref_idc = 1;
        }
        else // H264_SLICE_TYPE_B
        {
            s->current_picture_ptr->nal_ref_idc = 0;
        }

        //  We write headers here only if we run with 1 VC
		if ( avctx->num_of_codecs == 1 )
		{
			h264z_encode_slice_header( s, s->current_picture_ptr, slice_type, TRUE );

			// CABAC alignment
			if ( s->entropy_mode == H264_ENTROPY_MODE_CABAC )
			{
				cabac_alignment(&s->current_picture_ptr->pb);
			}

			if ( s->current_picture_ptr->skip_frame == 1 )
			{
				TimingMarkF( TIMING_GROUP_VIDEO_BRC_HRD, "[BRC TOOL] Skip frame, Frame %d",
							 s->ready_picture_ptr->coded_picture_number);

				h264_encode_skip_frame( s, s->current_picture_ptr );
			}
		}

        uiCodeSize = H264EncodeFrame( s );
    }
    else
    {
        // Get ready frame.
        uiCodeSize = H264EncodeFrame( s );
    }

	if ( s->ready_picture_ptr != NULL )
    {
        /* Check if need to change the frame to SKIP or/and add STUFFING bits */
		// The next part of code is only when only one VC is used
		// When we used two VC's, the decisions are made in Codec Level
		if ( ( s->brc_ctx.bHRDEnable == TRUE ) && ( avctx->num_of_codecs == 1 ) )
		{
			s->brc_ctx.bHRDBufferOverflown = ( ( s->brc_ctx.iHRDApproxNewBufFullness + ((INT) uiCodeSize * 8 ) )
										   > ( s->brc_ctx.iHRDBufferSize * 15 / 16 ) ? TRUE : FALSE );

				// we cannot skip first I frame
				// The last check is just to ensure we do not replace SKIP frame by SKIP frame - useless (Boris B.)
			if ( ( s->brc_ctx.bHRDBufferOverflown == TRUE ) &&
				 ( s->ready_picture_ptr->display_picture_number != 0 ) &&
				 ( s->ready_picture_ptr->skip_frame == 0 ) )
			{
                TimingMarkF( TIMING_GROUP_VIDEO_BRC_HRD, "Encoding OVERFLOW - Insert skip frame. Frame %d TYPE %s",
                             s->ready_picture_ptr->coded_picture_number,
                             s->ready_picture_ptr->pict_type == I_TYPE ? "I" :
                             s->ready_picture_ptr->pict_type == P_TYPE ? "P" :
                             s->ready_picture_ptr->pict_type == B_TYPE ? "B" : "ERROR" );

                if ( s->ready_picture_ptr->pict_type == I_TYPE )
				{
					// s->idr_frame is 1 for GOP - I Frames
					// We need to know if we skipped GOP I frame & how much of them
					s->brc_ctx.uiSkipedGopIFrames += s->ready_picture_ptr->idr_frame;
				}

				if ( ( s->ready_picture_ptr->pict_type == I_TYPE ) || ( s->ready_picture_ptr->pict_type == P_TYPE ) )
				{
					//s->brc_ctx.uiCPBSkipModeMissedIFlag = 1;
					//s->brc_ctx.bIFrameReplacedByP = TRUE;
					s->ready_picture_ptr->pict_type = P_TYPE;
					s->brc_ctx.uiCPBSkipModeFlag = 1;
				}


				if ( s->ready_picture_ptr->pict_type == B_TYPE )
				{
					s->brc_ctx.uiQscaleCurrent = 20;
					s->slice_qp_delta = (INT)s->brc_ctx.uiQscaleCurrent - s->pic_init_qp;
                    s->ready_picture_ptr->idr_frame = 0;
					s->ready_picture_ptr->nal_ref_idc = 0;
				}
				else
				{
					s->brc_ctx.uiQscaleCurrent = 20;
					s->slice_qp_delta = (INT)s->brc_ctx.uiQscaleCurrent - s->pic_init_qp;
					s->ready_picture_ptr->nal_ref_idc = 1;
					s->ready_picture_ptr->idr_frame = 0;

				}

                // assuming same buf_size every frame
				init_put_bits(&s->ready_picture_ptr->pb, s->ready_picture_ptr->code_buf, buf_size, NULL, NULL);

                h264_write_sei_messages( avctx, s->ready_picture_ptr, FALSE );

				h264z_encode_slice_header( s, s->ready_picture_ptr, h264z_get_slice_type( s->ready_picture_ptr->pict_type ), FALSE );

				// CABAC alignment
				if ( s->entropy_mode == H264_ENTROPY_MODE_CABAC )
				{
					cabac_alignment(&s->ready_picture_ptr->pb);
				}

				h264_encode_skip_frame( s, s->ready_picture_ptr );

				// Change Buffers if skipping here P or I frame
				if ( s->ready_picture_ptr->pict_type == P_TYPE )
				{
                    // Reference context of current frame is not valid, since we drop the frame.
                    // Use reference context of forward reference
                    CodecBuffersRelease( s->avctx, s->ready_picture_ptr->ref_context );
                    CodecBuffersGet( s->avctx, s->last_picture_ptr->ref_context, s->ready_picture_ptr->pict_type );
					s->ready_picture_ptr->ref_context = s->last_picture_ptr->ref_context;
                }

				uiCodeSize = pbBufPtr(&s->ready_picture_ptr->pb) - s->ready_picture_ptr->code_buf;

			} //end of if ( s->brc_ctx.bHRDBufferOverflown == TRUE )


			BrcHRDGetStuffingBitsNumber(&s->brc_ctx, uiCodeSize * 8 );

			s->brc_ctx.bHRDBufferUnderflown = ( ( ( s->brc_ctx.iStuffingBits > 0 ) && ( s->brc_ctx.bHRDBufferOverflown == FALSE ) )
												? TRUE : FALSE );

			h264z_add_stuffing_bits(s, s->ready_picture_ptr, uiCodeSize );

			if ( s->brc_ctx.iStuffingBits > 0)
			{
				uiCodeSize += ( 4 + s->brc_ctx.iStuffingBits / 8 );
			}

		} //if ( s->brc_ctx.bHRDEnable == TRUE )


		// Call Feedback here only if we run on 1 VC

		{
             UINT pict_type = original_pic_arg->pict_type;

             *original_pic_arg = *s->ready_picture_ptr;

             // For sample frames we need to return the original "sample" type.
             // Sample frames are processed in blocking method, therefore the type
	         // Can be taken from "original_pic_arg" frame.
		     if ( pict_type >= FF_SAMPLE_I4_TYPE_0 )
             {
                 original_pic_arg->pict_type = pict_type;
             }
        }

		if ( avctx->num_of_codecs == 1 )
		{
			original_pic_arg->pict_type = s->ready_picture_ptr->pict_type;
    
             // BRC calculation for next frame encoding
    		BrcFeedback( &s->brc_ctx, original_pic_arg, uiCodeSize * 8 );
    
    		// Did we really encoded frame GOP I ? If yes - stop incrementing frame_num & start new gop counter
    		if ( s->ready_picture_ptr->pict_type == I_TYPE )
    		{
    			s->brc_ctx.uiSkipedGopIFrames = 0;
    		}
    	}

			h264z_frame_end(s);
    }
    else
    {
        original_pic_arg->code_buf = NULL;
    }

    return uiCodeSize;
}

/*

    h264z_encode_verify_settings

    Verify legal setting for recording:
    1. Valid profile
    2. Valid max B-frames for baseline profile
    3. Valid entropy for baseline profile

*/

void
h264z_encode_verify_settings(AVCodecContext *avctx)
{
    H264zEncContext *s = avctx->priv_data;

    switch ( avctx->profile )
    {
        case H264_BASELINE_PROFILE:
        {
            // Verify valid max B-frames for given profile
            if ( avctx->max_b_frames > 0 )
            {
                OsSysHaltF("Invalid H264 settings: Baseline profile does not support B-frames. Max B frames is %d\n", avctx->max_b_frames  );
            }

            // Verify valid entropy for given profile
            if ( s->flags & CODEC_FLAG_CABAC )
            {
                OsSysHaltF("Invalid H264 settings: Baseline profile does not support CABAC.\n" );
            }
            break;
        }

        case H264_MAIN_PROFILE:
        {
            break;
        }

        default:
        {
            OsSysHaltF("Invalid H264 settings: profile %d\n", avctx->profile );
        }
    }

    return;
}

/*

    h264z_encode_init

*/

int
h264z_encode_init(AVCodecContext *avctx)
{
    H264zEncContext *s = avctx->priv_data;
    int i;
    int GOP_size;
    int log2_max_frame_num_minus4;

    avctx->pix_fmt    = PIX_FMT_YUV422Z;
    s->avctx          = avctx;

    if ( h264z_common_init(s) < 0 )
    {
        return -1;
    }

    h264z_encode_verify_settings( avctx );

    s->physical_width = avctx->physical_width;
    s->physical_height = avctx->physical_height;

    s->logical_width = avctx->logical_width;
    s->logical_height = avctx->logical_height;

    // Calculate the total number of macro block in frame.
    s->mb_in_frame    = ( s->physical_width / MB_WIDTH_IN_PIXELS ) *
                        ( s->physical_height / MB_HEIGHT_IN_PIXELS );

    s->max_b_frames   = avctx->max_b_frames;
    s->codec_id       = avctx->codec->id;

    s->encoding = 1;

    // Set buffer allocation functions.
    s->avctx->get_buffer     = encoder_get_buffer;
    s->avctx->release_buffer = encoder_release_buffer;

    // Initialize deblocking filter
    if ( s->flags & CODEC_FLAG_DEBLOCKING )
    {
        s->deblocking_filter = 1;
        s->slice_alpha_c0_offset = 0;
        s->slice_beta_offset = 0;
    }
    else
    {
        s->deblocking_filter = 0;
        s->slice_alpha_c0_offset = 0;
        s->slice_beta_offset = 0;
    }

    // Get the GOP size
    GOP_size = h264z_encode_get_gop_size(s);

    // Set entropy mode
    if ( s->flags & CODEC_FLAG_CABAC )
    {
        s->entropy_mode = H264_ENTROPY_MODE_CABAC;
    }
    else
    {
        s->entropy_mode = H264_ENTROPY_MODE_VLC;
    }

    // Set maximum frame num in an IDR
    log2_max_frame_num_minus4 = av_log2(GOP_size) - 4;

//    log2_max_frame_num_minus4++;

    if (log2_max_frame_num_minus4 < 0)
    {
        log2_max_frame_num_minus4 = 0;
    }

    s->sps.log2_max_frame_num = log2_max_frame_num_minus4 + 4;

	// From here max_frame_num will be as follows
	// For GOP <32 = max_frame_num = 15
	// It is not exactly to the spec (i think) but it is requested by SONY
    s->max_frame_num = 1 << s->sps.log2_max_frame_num;

    // Set frame_num to zero
    s->frame_num = 0;

    //
    s->sps.log2_max_poc_lsb = log2_max_frame_num_minus4 + 4;

    s->max_pic_order_cnt_lsb = 1 << s->sps.log2_max_poc_lsb;

    //
    s->gop_start = 0;

    // Set gaps_in_frame_num_value_allowed_flag
    s->gaps_in_frame_num_value_allowed_flag = 0;

    // Initialize picture and slice qp
    s->slice_qp_delta = 0;

    // Set headers_written to false
    s->headers_written = 0;

    s->sps.profile_idc = avctx->profile;
    s->sps.level_idc = avctx->level;

    // CPB size needed for Brc (h264 only)
    switch ( s->sps.level_idc )
    {
        // default hrd buffer size and input bit rate, according to h264 standard:
        // Level 3  : Size: 10Mbit, Rate: 10Mbit/s
        // Level 3.1: Size: 14Mbit, Rate: 14Mbit/s
        // Level 4  : Size: 25Mbit, Rate: 20Mbit/s

        case 41:
        {
            s->brc_ctx.iHRDBufferSize = 62500000;
            s->brc_ctx.iHRDInputRate  = 50000000;
            break;
        }

        case 40:
        {
            s->brc_ctx.iHRDBufferSize = 25000000;
            s->brc_ctx.iHRDInputRate  = 20000000;
            break;
        }

        case 31:
        {
            s->brc_ctx.iHRDBufferSize = 14000000;
            s->brc_ctx.iHRDInputRate  = 14000000;
            break;
        }

        case 30:
        {
            s->brc_ctx.iHRDBufferSize = 10000000;
            s->brc_ctx.iHRDInputRate  = 10000000;
            break;
        }

        default:
        {
            printf("ERROR: unsupported level for H264\n");
        }
    }

    if ( avctx->hrd_enable == TRUE )
    {
        s->avctx->syntax_elements_flag |= ( MPEG4_SYNTAX_ELEMENT_VUI     |
                                            MPEG4_SYNTAX_ELEMENT_NAL_HDR |
                                            MPEG4_SYNTAX_ELEMENT_VCL_HDR |
											MPEG4_SYNTAX_ELEMENT_BP_SEI  |
        									MPEG4_SYNTAX_ELEMENT_PT_SEI  |
        									MPEG4_SYNTAX_ELEMENT_RP_SEI );
    }
    else
    {
        s->avctx->syntax_elements_flag &= ~( MPEG4_SYNTAX_ELEMENT_NAL_HDR |
                                             MPEG4_SYNTAX_ELEMENT_VCL_HDR );
    }

    if ( avctx->flags & CODEC_FLAG_INTERLACED_MODE )
    {
        s->avctx->syntax_elements_flag |= ( MPEG4_SYNTAX_ELEMENT_VUI |
                                            MPEG4_SYNTAX_ELEMENT_BP_SEI |
                                            MPEG4_SYNTAX_ELEMENT_PT_SEI );
    }

    s->pic_init_qp = 28;

    // Initialize the BRC structure for this frame
    BrcInit( &(s->brc_ctx), avctx );

    if (s->brc_ctx.uiBrcMode == AV_BRC_MODE_CONST_QUALITY)
    {
        s->brc_ctx.iAvgActivity = avctx->iAvgActivity;
        assert( s->brc_ctx.iAvgActivity != 0 );
    }

    BrcDebugStart( &(s->brc_ctx) );

    // Set geometry for Ref Y and Ref C buffers.
    // Buffers geometry is set according to bandwidth optimization
    // options and deblocking filter option.
    for ( i = AV_BLOCK_BUFFER_Y; i < AV_BLOCK_BUFFER_NUM; i++ )
    {
       MepGetBufferGeometry( i,
                              s->physical_width,
                              s->physical_height,
                              s->flags,
                              FALSE,
                              (s->deblocking_filter == 1),
                              &s->avctx->block_buf[i] );
    }

    H264EncoderInit( s );

    return 0;

}

/**

%func

    h264z_encode_end

%desc

    The function ends the encoder mode: release the internal buffers...
    The function is called when all frames has been encoded.
*/

int
h264z_encode_end(AVCodecContext *avctx)
{
    H264zEncContext *s = avctx->priv_data;
    int i;

    BrcDebugEnd( &(s->brc_ctx) );
    h264z_common_end(s);

    // Release the reconstructre/reference buffers
    for ( i = 0; i < H264_MAX_PICTURE_COUNT; i++ )
    {
        AVFrame* pic = &s->picture[i];
        encoder_release_external_buffer( pic );
    }

    H264EncoderEnd( s );

    return 1;
}


/*
*/

#ifdef CONFIG_ENCODERS

AVCodec h264z_encoder = {
    "h264z",
    CODEC_TYPE_VIDEO,
    CODEC_ID_H264Z,
    sizeof(H264zEncContext),
    h264z_encode_init,
    h264z_encode_picture,
    h264z_encode_end,
};

#endif //CONFIG_ENCODERS

/*************************************************************************/
/*                                                                       */
/*                            H264 Decoder                               */
/*                                                                       */
/*************************************************************************/

#define END_NOT_FOUND -100

/**
 * finds the end of the current frame in the bitstream.
 * @return the position of the first byte of the next frame, or -1
 */
static int
h264z_find_frame_end(H264zEncContext *s, const uint8_t *buf, int buf_size){
    int i;
    uint32_t state;
    ParseContext *pc= &s->parse_context;

//    mb_addr= pc->mb_addr - 1;
    state= pc->state;

    for(i=0; i<=buf_size; i++){
        if((state&0xFFFFFF1F) == 0x101 || (state&0xFFFFFF1F) == 0x102 || (state&0xFFFFFF1F) == 0x105){
            if(pc->frame_start_found){
                // If there isn't one more byte in the buffer
                // the test on first_mb_in_slice cannot be done yet
                // do it at next call.
                if (i >= buf_size) break;
                if (buf[i] & 0x80) {
                    // first_mb_in_slice is 0, probably the first nal of a new
                    // slice
//                    printf("find_frame_end frame_end_found, state = %08x, pos = %d\n", state, i);
                    pc->state= (uint32_t)-1;
                    pc->frame_start_found= 0;
                    return i-4;
                }
           }
            pc->frame_start_found = 1;
        }
        if((state&0xFFFFFF1F) == 0x107 || (state&0xFFFFFF1F) == 0x108 || (state&0xFFFFFF1F) == 0x109
           || (state&0xFFFFFF1F) == 0x10A || (state&0xFFFFFF1F) == 0x10B ){
           if(pc->frame_start_found)
		   	{
                pc->state = (uint32_t)-1;
                pc->frame_start_found= 0;
                return i-4;
           }
        }
        if (i<buf_size)
            state= (state<<8) | buf[i];
    }

    pc->state= state;
    return END_NOT_FOUND;
}

/**
 * combines the (truncated) bitstream to a complete frame
 * @returns -1 if no complete frame could be created
 */
static int
h264z_combine_frame( H264zEncContext *s, int next, uint8_t **buf, int *buf_size)
{
    ParseContext *pc= &s->parse_context;

#if 0
    if(pc->overread)
    {
        printf("overread %d, state:%X next:%d index:%d o_index:%d\n", pc->overread, pc->state, next, pc->index, pc->overread_index);
        printf("%X %X %X %X\n", (*buf)[0], (*buf)[1],(*buf)[2],(*buf)[3]);
    }
#endif

    /* copy overreaded byes from last frame into buffer */
    for(; pc->overread>0; pc->overread--)
    {
        pc->buffer[pc->index++]= pc->buffer[pc->overread_index++];
    }

    pc->last_index= pc->index;

    /* copy into buffer end return */
    if(next == END_NOT_FOUND)
    {
        pc->buffer = av_fast_realloc(pc->buffer, (unsigned int *)&pc->buffer_size,
				(*buf_size) + pc->index + FF_INPUT_BUFFER_PADDING_SIZE);
        memcpy(&pc->buffer[pc->index], *buf, *buf_size);
        pc->index += *buf_size;
        return -1;
    }

    pc->overread_index= pc->index + next;

    /* append to buffer */
    if(pc->index)
    {
        pc->buffer= av_fast_realloc(pc->buffer, (unsigned int *)&pc->buffer_size,
				next + pc->index + FF_INPUT_BUFFER_PADDING_SIZE);
        memcpy(&pc->buffer[pc->index], *buf, next + FF_INPUT_BUFFER_PADDING_SIZE );
        pc->index = 0;
        *buf= pc->buffer;
        *buf_size= pc->last_index + next;
    }

    /* store overread bytes */
    for(;next < 0; next++)
    {
        pc->state = (pc->state<<8) | pc->buffer[pc->last_index + next];
        pc->overread++;
    }

#if 0
    if(pc->overread)
    {
        printf("overread %d, state:%X next:%d index:%d o_index:%d\n", pc->overread, pc->state, next, pc->index, pc->overread_index);
        printf("%X %X %X %X\n", (*buf)[0], (*buf)[1],(*buf)[2],(*buf)[3]);
    }
#endif

    return 0;
}

//////////////////////////////////////////////////////////////////////////////
// Decoder image buffers allocation/release functions
//////////////////////////////////////////////////////////////////////////////

static int
decoder_get_buffer(AVCodecContext *avctx,
                   AVFrame        *pic)
{
    H264zEncContext *s = avctx->priv_data;

    BOOL bEnableReconstructedClients = FALSE;

    if ( s->flags & CODEC_FLAG_INTERNAL_DRAM_MODE )
    {
        bEnableReconstructedClients = TRUE;
    }

    // Allocate buffer for decoded image
    pic->data[AV_BUF_YUV_IMAGE] = PoolAlloc( avctx->hPool,
                                             s->avctx->block_buf[AV_BLOCK_BUFFER_IMAGE].size,
                                             POOL_MEMORY_UNCACHED | POOL_WAIT_FOREVER | POOL_DISABLE_FAILURE_PRINT);

    pic->ref_context = CodecBuffersGet( avctx, NULL, pic->pict_type );

    if ( avctx->pix_fmt == PIX_FMT_YUV420Z )
    {
        // Allocate buffer for decoded image (chroma)
        pic->data[AV_BUF_YUV_IMAGE_C] = PoolAlloc( avctx->hPool,
                                                   s->avctx->block_buf[AV_BLOCK_BUFFER_IMAGE_C].size,
                                                   POOL_MEMORY_UNCACHED | POOL_WAIT_FOREVER | POOL_DISABLE_FAILURE_PRINT);

 		if ( bEnableReconstructedClients == FALSE )
		{
        	// In case we don't use recontructed clients - we should PoolGet converter data
        	// since we are going to use it as reference
        	// Convertor data will be used
            PoolGet( NULL, pic->data[AV_BUF_YUV_IMAGE] );
            PoolGet( NULL, pic->data[AV_BUF_YUV_IMAGE_C] );
		}
    }

    pic->skip_frame = 0;

    return 0;
}

static void
decoder_release_buffer(AVCodecContext *avctx,
                       AVFrame        *pic)
{
    H264zEncContext *s = avctx->priv_data;

    BOOL bEnableReconstructedClients = FALSE;

    if ( s->flags & CODEC_FLAG_INTERNAL_DRAM_MODE )
    {
        bEnableReconstructedClients = TRUE;
    }

    if ( pic->buf_allocated & AV_BUF_REFERENCE_USED )
    {
        // In case of yuv420 format we need to release convertor buffers.
        if ( (avctx->pix_fmt == PIX_FMT_YUV420Z) && (bEnableReconstructedClients == FALSE) )
        {
            // In case we don't use recontructed clients - we should free converter data
            // since we did PoolGet before (we use it as reference)
            PoolFree( NULL, pic->data[AV_BUF_YUV_IMAGE]);
            PoolFree( NULL, pic->data[AV_BUF_YUV_IMAGE_C]);
        }

        CodecBuffersRelease( avctx, pic->ref_context );

        // Mark the entry as references not used
        pic->buf_allocated &= ~AV_BUF_REFERENCE_USED;
    }
}


/////////////////////////////

/*

%func

  decode_nal

*/

static uint8_t *
decode_nal(H264zEncContext *s, uint8_t *src, int *dst_length, int *consumed, int length){
    int i, si, di;
    uint8_t *dst;

    s->nal_unit_type= src[0]&0x1F;

    s->nal_ref_idc = 0;

    if ( ( s->nal_unit_type != H264_NAL_PPS ) && ( s->nal_unit_type != H264_NAL_SPS ) ) 
    {
        s->nal_ref_idc= src[0]>>5;
    }

    src++; length--;

    // We assume there is only one slice nal.
    if ( ( s->nal_unit_type != H264_NAL_SLICE ) &&
         ( s->nal_unit_type != H264_NAL_IDR_SLICE ) )
    {
        for(i=0; i+1<length; i+=2){
            if(src[i]) continue;
            if(i>0 && src[i-1]==0) i--;
            if(i+2<length && src[i+1]==0 && src[i+2]<=3){
                if(src[i+2]!=3){
                    /* startcode, so we must be past the end */
                length=i;
                }
                break;
            }
        }

        // Start code prevention element was not enountered
        if(i>=length-1){
    *dst_length= length;
    *consumed= length+1; //+1 for the header
    return src;
}
    }
    else
    {
        // For slice unit we scan only the header. Slice data is parsed by HW.
        // We assume that slice header doesn't exceed 200 bytes
        length = 200;
        i = 0;
    }

    s->rbsp_buffer= av_fast_realloc(s->rbsp_buffer, &s->rbsp_buffer_size, length+FF_INPUT_BUFFER_PADDING_SIZE);
    dst = s->rbsp_buffer;

    if (dst == NULL){
        return NULL;
    }

    // Copy stream up-to start code prevention element
    memcpy(dst, src, i);
    si=di=i;
    while(si+2<length){
        //remove escapes (very rare 1:2^22)
        if(src[si+2]>3){
            dst[di++]= src[si++];
            dst[di++]= src[si++];
        }else if(src[si]==0 && src[si+1]==0){
            if(src[si+2]==3){ //escape
                dst[di++]= 0;
                dst[di++]= 0;
                si+=3;
                continue;
            }
            else{
                // next start code was encountered.
                break;
            }
        }
        dst[di++]= src[si++];
    }

    // in case we reached the end of buffer, copy last 2 bytes.
    if ( si + 2 == length )
    {
        dst[di++]= src[si++];
        dst[di++]= src[si++];
    }

    *dst_length= di;
    *consumed= si + 1;//+1 for the header
    return dst;
}

/*

%func

  decode_rbsp_trailing

%desc

 * identifies the exact end of the bitstream
 * @return the length of the trailing, or 0 if damaged

*/

static int
decode_rbsp_trailing(uint8_t *src)
{
    int v= *src;
    int r;

//    tprintf("rbsp trailing %X\n", v);

    for(r=1; r<9; r++){
        if(v&1) return r;
        v>>=1;
    }
    return 0;
}


/*

    h264z_encode_hrd

*/

static int
decode_h264z_hrd(H264zEncContext *s, HRD *hrd)
{
	int SchedSelIdx;

	hrd->cpb_cnt_minus1 = get_ue_golomb(&s->gb);

	hrd->bit_rate_scale = get_bits(&s->gb, 4);

	hrd->cpb_size_scale = get_bits(&s->gb, 4);

	for(SchedSelIdx = 0; SchedSelIdx <= hrd->cpb_cnt_minus1; SchedSelIdx ++){
		hrd->bit_rate_value_minus1[SchedSelIdx] = get_ue_golomb(&s->gb);
		hrd->cpb_size_value_minus1[SchedSelIdx] = get_ue_golomb(&s->gb);
		hrd->cbr_flag[SchedSelIdx] = get_bits1(&s->gb);
	}

	hrd->initial_cpb_removal_delay_length_minus1 = get_bits(&s->gb, 5);
	hrd->cpb_removal_delay_length_minus1 = get_bits(&s->gb, 5);
	hrd->dpb_output_delay_length_minus1 = get_bits(&s->gb, 5);
	hrd->time_offset_length = get_bits(&s->gb, 5);

	return 1;
}

/*

    h264z_encode_vui

*/

static int
decode_h264z_vui(H264zEncContext *s, VUI *vui)
{
    vui->aspect_ratio_info_present_flag = get_bits1(&s->gb);

	if(vui->aspect_ratio_info_present_flag){
		vui->aspect_ratio_idc = get_bits(&s->gb, 8);

		if(vui->aspect_ratio_idc == 255){
			vui->sar_width =  get_bits(&s->gb, 16);
			vui->sar_height =  get_bits(&s->gb, 16);
		}
	}

	vui->overscan_info_present_flag = get_bits1(&s->gb);

	if(vui->overscan_info_present_flag){
		vui->overscan_appropriate_flag = get_bits1(&s->gb);
	}

	vui->video_signal_type_present_flag = get_bits1(&s->gb);

	if(vui->video_signal_type_present_flag){
		vui->video_format = get_bits(&s->gb, 3);
		vui->video_full_range_flag = get_bits1(&s->gb);
		vui->colour_description_present_flag = get_bits1(&s->gb);

		if(vui->colour_description_present_flag){
			vui->colour_primaries = get_bits(&s->gb, 8);
			vui->transfer_characteristics = get_bits(&s->gb, 8);
			vui->matrix_coefficients = get_bits(&s->gb, 8);
		}
	}

	vui->chroma_loc_info_present_flag = get_bits1(&s->gb);

	if(vui->chroma_loc_info_present_flag){
		vui->chroma_sample_loc_type_top_field = get_ue_golomb(&s->gb);
		vui->chroma_sample_loc_type_bottom_field = get_ue_golomb(&s->gb);
	}

	vui->timing_info_present_flag = get_bits1(&s->gb);

	if(vui->timing_info_present_flag){
		vui->num_units_in_tick = get_bits_long(&s->gb, 32);
		vui->time_scale = get_bits_long(&s->gb, 32);
		vui->fixed_frame_rate_flag = get_bits1(&s->gb);
	}

	vui->nal_hrd_parameters_present_flag = get_bits1(&s->gb);

	if(vui->nal_hrd_parameters_present_flag){
		decode_h264z_hrd(s, &vui->nal_hrd_parameters);
	}

	vui->vcl_hrd_parameters_present_flag = get_bits1(&s->gb);

	if(vui->vcl_hrd_parameters_present_flag){
		decode_h264z_hrd(s, &vui->vcl_hrd_parameters);
	}

	if(vui->nal_hrd_parameters_present_flag || vui->vcl_hrd_parameters_present_flag){
		vui->low_delay_hrd_flag = get_bits1(&s->gb);
	}

	vui->pic_struct_present_flag = get_bits1(&s->gb);

	vui->bitstream_restriction_flag = get_bits1(&s->gb);

	if(vui->bitstream_restriction_flag){
		vui->motion_vectors_over_pic_boundaries_flag = get_bits1(&s->gb);
		vui->max_bytes_per_pic_denom = get_ue_golomb(&s->gb);
		vui->max_bits_per_mb_denom = get_ue_golomb(&s->gb);
		vui->log2_max_mv_length_horizontal = get_ue_golomb(&s->gb);
		vui->log2_max_mv_length_vertical = get_ue_golomb(&s->gb);
		vui->num_reorder_frames = get_ue_golomb(&s->gb);
		vui->max_dec_frame_buffering = get_ue_golomb(&s->gb);
	}

    return 1;
}


static int
decode_seq_parameter_set(H264zEncContext *s)
{
    int profile_idc, level_idc;
    int sps_id;
    SPS *sps;

    profile_idc= get_bits(&s->gb, 8);
    get_bits1(&s->gb);   //constraint_set0_flag
    get_bits1(&s->gb);   //constraint_set1_flag
    get_bits1(&s->gb);   //constraint_set2_flag
    get_bits1(&s->gb);   //constraint_set3_flag
    get_bits(&s->gb, 4); // reserved
    level_idc= get_bits(&s->gb, 8);
    sps_id= get_ue_golomb(&s->gb);

    sps= &s->sps_buffer[ sps_id ];
    sps->profile_idc= profile_idc;
    sps->level_idc= level_idc;

    sps->scaling_matrix_present = 0;

    sps->log2_max_frame_num= get_ue_golomb(&s->gb) + 4;
    sps->poc_type= get_ue_golomb(&s->gb);

    if(sps->poc_type == 0)
    { //FIXME #define
        sps->log2_max_poc_lsb= get_ue_golomb(&s->gb) + 4;
    }

    if(sps->poc_type > 2)
    {
        printf("illegal POC type %d\n", sps->poc_type);
        return -1;
    }

    sps->ref_frame_count= get_ue_golomb(&s->gb);

    // For C10 - max value of ref_frame_count is 2 for B frames
    if(sps->ref_frame_count > 2){
        printf("too many reference frames\n");
    }
    sps->gaps_in_frame_num_allowed_flag= get_bits1(&s->gb);
    sps->mb_width= get_ue_golomb(&s->gb) + 1;
    sps->mb_height= get_ue_golomb(&s->gb) + 1;

    if ( sps->mb_width > H264_MAX_WIDTH/16 || sps->mb_height > H264_MAX_HEIGHT/16 )
    {
        printf ("Picture too big: %dx%d (MAX: %dx%d)\n", sps->mb_width*16, sps->mb_height*16, H264_MAX_WIDTH, H264_MAX_HEIGHT );
        return -1;
    }

    sps->frame_mbs_only_flag= get_bits1(&s->gb);
    if(!sps->frame_mbs_only_flag)
        sps->mb_aff= get_bits1(&s->gb);
    else
        sps->mb_aff= 0;

    sps->direct_8x8_inference_flag= get_bits1(&s->gb);

#ifndef ALLOW_INTERLACE
    if(sps->mb_aff)
        printf("MBAFF support not included; enable it compilation time\n");
#endif
    if(!sps->direct_8x8_inference_flag && sps->mb_aff)
        printf("MBAFF + !direct_8x8_inference is not implemented\n");

    sps->crop= get_bits1(&s->gb);
    if(sps->crop){

        sps->crop_left  = get_ue_golomb(&s->gb);
        sps->crop_right = get_ue_golomb(&s->gb);
        sps->crop_top   = get_ue_golomb(&s->gb);
        sps->crop_bottom= get_ue_golomb(&s->gb);
    }else{
        sps->crop_left  =
        sps->crop_right =
        sps->crop_top   =
        sps->crop_bottom= 0;
    }

    // Update crop_top & crop_left in AvContext
    s->avctx->crop_top = sps->crop_top * 2;
    s->avctx->crop_left = sps->crop_left * 2;

    sps->vui_parameters_present_flag= get_bits1(&s->gb);

	if(sps->vui_parameters_present_flag){
		decode_h264z_vui(s, &sps->vui_parameters);
	}

#ifdef H264Z_DEBUG_PRINT
    printf("sps:%d profile:%d/%d poc:%d ref:%d %dx%d %s %s crop:%d/%d/%d/%d %s\n",
           sps_id, sps->profile_idc, sps->level_idc,
           sps->poc_type,
           sps->ref_frame_count,
           sps->mb_width, sps->mb_height,
           sps->frame_mbs_only_flag ? "FRM" : (sps->mb_aff ? "MB-AFF" : "PIC-AFF"),
           sps->direct_8x8_inference_flag ? "8B8" : "",
           sps->crop_left, sps->crop_right,
           sps->crop_top, sps->crop_bottom,
           sps->s->sps.vui_parameters_present_flag ? "VUI" : "" );
#endif

    s->physical_width = sps->mb_width * 16;
    s->physical_height = sps->mb_height * 16;
    s->logical_width = s->physical_width - ( sps->crop_right + sps->crop_left );
    s->logical_height = s->physical_height - ( sps->crop_bottom + sps->crop_top );

    return 0;
}

/*

%func

  decode_picture_parameter_set

*/

static int
decode_picture_parameter_set(H264zEncContext *s, int bit_length)
{
    int pps_id= get_ue_golomb(&s->gb);
    PPS *pps= &s->pps_buffer[pps_id];

    pps->sps_id= get_ue_golomb(&s->gb);
    pps->cabac= get_bits1(&s->gb);
    s->entropy_mode = pps->cabac;

    pps->pic_order_present= get_bits1(&s->gb);
    pps->slice_group_count= get_ue_golomb(&s->gb) + 1;

    pps->ref_count[0]= get_ue_golomb(&s->gb) + 1;
    pps->ref_count[1]= get_ue_golomb(&s->gb) + 1;

    // In C10 we support only one forward reference frame and one backward reference frame
    if(pps->ref_count[0] > 1 || pps->ref_count[1] > 1)
    {
        printf("reference overflow (pps)\n");
        return -1;
    }

    pps->weighted_pred= get_bits1(&s->gb);
    pps->weighted_bipred_idc= get_bits(&s->gb, 2);
    pps->init_qp= get_se_golomb(&s->gb) + 26;
    pps->init_qs= get_se_golomb(&s->gb) + 26;
    pps->chroma_qp_index_offset= get_se_golomb(&s->gb);
    pps->deblocking_filter_parameters_present= get_bits1(&s->gb);
    pps->constrained_intra_pred= get_bits1(&s->gb);
    pps->redundant_pic_cnt_present = get_bits1(&s->gb);

    pps->transform_8x8_mode= 0;

#ifdef H264Z_DEBUG_PRINT
    printf("pps:%d sps:%d %s slice_groups:%d ref:%d/%d %s qp:%d/%d/%d %s %s %s %s\n",
           pps_id, pps->sps_id,
           pps->cabac ? "CABAC" : "CAVLC",
           pps->slice_group_count,
           pps->ref_count[0], pps->ref_count[1],
           pps->weighted_pred ? "weighted" : "",
           pps->init_qp, pps->init_qs, pps->chroma_qp_index_offset,
           pps->deblocking_filter_parameters_present ? "LPAR" : "",
           pps->constrained_intra_pred ? "CONSTR" : "",
           pps->redundant_pic_cnt_present ? "REDU" : "",
           pps->transform_8x8_mode ? "8x8DCT" : ""
           );
#endif

    return 0;
}



#if 0 /* unused */
/*
%func
  decode_SEI_message_default
*/

static int
decode_SEI_message_default(H264zEncContext *s, int payload_size)
{
	int i;

    for(i = 0; i < payload_size; i++){
		get_bits(&s->gb, 8);
	}

	return 1;
}
#endif /* 0 */

#if 0
/*
%func
  decode_picture_parameter_set
*/

static int
decode_nal_SEI_message(H264zEncContext *s)
{
	int payload_type = 0;
	int payload_size = 0;
	int offset = 1;
	uint8_t tmp_byte;

	do
	{
		// sei_message();
		payload_type = 0;

		tmp_byte = get_bits(&s->gb, 8);

		while (tmp_byte == 0xFF){
			payload_type += 255;
			tmp_byte = get_bits(&s->gb, 8);
		}

		payload_type += tmp_byte;   // this is the last byte

		payload_size = 0;
		tmp_byte = get_bits(&s->gb, 8);

		while (tmp_byte == 0xFF){
			payload_size += 255;
			tmp_byte = get_bits(&s->gb, 8);
		}

		payload_size += tmp_byte;   // this is the last byte

		switch ( payload_type )     // sei_payload( type, size );
		{
			case  0://SEI_BUFFERING_PERIOD:
			//	decode_SEI_buffering_period(s, payload_size);
			//	break;
			case  1://SEI_PIC_TIMING:
			//	decode_SEI_picture_timing(s, payload_size);
			//	break;
			case  6://SEI_RECOVERY_POINT:
			//	decode_SEI_recovery_point(s, payload_size);
			//	break;
			default:
				decode_SEI_message_default(s, payload_size);
			break;
		}

		offset += payload_size;

	} while( show_bits(&s->gb, 8) != 0x80 );    // more_rbsp_data()

    return 0;
}
#endif /*0 */

/*

%func

  get_pict_type_char

*/

char
get_pict_type_char(int pict_type)
{
    switch(pict_type)
    {
        case I_TYPE: return 'I';
        case P_TYPE: return 'P';
        case B_TYPE: return 'B';
        default:     return '?';
    }
}

/*

%func

  decode_ref_pic_list_reordering

*/

static int
decode_ref_pic_list_reordering(H264zEncContext *s)
{
	if ( s->slice_type != H264_SLICE_TYPE_I )
    {
        // ref_pic_list_reordering_flag_l0
        get_bits1(&s->gb);
    }

	if ( s->slice_type == H264_SLICE_TYPE_B )
    {
        // ref_pic_list_reordering_flag_l1
        get_bits1(&s->gb);
    }


    return 0;
}

/*

%func

  decode_ref_pic_marking

*/

static int
decode_ref_pic_marking(H264zEncContext *s)
{
    if(s->nal_unit_type == H264_NAL_IDR_SLICE)
    {
        // no_output_of_prior_pics_flags
        get_bits1(&s->gb);
        // long_term_reference_flag
        get_bits1(&s->gb);
    }
    else
    {
        // adaptive_ref_pic_marking_mode_flag
        get_bits1(&s->gb);
    }

    return 0;
}

/*

%func

  decode_nal_units

%desc

 decodes a slice header.
 this will allso call MPV_common_init() and frame_start() as needed

*/

static int
decode_slice_header(H264zEncContext *s)
{
    int first_mb_in_slice;
    int num_ref_idx_active_override_flag;

	int pps_id;

    //static const uint8_t slice_type_map[3]= {P_TYPE, B_TYPE, I_TYPE};
	static const uint8_t slice_type_map[]= {	H264_SLICE_TYPE_P, // 0
												H264_SLICE_TYPE_B, // 1
												H264_SLICE_TYPE_I, // 2
												H264_SLICE_TYPE_SP,// 3
												H264_SLICE_TYPE_SI,// 4
												H264_SLICE_TYPE_P, // 5
												H264_SLICE_TYPE_B, // 6
												H264_SLICE_TYPE_I, // 7
												H264_SLICE_TYPE_SP,// 8
												H264_SLICE_TYPE_SI};//9
	static const uint8_t pict_type_map[3]= {P_TYPE, B_TYPE, I_TYPE};
    int slice_type;
    int direct_spatial_mv_pred;

    first_mb_in_slice= get_ue_golomb(&s->gb);

    slice_type= get_ue_golomb(&s->gb);

	// check for correct range
    //if(slice_type > 2){
	if(slice_type > H264_SLICE_TYPE_SI){
        printf("slice type too large (%d)\n", slice_type );
        return -1;
    }

	slice_type = slice_type_map[ slice_type ];

	// check for SP/SI
	if(slice_type > H264_SLICE_TYPE_I){
        printf("slice type too large (%d)\n", slice_type );
        return -1;
    }

/*
    // TODO: Check this!
    if (slice_type == I_TYPE
        || (s->frame_num != 0 && slice_type == s->slice_type) ) {
        default_ref_list_done = 1;
    }
*/

    s->slice_type = slice_type;
    s->pict_type  = pict_type_map[slice_type - H264_SLICE_TYPE_P];

    pps_id= get_ue_golomb(&s->gb);
    if(pps_id>255){
        printf("pps_id out of range\n");
        return -1;
    }
    s->pps= s->pps_buffer[pps_id];
    if(s->pps.slice_group_count == 0){
        printf("non existing PPS referenced\n");
        return -1;
    }

    s->sps= s->sps_buffer[ s->pps.sps_id ];
    if(s->sps.log2_max_frame_num == 0){
        printf("non existing SPS referenced\n");
        return -1;
    }

    s->physical_width = 16*s->sps.mb_width - 2*(s->sps.crop_left + s->sps.crop_right );
    if(s->sps.frame_mbs_only_flag)
        s->physical_height= 16*s->sps.mb_height - 2*(s->sps.crop_top  + s->sps.crop_bottom);
    else
        s->physical_height= 16*s->sps.mb_height - 4*(s->sps.crop_top  + s->sps.crop_bottom); //FIXME recheck

    s->frame_num= get_bits(&s->gb, s->sps.log2_max_frame_num);

    if ( s->nal_unit_type == H264_NAL_IDR_SLICE )
    {
        // IDR_PIC_ID
        get_ue_golomb(&s->gb);
    }

    if ( s->sps.poc_type==0)
    {
        // PIC_ORDER_CNT_LSB
        s->pic_order_cnt = get_bits(&s->gb, s->sps.log2_max_poc_lsb);
    }

    if(s->slice_type == H264_SLICE_TYPE_P || s->slice_type == H264_SLICE_TYPE_B)
    {
        if(s->slice_type == H264_SLICE_TYPE_B)
        {
            direct_spatial_mv_pred= get_bits1(&s->gb);
            if(s->sps.mb_aff && direct_spatial_mv_pred)
                printf("MBAFF + spatial direct mode is not implemented\n");
        }
        num_ref_idx_active_override_flag = get_bits1(&s->gb);
    }

    if(decode_ref_pic_list_reordering(s) < 0)
    {
            return -1;
    }

    if ( s->nal_ref_idc != 0 )
    {
            decode_ref_pic_marking(s);
    }

    if( s->slice_type != H264_SLICE_TYPE_I && s->pps.cabac )
        s->cabac_init_idc = get_ue_golomb(&s->gb);

    // SLICE_QP_DELTA
    s->qscale = s->pps.init_qp + get_se_golomb(&s->gb);
    if(s->qscale<0 || s->qscale>51){
        printf("QP %d out of range\n", s->qscale);
        return -1;
    }

    // We calculate this for each MB
    s->chroma_qp = s->pps.chroma_qp_index_offset + s->qscale;

    s->deblocking_filter = 1;
    s->slice_alpha_c0_offset = 0;
    s->slice_beta_offset = 0;

    if( s->pps.deblocking_filter_parameters_present )
    {
        // DISABLE_DEBLOCKING_FILTER_IDC
        s->deblocking_filter= get_ue_golomb(&s->gb);

        if(s->deblocking_filter < 2)
        {
            s->deblocking_filter^= 1; // 1<->0
        }

        if( s->deblocking_filter )
        {
            s->slice_alpha_c0_offset = get_se_golomb(&s->gb) << 1;
            s->slice_beta_offset = get_se_golomb(&s->gb) << 1;
        }
    }

    s->frame_num++;



#ifdef H264Z_DEBUG_PRINT
printf("Debug Info: frame_type = %c, frame_num = %d, pic_order_cnt = %d\n",
       (s->slice_type == H264_SLICE_TYPE_B) ? 'B' :
       (s->slice_type == H264_SLICE_TYPE_P) ? 'P' : 'I',
       s->frame_num,
       s->pic_order_cnt);

        printf("slice:%d mb:%d %c pps:%d frame:%d qp:%d loop:%d:%d:%d\n",
               s->frame_num,
               first_mb_in_slice,
               get_pict_type_char(s->pict_type),
               pps_id, s->frame_num,
//               ref_count[0], ref_count[1],
               s->qscale,
               s->deblocking_filter, s->slice_alpha_c0_offset/2, s->slice_beta_offset/2
               );
#endif /* H264Z_DEBUG_PRINT */

    return 0;
}

/*

%func

  decode_nal_units

*/

static int
decode_nal_units(H264zEncContext *s, uint8_t *buf, int buf_size)
{
    int buf_index=0;
    int found_slice = 0;

    if ( (s->avctx->priv_data == NULL) || (buf == NULL) )
    {
        return -1;
    }

    for(;;)
    {
        int consumed = 0;
        int dst_length;
        int bit_length;
        uint8_t *ptr;

        if ( found_slice == 1 )
        {
            break;
        }

        // start code prefix search
        for(; buf_index + 3 < buf_size; buf_index++)
        {
            // this should allways succeed in the first iteration
            if(buf[buf_index] == 0 && buf[buf_index+1] == 0 && buf[buf_index+2] == 1)
                break;
        }

        if(buf_index+3 >= buf_size)
            break;

        buf_index+=3;

        // 1 byte for first byte (nal_unit_type)
        s->slice_header_offset = buf_index + 1;

        ptr= decode_nal(s, buf + buf_index, &dst_length, &consumed, buf_size - buf_index);

        if ( ptr == NULL )
        {
            return -1;
        }

        if(ptr[dst_length - 1] == 0) dst_length--;
        bit_length= 8*dst_length - decode_rbsp_trailing(ptr + dst_length - 1);

        buf_index += consumed;

        switch(s->nal_unit_type)
        {
            case H264_NAL_IDR_SLICE:
            {
//                idr(h); //FIXME ensure we don't loose some frames if there is reordering
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_IDR_SLICE\n");
#endif
            }
            case H264_NAL_SLICE:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_SLICE\n");
#endif
                init_get_bits(&s->gb, ptr, bit_length);

                if(decode_slice_header(s) < 0){
                    printf("decode_slice_header error\n");
                    break;
                }

                s->slice_header_bits = get_bits_count(&s->gb);

                // Senity check assumption that slice header cannot be bigger than 200 bytes.
                assert( ( s->slice_header_bits / 8 ) <= 200 );

                found_slice = 1;
                break;
            }
            case H264_NAL_SPS:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_SPS\n");
#endif
                init_get_bits(&s->gb, ptr, bit_length);
                decode_seq_parameter_set(s);

                if(s->flags& CODEC_FLAG_LOW_DELAY)
                {
                    s->low_delay=1;
                }   

                if(s->max_b_frames < 2)
                    s->max_b_frames= !s->low_delay;
                break;
            }
            case H264_NAL_PPS:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_PPS\n");
#endif
                init_get_bits(&s->gb, ptr, bit_length);
                decode_picture_parameter_set(s, bit_length);
                s->headers_read = 1;
                break;
            }
			case H264_NAL_SEI:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_SEI\n");
#endif
				// for now ignore SEI messages.
                break;
            }
			case H264_NAL_AUD:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_AUD\n");
#endif
				// ignore access unit delimiter.
                break;
            }
            case H264_NAL_END_SEQUENCE:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_END_SEQUENCE\n");
#endif
            }
            case H264_NAL_END_STREAM:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_END_STREAM\n");
#endif
                break;
            }
            default:
            {
                printf("Unknown NAL code: %d\n", s->nal_unit_type);
                break;
            }
        }
    }

    return buf_index;
}

/**

%func

  h264z_decode_init

*/
int
h264z_decode_init(AVCodecContext *avctx)
{
    // Get encoding context
    H264zEncContext *s = avctx->priv_data;

    s->avctx = avctx;

    s->headers_read = 0;

    s->entropy_mode = 0;

    s->physical_width  = avctx->physical_width;
    s->physical_height = avctx->physical_height;

    s->logical_width = avctx->logical_width;
    s->logical_height = avctx->logical_height;

    s->buffered_picture_ptr[0] = NULL;
    s->buffered_picture_ptr[1] = NULL;

    if ( h264z_common_init(s) < 0 )
    {
        return -1;
    }

    if ( s->flags & CODEC_FLAG_PIX_FORMAT_YUV420 )
    {
        avctx->pix_fmt= PIX_FMT_YUV420Z;
    }
    else
    {
        avctx->pix_fmt= PIX_FMT_YUV422Z;
    }

    //default, might be overriden in the vol header during header parsing
    s->low_delay = 0;
    s->codec_id  = avctx->codec->id;

    // Set buffer allocation functions.
    s->avctx->get_buffer     = decoder_get_buffer;
    s->avctx->release_buffer = decoder_release_buffer;

    // Initialize block buffers - assume worst case (with deblocking)
    // We need to initialize the YUV buffer size.
    // Accurate YUV buffer size will be set after first frame will
    // be decoded.
    MepGetBufferGeometry( AV_BLOCK_BUFFER_IMAGE,
                          s->physical_width,
                          s->physical_height,
                          s->flags,
                          TRUE,
                          TRUE,
                          &s->avctx->block_buf[AV_BLOCK_BUFFER_IMAGE] );

    MepGetBufferGeometry( AV_BLOCK_BUFFER_IMAGE_C,
                          s->physical_width,
                          s->physical_height,
                          s->flags,
                          TRUE,
                          TRUE,
                          &s->avctx->block_buf[AV_BLOCK_BUFFER_IMAGE_C] );

    // Decoder driver will be initialized at the first frame of the clip.
    // We can not initialize it now, because frame size and entropy mode,
    // are not known yet.
    s->decoder_init_done = 0;

    return 0;
}


/**

%func

  h264z_decode_end

%desc

*/
int
h264z_decode_end(AVCodecContext *avctx)
{
    H264zEncContext *s = avctx->priv_data;

    ParseContext *pc= &s->parse_context;

    h264z_common_end(s);

    av_free(pc->buffer);
    av_free(s->rbsp_buffer);

    if ( s->decoder_init_done == 1 )
    {
        UINT i;

        H264DecoderEnd( s );

        // Release the reconstructre/reference buffers
        for ( i = 0; i < H264_MAX_PICTURE_COUNT; i++ )
        {
            AVFrame* pic = &s->picture[i];

            decoder_release_external_buffer( pic );
        }
    }

    return 0;
}

/*

%func

  h264z_decode_frame_actual: the decoding process starts from this function

*/

int
h264z_decode_frame(AVCodecContext *avctx,
                    void *data,
                    int *data_size,
                    uint8_t *buf,
                    int buf_size)
{
    H264zEncContext *s = avctx->priv_data;
    int i;
    AVFrame *pict = data;
    int next = buf_size;

    *data_size = 0;

    s->flags = avctx->flags;

	avctx->last_decoded_frame_size = next;

    if ( buf_size == 0 )
    {
        assert( buf == NULL );
    }

    if ( buf != NULL )
    {
        if ( s->flags & CODEC_FLAG_TRUNCATED )
        {
            next = h264z_find_frame_end(s, buf, buf_size);

			avctx->last_decoded_frame_size = next;

            if( h264z_combine_frame(s, next, &buf, &buf_size) < 0 )
            {
                // Return number of consumed bytes (assuming the whole buffer)
                return buf_size;
            }
        }


        decode_nal_units( s, buf, buf_size );

        if ( s->flags & CODEC_FLAG_LOW_DELAY )
        {
            s->low_delay = 1;
        }

        // Update actual frame size.
        s->physical_width = avctx->physical_width;
        s->physical_height = avctx->physical_height;

        s->logical_width = avctx->logical_width;
        s->logical_height = avctx->logical_height;

        // In h264 we can not init the decoder at the "open stage" of the codec
        // since, the frame size and entropy mode are not known yet.
        if ( s->decoder_init_done == 0 )
        {
            // Set geometry for Ref Y and Ref C buffers.
            // Buffers geometry is set according to bandwidth optimization
            // options and deblocking filter option.
            for ( i = 0; i < AV_BLOCK_BUFFER_NUM; i++ )
            {
               MepGetBufferGeometry( i,
                                      s->physical_width,
                                      s->physical_height,
                                      s->flags,
                                      TRUE,
                                      (s->deblocking_filter == 1),
                                      &s->avctx->block_buf[i] );
            }

            H264DecoderInit( s );

            s->decoder_init_done = 1;
        }


        /* Skip b frames if we dont have reference frames */
        if ( (s->last_picture_ptr == NULL) && (s->pict_type == B_TYPE) )
        {
            printf( "Returning since s->last_picture_ptr==NULL && s->pict_type==B_TYPE\n" );
            // Return number of consumed bytes (assuming the whole buffer)
            return buf_size;
        }

        /* Skip b frames if we are in a hurry */
        if (avctx->hurry_up && s->pict_type==B_TYPE)
        {
            // Return number of consumed bytes (assuming the whole buffer)
            return buf_size;
        }

        /* skip everything if we are in a hurry>=5 */
        if (avctx->hurry_up>=5)
        {
            printf( "Returning since avctx->hurry_up>=5\n" );
            // Return number of consumed bytes (assuming the whole buffer)
            return buf_size;
        }
        if ( h264z_frame_start(s, avctx) < 0 )
        {
            printf( "h264z_frame_start failed, returning\n" );
            return -1;
        }


        // Update code buffer (for current and ready)
        s->current_picture_ptr->code_buf = buf;
        // Update flags, timestamp, duration and segment
        s->current_picture_ptr->flags = pict->flags;
        s->current_picture_ptr->timestamp = pict->timestamp;
        s->current_picture_ptr->duration = pict->duration;
        s->current_picture_ptr->segment = pict->segment;

        

        // Syncronize dram-cache consistency of code buffer.
        CACHE_WriteBackInvalidateDCacheRegion( (ULONG)buf, buf_size );
    }
    else
    {
	    // NULL code buffer indicates flush request, in which case the
    	// remaining reference image is returned

        s->current_picture_ptr = NULL;
        s->ready_picture_ptr = NULL;

        if ( s->decoder_init_done == 0 )
        {
            return 0; //
        }
    }

    avctx->error_code = H264DecodeFrame( s );



    // If we have a decoded frame
    if ( s->ready_picture_ptr != NULL )
    {
        if ( (s->ready_picture_ptr->pict_type == B_TYPE) || ( s->low_delay != 0 ) )
        {
            // If frame is B, we return the ready picture
            *pict= *s->ready_picture_ptr;
			*data_size = sizeof(AVFrame);
        }
        else
        {
            // If frame is not B, we return the last frame (buffered frame)
			if( s->buffered_picture_ptr[0] != NULL )
            {
                *pict= *s->buffered_picture_ptr[0];
				*data_size = sizeof(AVFrame);
			}
        }

        h264z_frame_end(s);

        if ( avctx->error_code != AVCODEC_DECODER_SUCCESS )
        {
            return -1;
        }

    }
    else
    {
        // No ready frame, but there is a buffered frame we need to send
        // in case of low_delay = 0 we always send the ready frame, never the buffered frame
        if( s->low_delay == 0 && s->buffered_picture_ptr[0] != NULL )
        {
            *pict= *s->buffered_picture_ptr[0];
            *data_size = sizeof(AVFrame);

            // We've just sent this picture, put NULL, so we won't send it again
            s->buffered_picture_ptr[0] = NULL;
        }
        else
        {
            // No ready frame and no buffered frame, send NULL
            pict = NULL;
            *data_size = 0;
        }

        if ( buf == NULL )
        {
            // No code was given, and no frame was returned
			// We should free the last two reference frames we saved
            if ( s->next_picture_ptr != NULL )
            {
                free_picture( s, s->next_picture_ptr );
            }

            if ( s->last_picture_ptr != NULL )
            {
                free_picture( s, s->last_picture_ptr );
            }

			s->next_picture_ptr = s->last_picture_ptr = s->buffered_picture_ptr[0] = NULL;
        }
    }

    // Return number of consumed bytes.
    // When combine frames "start code" might be divided between
    // two packes. In this case the "frame end index" (next) will
    // be negative. But the number of consumed bytes is stil 0.
    return MAX( 0, next);
}



AVCodec h264z_decoder =
{
    "h264z",
    CODEC_TYPE_VIDEO,
    CODEC_ID_H264Z,
    sizeof(H264zEncContext),
    h264z_decode_init,
    NULL,
    h264z_decode_end,
    h264z_decode_frame,
    CODEC_CAP_TRUNCATED,
};

