/*
 * Misc image convertion routines
 * Copyright (c) 2001, 2002, 2003 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file imgconvert.c
 * Misc image convertion routines.
 */

/* TODO:
 * - write 'ffimg' program to test all the image related stuff
 * - move all api to slice based system
 * - integrate deinterlacing, postprocessing and scaling in the conversion process
 */

#include "avcodec.h"
#include "Av.h"
/* Picture field are filled with 'ptr' addresses. Also return size */
int Av_avpicture_fill(AVPicture *picture, 
                      uint8_t *ptr,
                      int pix_fmt, 
                      int width, 
                      int height)
{
    switch(pix_fmt)
    {
        case PIX_FMT_YUV422Z:
        case PIX_FMT_RAW16Z:
        {
            picture->data[0] = ptr;
            picture->data[AV_BUF_YUV_IMAGE_C] = NULL;
            picture->data[2] = NULL;
            picture->linesize[0] = width * 2;
            return width * height * 2;
        }

        case PIX_FMT_YUV420P:
        {
            int i;

            uint8_t *chroma_start_ptr = ptr + width * height;
            uint8_t *src_ptr_u = chroma_start_ptr;
            uint8_t *src_ptr_v = chroma_start_ptr + width * height / 4;
            uint8_t *dst_ptr_uv = picture->data[AV_BUF_YUV_IMAGE_C];

            // Set luma pointer
            picture->data[AV_BUF_YUV_IMAGE] = ptr;

            // Chroma component is converted to Zoran format
            for (i = 0; i < width * height / 4; i ++ )
            {
                dst_ptr_uv[2*i] = src_ptr_u[i];
                dst_ptr_uv[2*i + 1] = src_ptr_v[i];
            }

            picture->data[2] = NULL;
            picture->linesize[AV_BUF_YUV_IMAGE] = width;
            picture->linesize[AV_BUF_YUV_IMAGE_C] = width;

            return width * height * 12 / 8;
        }

        default:
        {
            picture->data[0] = NULL;
            picture->data[1] = NULL;
            picture->data[2] = NULL;
            picture->data[3] = NULL;
            return -1;
        }
    }
}

int avpicture_get_size(int pix_fmt, int width, int height)
{
    AVPicture dummy_pict;
    return Av_avpicture_fill(&dummy_pict, NULL, pix_fmt, width, height);
}


