/*
 * default memory allocator for libavcodec
 * Copyright (c) 2002 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "avcodec.h"
#include "zav.h"

/* here we can use OS dependant allocation functions */
#undef malloc
#undef free
#undef realloc

/* memory alloc */
void *av_malloc_ex(unsigned int size, char* file, int line)
{
    // :ZZZ: Should be fixed in the component
    if ( size == 0 )
    {
        size = 1;
    }
    return PoolAllocEx( NULL, size, POOL_MEMORY_CACHED | POOL_WAIT_FOREVER | POOL_DISABLE_FAILURE_PRINT, file, line );
}

void *av_realloc_ex(void* ptr, unsigned int size, char* file, int line)
{
    return PoolReallocEx( NULL, ptr, size, POOL_MEMORY_CACHED, file, line );
}

void av_free_ex(void *ptr, char* file, int line)
{
    PoolFreeEx( NULL, ptr, file, line );
}

void av_mem_lock(void *ptr)
{
    PoolGet( NULL, ptr );
}
