#ifndef MOV_H
#define MOV_H

#include "avformat.h"
#include "avi.h"
#include "avio.h"
#include "Features.h"

#define MKTAG(a,b,c,d) (a | (b << 8) | (c << 16) | (d << 24))
#define MKBETAG(a,b,c,d) (d | (c << 8) | (b << 16) | (a << 24))

//////////////////////////////////////////////////////////////////////////////
// Definitions and Structures for MOV Encoder 
//////////////////////////////////////////////////////////////////////////////
#define MOV_INDEX_CLUSTER_SIZE 16384
#define globalTimescale 600
#define MP4_GLOBAL_TIME_SCALE 90000

/* Indicates that debug information (current change-list number) will be added 
   into reserved area of MOV/MP4 container. Please comment this definition out 
   for final [GM] version of CW. */
//#define MOV_INCLUDE_DEBUG_INFO 

// MOV timestamp (Mac Os format) definitions
#define MOV_TIMESTAMP_GMTIME_REL_YEARS          70
#define MOV_TIMESTAMP_MAC_FORMAT_BASE_YEAR      1904
#define MOV_TIMESTAMP_DAYS_IN_NON_BISSEXT_YEAR  365
#define MOV_TIMESTAMP_SEC_IN_MIN                60
#define MOV_TIMESTAMP_MIN_IN_HOUR               60
#define MOV_TIMESTAMP_HOURS_IN_DAY              24
#define MOV_TIMESTAMP_SEC_IN_HOUR \
            (MOV_TIMESTAMP_SEC_IN_MIN*MOV_TIMESTAMP_MIN_IN_HOUR)
#define MOV_TIMESTAMP_SEC_IN_DAY  \
            (MOV_TIMESTAMP_HOURS_IN_DAY*MOV_TIMESTAMP_SEC_IN_HOUR)

// MOV Tag max modification
#define MOV_ATOM_MAX_CONFIG_NUM                 4

// MOV Tag configuration type
#define MOV_ATOM_STREAM_VIDEO                   0x00000001
#define MOV_ATOM_STREAM_AUDIO                   0x00000002
#define MOV_ATOM_GENERAL                        0x00000003

// MOV Atom configuration
#define MOV_ATOM_DEL                            0x00000001
#define MOV_ATOM_INITIAL                        0x00000002
#define MOV_ATOM_REGISTER_CALLBACK              0x00000004

// MOV Tag Id
#define MOV_ATOM_ID_STSS                        0x00000001
#define MOV_ATOM_ID_SKIP                        0x00000002
#define MOV_ATOM_ID_CUSTOM                      0x00000003

typedef enum
{
    MOV_HEADER_NO_INDEX = 0,
    MOV_HEADER_INDEX_FIRST_ENTRY_ONLY,
    MOV_HEADER_FULL_INDEX
} MOV_HEADER_INDEX_FLAG;

#if( FEATURE_MOVI_ENTRY_LOW_MEM() == FEATURE_EN )

typedef struct MOVIentry {
    unsigned int   pos;
    unsigned int   size;    
    unsigned short duration;

    unsigned short samplesInChunk:8;
    unsigned short dependency: 7;
    unsigned short key_frame :1 ;
} MOVIentry;

#else

typedef struct MOVIentry {
    offset_t pos;
    unsigned int samplesInChunk;
    unsigned int size;    
    unsigned int duration;
    
    unsigned short dependency: 15;
    unsigned short key_frame :1 ;
} MOVIentry;

#endif

typedef struct MOVTrack { 
    int         entry;
    uint64_t    mdat_size;
    int         ents_allocated;
    long        timescale;
    ULONG       time;
    uint64_t    trackDuration;
    long        sampleCount;
    long        sampleDuration;
    int         hasKeyframes;
    int         trackID;
    int         chunksCount;
    unsigned int maxSamplesPerChunk;
    unsigned int SamplesInChunk;
    AVCodecContext *enc;

    int         vosLen;
    uint8_t     *vosData;
    MOVIentry** cluster;
} MOVTrack;

typedef struct MOVEncodeContext{
    int64_t   time;
    int       nb_streams;
    int       mdat_written;
    offset_t  mdat_pos;
    offset_t  movi_list;
    long      timescale;
    MOVTrack  tracks[MAX_STREAMS];
    BOOL      bThumbSpaceCreated;
    offset_t  skip_atom_offset;
    offset_t  custom_atom_offset;
    ULONG     ulEncodingFlags;
    BOOL      bSkipAtomSpaceCreated;
    BOOL      bCustomAtomSpaceCreated;
} MOVEncodeContext;

typedef struct {
    UINT   tag_id;
    UINT   stream;
    UINT   op;
    UINT   size;
    PVOID  data;
} MOVAtomContent;

//Type definition for skip atom callback function 
typedef BOOL (*FMOVSKIPATOMCB)( AVMOVIEINFO *, unsigned char * data,  unsigned long size );

//////////////////////////////////////////////////////////////////////////////
// Definitions and Structures for MOV Decoder
//////////////////////////////////////////////////////////////////////////////
/* allows chunk splitting - should work now... */
#define MOV_SPLIT_CHUNKS

/* some streams in QT (and in MP4 mostly) aren't either video nor audio */
/* so we first list them as this, then clean up the list of streams we give back, */
/* getting rid of these */
#define CODEC_TYPE_MOV_OTHER    (enum CodecType) 3

static const CodecTag mov_video_tags[] = {
/*  { CODEC_ID_, MKTAG('c', 'v', 'i', 'd') }, *//* Cinepak */
/*  { CODEC_ID_H263, MKTAG('r', 'a', 'w', ' ') }, *//* Uncompressed RGB */
/*  { CODEC_ID_H263, MKTAG('Y', 'u', 'v', '2') }, *//* Uncompressed YUV422 */
/*    { CODEC_ID_RAWVIDEO, MKTAG('A', 'V', 'U', 'I') }, *//* YUV with alpha-channel (AVID Uncompressed) */
/* Graphics */
/* Animation */
/* Apple video */
/* Kodak Photo CD */
    { CODEC_ID_MJPEG, MKTAG('j', 'p', 'e', 'g') }, /* PhotoJPEG */
    { CODEC_ID_MPEG1VIDEO, MKTAG('m', 'p', 'e', 'g') }, /* MPEG */
    { CODEC_ID_MJPEG, MKTAG('m', 'j', 'p', 'a') }, /* Motion-JPEG (format A) */
    { CODEC_ID_MJPEG, MKTAG('m', 'j', 'p', 'b') }, /* Motion-JPEG (format B) */
    { CODEC_ID_MJPEG, MKTAG('A', 'V', 'D', 'J') }, /* MJPEG with alpha-channel (AVID JFIF meridien compressed) */
/*    { CODEC_ID_MJPEG, MKTAG('A', 'V', 'R', 'n') }, *//* MJPEG with alpha-channel (AVID ABVB/Truevision NuVista) */
/*    { CODEC_ID_GIF, MKTAG('g', 'i', 'f', ' ') }, *//* embedded gif files as frames (usually one "click to play movie" frame) */
/* Sorenson video */
    { CODEC_ID_SVQ1, MKTAG('S', 'V', 'Q', '1') }, /* Sorenson Video v1 */
    { CODEC_ID_SVQ1, MKTAG('s', 'v', 'q', '1') }, /* Sorenson Video v1 */
    { CODEC_ID_SVQ1, MKTAG('s', 'v', 'q', 'i') }, /* Sorenson Video v1 (from QT specs)*/
    { CODEC_ID_SVQ3, MKTAG('S', 'V', 'Q', '3') }, /* Sorenson Video v3 */
    { CODEC_ID_MPEG4Z, MKTAG('m', 'p', '4', 'v') },
    { CODEC_ID_MPEG4, MKTAG('D', 'I', 'V', 'X') }, /* OpenDiVX *//* sample files at http://heroinewarrior.com/xmovie.php3 use this tag */
/*    { CODEC_ID_, MKTAG('I', 'V', '5', '0') }, *//* Indeo 5.0 */
    { CODEC_ID_H263, MKTAG('h', '2', '6', '3') }, /* H263 */
    { CODEC_ID_H263, MKTAG('s', '2', '6', '3') }, /* H263 ?? works */
    { CODEC_ID_DVVIDEO, MKTAG('d', 'v', 'c', ' ') }, /* DV NTSC */
    { CODEC_ID_DVVIDEO, MKTAG('d', 'v', 'c', 'p') }, /* DV PAL */
/*    { CODEC_ID_DVVIDEO, MKTAG('A', 'V', 'd', 'v') }, *//* AVID dv */
    { CODEC_ID_VP3, MKTAG('V', 'P', '3', '1') }, /* On2 VP3 */
    { CODEC_ID_H264Z, MKTAG('a', 'v', 'c', '1') }, /* Advanced Video Coding 1 (H.264) */
    { CODEC_ID_NONE, 0 },
};

static const CodecTag mov_audio_tags[] = {
/*    { CODEC_ID_PCM_S16BE, MKTAG('N', 'O', 'N', 'E') }, *//* uncompressed */
    { CODEC_ID_PCM_S16BE, MKTAG('t', 'w', 'o', 's') },  /* signed 8/16 bits big-endian  */
    { CODEC_ID_PCM_U8, MKTAG('r', 'a', 'w', ' ') },     /* unsigned 8 bits(signed 16 bits big-endian */
    { CODEC_ID_PCM_S16LE, MKTAG('s', 'o', 'w', 't') },  /* signed 16 bits little-endian */
    { CODEC_ID_PCM_MULAW, MKTAG('u', 'l', 'a', 'w') }, /*  */
    { CODEC_ID_PCM_ALAW, MKTAG('a', 'l', 'a', 'w') }, /*  */
    { CODEC_ID_ADPCM_IMA_QT, MKTAG('i', 'm', 'a', '4') }, /* IMA-4 ADPCM */
    { CODEC_ID_ADPCM_MS, MKTAG('m', 's', 0, 2) }, /* MS ADPCM */
    { CODEC_ID_ADPCM_IMA_WAV, MKTAG('m', 's', 0, 0x11) }, /* DVI ADPCM */
    { CODEC_ID_MACE3, MKTAG('M', 'A', 'C', '3') }, /* Macintosh Audio Compression and Expansion 3:1 */
    { CODEC_ID_MACE6, MKTAG('M', 'A', 'C', '6') }, /* Macintosh Audio Compression and Expansion 6:1 */

    { CODEC_ID_MP2, MKTAG('.', 'm', 'p', '3') }, /* MPEG layer 3 */ /* sample files at http://www.3ivx.com/showcase.html use this tag */
    { CODEC_ID_MP2, 0x6D730055 }, /* MPEG layer 3 */
    { CODEC_ID_MP2, 0x5500736D }, /* MPEG layer 3 *//* XXX: check endianness */
/*    { CODEC_ID_OGG_VORBIS, MKTAG('O', 'g', 'g', 'S') }, *//* sample files at http://heroinewarrior.com/xmovie.php3 use this tag */
/* MP4 tags */
    { CODEC_ID_AAC, MKTAG('m', 'p', '4', 'a') }, /* MPEG 4 AAC or audio ? */
    /* The standard for mpeg4 audio is still not normalised AFAIK anyway */
    { CODEC_ID_MP2, MKTAG(' ', ' ', ' ', ' ') }, /* AMR-NB 3gp */
    { CODEC_ID_NONE, 0 },
};

/* the QuickTime file format is quite convoluted...
 * it has lots of index tables, each indexing something in another one...
 * Here we just use what is needed to read the chunks
 */

typedef struct MOV_sample_to_chunk_tbl {
    long first;
    long count;
    long id;
} MOV_sample_to_chunk_tbl;

typedef struct {
    uint32_t type;
    int64_t offset;
    int64_t size; /* total size (excluding the size and type fields) */
} MOV_atom_t;

typedef struct {
    int seed;
    int flags;
    int size;
    void* clrs;
} MOV_ctab_t;

typedef struct {
    int sample_count;
    uint32_t sample_offset;
} MOV_ctts_tbl;

typedef struct {
    uint8_t  version;
    uint32_t flags; // 24bit

    /* 0x03 ESDescrTag */
    uint16_t es_id;
#define MP4ODescrTag            0x01
#define MP4IODescrTag           0x02
#define MP4ESDescrTag           0x03
#define MP4DecConfigDescrTag        0x04
#define MP4DecSpecificDescrTag      0x05
#define MP4SLConfigDescrTag     0x06
#define MP4ContentIdDescrTag        0x07
#define MP4SupplContentIdDescrTag   0x08
#define MP4IPIPtrDescrTag       0x09
#define MP4IPMPPtrDescrTag      0x0A
#define MP4IPMPDescrTag         0x0B
#define MP4RegistrationDescrTag     0x0D
#define MP4ESIDIncDescrTag      0x0E
#define MP4ESIDRefDescrTag      0x0F
#define MP4FileIODescrTag       0x10
#define MP4FileODescrTag        0x11
#define MP4ExtProfileLevelDescrTag  0x13
#define MP4ExtDescrTagsStart        0x80
#define MP4ExtDescrTagsEnd      0xFE
    uint8_t  stream_priority;

    /* 0x04 DecConfigDescrTag */
    uint8_t  object_type_id;
    uint8_t  stream_type;
    /* XXX: really streamType is
     * only 6bit, followed by:
     * 1bit  upStream
     * 1bit  reserved
     */
    uint32_t buffer_size_db; // 24
    uint32_t max_bitrate;
    uint32_t avg_bitrate;

    /* 0x05 DecSpecificDescrTag */
    uint8_t  decoder_cfg_len;
    uint8_t *decoder_cfg;

    /* 0x06 SLConfigDescrTag */
    uint8_t  sl_config_len;
    uint8_t *sl_config;
} MOV_esds_t;

struct MOVParseTableEntry;

typedef struct MOVStreamContext {
    int ffindex; /* the ffmpeg stream id */
    int is_ff_stream; /* Is this stream presented to ffmpeg ? i.e. is this an audio or video stream ? */
    long next_chunk;
    long chunk_count;
    offset_t *chunk_offsets;  // The maximal file size is above 4GB
//    uint32_t *chunk_offsets;
    long sample_to_chunk_sz;
    MOV_sample_to_chunk_tbl *sample_to_chunk;
    long sample_to_chunk_index;
    long sample_size;
    long sample_count;
    long *sample_sizes;
    int time_scale;
    long current_sample;
    long left_in_chunk; /* how many samples before next chunk */
    /* specific MPEG4 header which is added at the beginning of the stream */
    int header_len;
    uint8_t *header_data;
    MOV_esds_t esds;
    int frame_rate;
    long *sync_samples;
    long sync_samples_count;
    long sample_duration_num; // number of entries
    long *sample_duration;    // duration per entry
    MOV_ctts_tbl *ctts_table;
    long ctts_entry_num;  
} MOVStreamContext;

typedef struct MOVDecodeContext {
    int mp4; /* set to 1 as soon as we are sure that the file is an .mp4 file (even some header parsing depends on this) */
    AVFormatContext *fc;
    int time_scale;
    int duration; /* duration of the longest track */
    int video_bitrate; /* average video bitrate retreaved from uuid prof*/
    int found_moov; /* when both 'moov' and 'mdat' sections has been found */
    int found_mdat; /* we suppose we have enough data to read the file */
    uint64_t mdat_size;
    offset_t mdat_offset;
    int total_streams;
    /* some streams listed here aren't presented to the ffmpeg API, since they aren't either video nor audio
     * but we need the info to be able to skip data from those streams in the 'mdat' section
     */
    MOVStreamContext *streams[MAX_STREAMS];

    offset_t next_chunk_offset;
    MOVStreamContext *partial; /* != 0 : there is still to read in the current chunk */
    int partial_stream_num;
    int ctab_size;
    MOV_ctab_t **ctab;           /* color tables */
    const struct MOVParseTableEntry *parse_table; /* could be eventually used to change the table */
    MOV_HEADER_INDEX_FLAG eLoadIndex;
    int ctts_entries; /* There is no entry entity for sdtp so need to share */
    unsigned char *dependency;
    /* NOTE: for recursion save to/ restore from local variable! */
} MOVDecodeContext;

/* XXX: it's the first time I make a recursive parser I think... sorry if it's ugly :P */

/* those functions parse an atom */
/* return code:
 1: found what I wanted, exit
 0: continue to parse next atom
 -1: error occured, exit
 */
typedef int (*mov_parse_function)(MOVDecodeContext *ctx, ByteIOContext *pb, MOV_atom_t atom);

/* links atom IDs to parse functions */
typedef struct MOVParseTableEntry {
    uint32_t type;
    mov_parse_function func;
} MOVParseTableEntry;

typedef enum
{
    TFHD_FLAG_BASE_DATA_OFFSET          = 0x0001,
    TFHD_FLAG_SAMPLE_DESCRIPTION_INDEX  = 0x0002,
    TFHD_FLAG_DEFAULT_SAMPLE_DURATION   = 0x0008,
    TFHD_FLAG_DEFAULT_SAMPLE_SIZE       = 0x0010,
    TFHD_FLAG_DURATION_IS_EMPTY         = 0x0100
} MOV_TFHD_ATOM_FLAG;

typedef enum
{
    TRUN_FLAG_DATA_OFFSET                       = 0x01,
    TRUN_FLAG_FIRST_SAMPLE_FLAGS                = 0x04,
    TRUN_FLAG_SAMPLE_DURATION                   = 0x08,
    TRUN_FLAG_SAMPLE_SIZE                       = 0x10,
    TRUN_FLAG_SAMPLE_FLAGS                      = 0x20,
    TRUN_FLAG_SAMPLE_COMPOSITION_TIME_OFFSETS   = 0x40
} MOV_TRUN_ATOM_FLAG;

typedef struct 
{
    unsigned short  flags;
    unsigned int    track_id;
    long long       base_data_offset;
    unsigned int    sample_description_index;
    unsigned int    default_sample_duration;
    unsigned int    default_sample_size;
    unsigned int    default_sample_flags;
} MOV_tfhd_t;

typedef struct 
{
    unsigned int sample_duration;
    unsigned int sample_size;
    unsigned int sample_flags;
    unsigned int sample_composition_time_offset;    
} MOV_trun_sample_t;

typedef struct
{
    unsigned short      flags;
    unsigned int        sample_count;
    int                 data_offset;
    unsigned int        first_sample_flags;
    MOV_trun_sample_t*  samples;
} MOV_trun_t;

#define MOV_ATOM_DEFAULT                0x0000
#define MOV_ATOM_INCLUDE_EXTENDED_FTYP  0x0001
#define MOV_ATOM_INCLUDE_UUID_PROF      0x0002
#define MOV_ATOM_INCLUDE_EDTS           0x0004
#define MOV_ATOM_REMOVE_HDLR_ALIS       0x0008
#define MOV_ATOM_INCLUDE_DREF_URL       0x0010
#define MOV_ATOM_INCLUDE_UUID_USMT_MTDT 0x0020
#define MOV_ATOM_INCLUDE_UUID_USMT_MTDF 0x0040
#define MOV_ATOM_INCLUDE_IODS           0x0080
#define MOV_ATOM_MOVE_STSS_TO_STBL_END  0x0100
#define MOV_ATOM_INCLUDE_FREE           0x0200
#define MOV_ATOM_INCLUDE_TREF_SYNC      0x0400
#define MOV_ATOM_INCLUDE_TREF_MPOD      0x0800
#define MOV_ATOM_INCLUDE_MVEX           0x1000
#define MOV_ATOM_INCLUDE_MOOF           0x2000
#define MOV_ATOM_REMOVE_EX_SIZE         0x4000

void mov_set_globalTimescale(long Timescale);

#endif // #ifndef MOV_H

