/*
 * AVI decoder.
 * Copyright (c) 2001 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "av.h"
#include "avformat.h"
#include "AvApi.h"

//#define DEBUG


////////////////////
// Exif - DECODER //
////////////////////

static int exif_read_header( AVFormatContext *s, AVFormatParameters *ap )
{
    AVStream* st;

    s->nb_streams = 1;

    st = avf_mallocz( sizeof( AVStream ) );
    if ( !st )
    {
        return -1;
    }

    AvApi_avcodec_get_context_defaults( &st->codec );

    st->codec.codec_type = CODEC_TYPE_VIDEO;

    st->codec.codec_id = CODEC_ID_JPEGZ;

    st->codec.extradata_size = 0;

    st->codec.extradata= NULL;

    s->streams[0] = st;

    s->packet_buffer = NULL;

    return 0;
}


static int exif_read_packet( AVFormatContext *s, AVPacket *pkt )
{
    ByteIOContext *pb = &s->pb;
    unsigned long file_size;
    offset_t current;
    int read;

    current = url_ftell( pb );

    file_size = Avf_url_fseek( pb, 0xFFFFFFF, SEEK_SET );

    Avf_url_fseek( pb, current, SEEK_SET );

    av_new_packet( pkt, file_size );

    read = get_buffer( pb, pkt->data, pkt->size );

    if ( read != pkt->size )
    {
        Avf_free_packet( pkt );
        return -1;
    }
    
    // Every packet is a key frame
    pkt->flags |= PKT_FLAG_KEY;

    return 0;
}


#if 0 /* unused */
static int exif_read_trailer( AVFormatContext *s )
{
    return 0;
}
#endif /* 0 */

static int exif_probe( AVProbeData *pd )
{
    if (pd->buf_size >= 64 &&
        pd->buf[0] == 0xff && pd->buf[1] == 0xd8 && pd->buf[2] == 0xff)
    {
        return AVPROBE_SCORE_MAX;
    }
    else
        return 0;
}


static int exif_read_close( AVFormatContext *s )
{
    return 0;
}


static int exif_read_seek( AVFormatContext *s, int64_t pts )
{
    return 0;
}


AVInputFormat exif_decode = {
    "exif",
    "Exif format",
    0,
    exif_probe,
    exif_read_header,
    exif_read_packet,
    exif_read_close,
    exif_read_seek,
};

int Avf_exif_init(void)
{
    av_register_input_format( &exif_decode );
    return 0;
}


