/*
 * MOV, 3GP, MP4 encoder.
 * Copyright (c) 2003 Thomas Raivio.
 * Enhancements by Gildas Bazin <gbazin@netcourrier.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "av.h"
#include "mov.h"
#include "AvOsStub.h"

#undef NDEBUG
#include <assert.h>

#ifdef ZAVF
#include "Rtc.h"
#endif

#include "../Av/Include/golomb.h" // storing information about H.264 frame dependancy

//#define MOV_DEBUG

#define MOV_MAX_CHUNK_SIZE 1024

#ifndef UINT32_MAX
#define UINT32_MAX      0xffffffff
#endif

static MOVAtomContent atom[MOV_ATOM_MAX_CONFIG_NUM] = {{0}};
static FMOVSKIPATOMCB fnMovSkipAtomCb  = NULL;
static FMOVSKIPATOMCB fnMovCustomAtomCb  = NULL;

static int mov_write_esds_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s);
static int mov_write_hdlr_alis_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s);
static int mov_write_audio_mute_end(ByteIOContext *pb, MOVEncodeContext *mov, int size, int data);
static int mov_query_audio_sample_pos(MOVTrack* track, int sampleSize, int* chunkNo, int* byte);
static int mov_seek_chunk_pos(ByteIOContext *pb, MOVIentry* enc, int byte);
static int mov_seek_last_chunk_end_pos(AVFormatContext *s);
static int mov_estimate_trailer_size(AVFormatContext *s);
static offset_t mov_get_last_chunk_end( AVFormatContext *s, offset_t* mdat_pos );

static int mov_write_uuid_prof_tag(ByteIOContext *pb, AVFormatContext *s);
static int mov_write_edts_tag(ByteIOContext *pb, MOVTrack *track, AVFormatContext *s);
static int mov_write_uuid_usmt(ByteIOContext *pb, MOVEncodeContext *mov);
static int mov_write_tref_tag2(ByteIOContext *pb, MOVEncodeContext *mov, int syncID, int odID);
static int mov_write_nmhd_tag(ByteIOContext *pb);
static int mov_write_mvex_tag(ByteIOContext *pb);
static int mov_write_moof_tag(ByteIOContext *pb, MOV_tfhd_t tfhd_data, MOV_trun_t trun_data);

// For mov to config the time scale
static long gl_globalTimescale = globalTimescale;
void mov_set_globalTimescale(long Timescale)
{
    gl_globalTimescale = Timescale;
}

//FIXME A stub for Sony framerate request
static unsigned int s_vprf_frame_rate_value = 0;

unsigned int mov_set_vprf_frame_rate(unsigned int frame_rate_value)
{
    s_vprf_frame_rate_value = frame_rate_value;

    return s_vprf_frame_rate_value ;
}

static unsigned int mov_get_vprf_frame_rate(AVCodecContext *pCodec)
{
    unsigned int frameRate = 0;

    if(pCodec == NULL)
    {
        return s_vprf_frame_rate_value;
    }

    if(pCodec->codec_type != CODEC_TYPE_VIDEO)
    {
        return s_vprf_frame_rate_value;
    }

    if (s_vprf_frame_rate_value == 0)
    {
        //if ( strcmp("mp4", s->oformat->name) == 0 )
        //{
            //((videoCodec.time_base.den) * (0x10000))/ (VideoCodec.time_base.num);
            frameRate = pCodec->frame_rate * 0x10000 / pCodec->frame_rate_base;
        //}
        //else
        //{
        //	frameRate = pVideoCodec->frame_rate;
        //}

        if ( frameRate == 0x001DF851 )
        {
            frameRate = 0x001DF853;//(29.97003 fps :(~ 30000/1001))
        }
    }
    else
    {
        frameRate = s_vprf_frame_rate_value;
    }

    return frameRate;
}

//FIXME supprt 64bit varaint with wide placeholders
static offset_t updateSize (ByteIOContext *pb, offset_t pos)
{
    offset_t curpos = url_ftell(pb);
    Avf_url_fseek(pb, pos, SEEK_SET);
    put_be32(pb, curpos - pos); /* rewrite size */
    Avf_url_fseek(pb, curpos, SEEK_SET);

    return curpos - pos;
}

static int updateValue (ByteIOContext *pb, offset_t pos, int value)
{
    offset_t curpos = url_ftell(pb);
    Avf_url_fseek(pb, pos, SEEK_SET);
    put_be32(pb, value); /* rewrite size */
    Avf_url_fseek(pb, curpos, SEEK_SET);

    return 0;
}

/* Preview atom */
static int mov_write_pnot_tag(ByteIOContext *pb, MOVEncodeContext* mov)
{
    offset_t pos = url_ftell(pb);
    put_be32(pb, 0);            /* size */
    put_tag(pb, "pnot");
    put_be32(pb, mov->time);    /* modification date */
    put_be16(pb, 0);            /* version number */
    put_tag(pb, "PICT");        /* atom type */
    put_be16(pb, 1);            /* atom index */

    return updateSize (pb, pos);
}

/* PICT atom */
static int mov_write_pict_tag(ByteIOContext *pb, AVFormatContext *s)
{
    char* str;
    offset_t pos = url_ftell(pb);
    offset_t pos1, pos2, curpos;

    put_be32(pb, 0);            /* size */
    put_tag(pb, "PICT");

    pos1 = url_ftell(pb);

    put_be16(pb, 0);            /* size */
    put_be32(pb, 0x00000000); put_be32(pb, 0x007800A0); /* frame size */
    put_be32(pb, 0x001102FF);   /* version */
    put_be32(pb, 0x0C00FFFE);   /* version2 */
    put_be16(pb, 0);            /* reserverd */
    put_be32(pb, 0x00480000); put_be32(pb, 0x00480000); /* resolution */
    put_be32(pb, 0x00000000); put_be32(pb, 0x007800A0); /* rectangle */
    put_be32(pb, 0x00000000);   /* reserved */
    put_be32(pb, 0x001F0080); put_be32(pb, 0x00800080); /* RGB color */
    put_be16(pb, 0x1E);         /* highlight */
    put_be32(pb, 0x0001000A); put_be32(pb, 0x00000000); put_be32(pb, 0x007800A0);  /*clip area*/
    put_be16(pb, 0x8200);       /* pict */

    pos2 = url_ftell(pb);

    put_be32(pb, 0);            /* size */
    put_be16(pb, 0);            /* version */

    // put matrix
    put_be32(pb, 0x00010000); put_be32(pb, 0x00000000); put_be32(pb, 0x00000000);
    put_be32(pb, 0x00000000); put_be32(pb, 0x00010000); put_be32(pb, 0x00000000);
    put_be32(pb, 0x00000000); put_be32(pb, 0x00000000); put_be32(pb, 0x40000000);

    put_be32(pb, 0);            /* mat size */
    put_be32(pb, 0); put_be32(pb, 0);   /* mat rect */
    put_be16(pb, 0x0040);       /* mode */
    put_be32(pb, 0x00000000); put_be32(pb, 0x007800A0); /* rectangle */
    put_be32(pb, 0x00000300);   /* accuracy */
    put_be32(pb, 0);            /* mask size */
    put_be32(pb, 0x56);         /* in size */
    put_tag(pb, "jpeg");
    put_be32(pb, 0); put_be16(pb, 0);   /* reserved */
    put_be16(pb, 0);            /* dataRefIndex */
    put_le16(pb, s->uwModel);   /* version */
    put_le16(pb, s->uwRevision);        /* revision */
    put_le32(pb, s->ulVendorName);      /* a four character vendor id */
    put_be32(pb, 0);            /*temporal quality*/
    put_be32(pb, 0x00000200);   /*spatial quality*/
    put_be16(pb, 0xA0);         /* width */
    put_be16(pb, 0x78);         /* height */
    put_be32(pb, 0x00480000);   /* hRes */
    put_be32(pb, 0x00480000);   /* vRes */
    put_be32(pb, 0);            /* datasize (unknown) */
    put_be16(pb, 1);            /* frame count */

    /*name*/
    str = "Photo - JPEG";
    put_pascal_str( pb, str, 32 );

    put_be16(pb, 0x18);         /* depth */
    put_be16(pb, 0xFFFF);       /* depth */

    // Put the thumbnail image
    put_buffer(pb, s->pbyThumbnailImage, s->ulThumbnailImageSize);

    // Align
    while (url_ftell(pb) % 4 != 2)
        put_byte(pb, 0x00);

    // Update pos2 size
    curpos = url_ftell(pb);
    Avf_url_fseek(pb, pos2, SEEK_SET);
    put_be32(pb, curpos - pos2 - 4); /* rewrite size */
    Avf_url_fseek(pb, curpos, SEEK_SET); /* return to current offset */

    put_be32(pb, 0x00000000); put_be16(pb, 0x00FF); /* endpic */

    curpos = url_ftell(pb);
    Avf_url_fseek(pb, pos1, SEEK_SET);
    put_be16(pb, curpos - pos1);    /* rewrite size */
    Avf_url_fseek(pb, curpos, SEEK_SET);

    return updateSize (pb, pos);
}

/* User Data atom */
static int mov_write_udta_tag(ByteIOContext *pb, PBYTE pbyBuffer)
{
    int atomSize = SWAP_WORDS(((PULONG)pbyBuffer)[0]);

    put_buffer(pb, pbyBuffer, atomSize);

    return 0;
}

/* Chunk offset atom */
static int mov_write_stco_tag(ByteIOContext *pb, MOVTrack* track)
{
    int i;
    offset_t next_frame_pos;
    offset_t pos = url_ftell(pb);
    int mode64 = 0;
    unsigned int samples_in_chunk = 1;
    int uiCountWritten = 1; //Start from 1, since 1st one is written before for-loop

    put_be32(pb, 0); /* size */

    if (pos > UINT32_MAX) {
        mode64 = 1;
        put_tag(pb, "co64");
    } 
    else
    {
        put_tag(pb, "stco");
    }
    put_be32(pb, 0); /* version & flags */
    put_be32(pb, track->chunksCount); // entry count

    if(mode64)
    {
        put_be64(pb, track->cluster[0][0].pos);
    }
    else
    {
        put_be32(pb, track->cluster[0][0].pos);
    }

    next_frame_pos = track->cluster[0][0].pos + track->cluster[0][0].size;

    samples_in_chunk = track->cluster[0][0].samplesInChunk;

    for (i=1; i<track->entry; i++) {
        int cl = i / MOV_INDEX_CLUSTER_SIZE;
        int id = i % MOV_INDEX_CLUSTER_SIZE;

        if ((next_frame_pos != track->cluster[cl][id].pos) || (samples_in_chunk >= track->maxSamplesPerChunk))
        {
            // was a gap, starting a new chunk:
            if(mode64)
            {
                put_be64(pb, track->cluster[cl][id].pos);
            }
            else
            {
                put_be32(pb, track->cluster[cl][id].pos);
            }

            samples_in_chunk = 0;
            uiCountWritten++;
        }

        samples_in_chunk += track->cluster[cl][id].samplesInChunk;

        next_frame_pos = track->cluster[cl][id].pos + track->cluster[cl][id].size;
    }

    if (uiCountWritten != track->chunksCount)
    {
        offset_t curpos = url_ftell(pb);
        Avf_url_fseek(pb, pos+12, SEEK_SET);
        put_be32(pb, uiCountWritten); /* rewrite size */
        Avf_url_fseek(pb, curpos, SEEK_SET);
        printf("stco entry %c:%d(%d)\n", (track->enc->codec_type == CODEC_TYPE_VIDEO) ? 'V' :'A', 
                                           uiCountWritten,track->chunksCount );
    }

    return updateSize (pb, pos);
}

/* Sample size atom */
static int mov_write_stsz_tag(ByteIOContext *pb, MOVTrack* track)
{
    //note: in this function, use samplesInChunk instead of entries, and entries
    //was removed from structure MOVIentry

    int equalChunks = 1;
    int i, entries = 0, tst = -1, oldtst = -1;
    unsigned int j;

    offset_t pos = url_ftell(pb);
    put_be32(pb, 0); /* size */
    put_tag(pb, "stsz");
    put_be32(pb, 0); /* version & flags */

    if( (track->enc->codec_id == CODEC_ID_ADPCM_MS) ||
        (track->enc->codec_id == CODEC_ID_ADPCM_IMA_WAV) )
    {
        put_be32(pb, 1); // sample size
        put_be32(pb, track->cluster[0][0].samplesInChunk * track->entry); // sample count
        return updateSize (pb, pos);
    }

    for (i=0; i<track->entry; i++) {
        int cl = i / MOV_INDEX_CLUSTER_SIZE;
        int id = i % MOV_INDEX_CLUSTER_SIZE;

        // Prevent divizion by zero
        if ( track->cluster[cl][id].samplesInChunk == 0 )
            continue;

        tst = track->cluster[cl][id].size/track->cluster[cl][id].samplesInChunk;
        if(oldtst != -1 && tst != oldtst) {
            equalChunks = 0;
        }
        oldtst = tst;
        entries += track->cluster[cl][id].samplesInChunk;
    }
    if (equalChunks) {
        int sSize = track->cluster[0][0].size/track->cluster[0][0].samplesInChunk;
        put_be32(pb, sSize); // sample size
        put_be32(pb, track->sampleCount); // sample count
    }
    else {
        put_be32(pb, 0); // sample size
        put_be32(pb, track->sampleCount); // sample count
        for (i=0; i<track->entry; i++) {
            int cl = i / MOV_INDEX_CLUSTER_SIZE;
            int id = i % MOV_INDEX_CLUSTER_SIZE;
            for ( j=0; j<track->cluster[cl][id].samplesInChunk; j++) {
                // Prevent divizion by zero
                if ( track->cluster[cl][id].samplesInChunk == 0 )
                    continue;

                put_be32(pb, track->cluster[cl][id].size /
                         track->cluster[cl][id].samplesInChunk);
            }
        }
    }
    return updateSize (pb, pos);
}

/* Sample to chunk atom */
static int mov_write_stsc_tag(ByteIOContext *pb, MOVTrack* track)
{
    int index = 1, written = 0, i;
    offset_t entryPos, curpos;
    unsigned int oldval = (unsigned int)-1;
    unsigned int prev_cl = 0;
    unsigned int prev_id = 0;
    unsigned int samples_in_chunk;

    offset_t pos = url_ftell(pb);
    put_be32(pb, 0); /* size */
    put_tag(pb, "stsc");
    put_be32(pb, 0); // version & flags
    entryPos = url_ftell(pb);

    put_be32(pb, track->chunksCount); // entry count

    samples_in_chunk = track->cluster[0][0].samplesInChunk;

    for (i=1; i<track->entry; i++) {
        int cl = i / MOV_INDEX_CLUSTER_SIZE;
        int id = i % MOV_INDEX_CLUSTER_SIZE;

        if( (track->cluster[cl][id].pos >
            (track->cluster[prev_cl][prev_id].pos + track->cluster[prev_cl][prev_id].size)) ||
            (samples_in_chunk >= track->maxSamplesPerChunk) )
        {
            //starting a new chunk
            if(oldval != samples_in_chunk)
            {
                put_be32(pb, index);            // first chunk
                put_be32(pb, samples_in_chunk); // samples per chunk
                put_be32(pb, 0x1);              // sample description index

                oldval = samples_in_chunk;
                written++;
            }

            samples_in_chunk = 0;
            index++;
        }

        samples_in_chunk += track->cluster[cl][id].samplesInChunk;

        prev_cl = cl;
        prev_id = id;
    }

    // write a last chunk info:

    if(oldval != samples_in_chunk){
        put_be32(pb, index);            // first chunk
        put_be32(pb, samples_in_chunk); // samples per chunk
        put_be32(pb, 0x1);              // sample description index
        written ++;
    }

    curpos = url_ftell(pb);
    Avf_url_fseek(pb, entryPos, SEEK_SET);
    put_be32(pb, written); // rewrite size
    Avf_url_fseek(pb, curpos, SEEK_SET);

    return updateSize (pb, pos);
}

/* Sync sample atom */
static int mov_write_stss_tag(ByteIOContext *pb, MOVTrack* track)
{
    offset_t curpos, entryPos = 0;
    int i, index = 0;
    int have_key = 0;
    offset_t pos = url_ftell(pb);
    BOOL chk_gop = ( track->enc->codec_id == CODEC_ID_MJPEG ? FALSE : TRUE );

    put_be32(pb, 0); // size
    put_tag(pb, "stss");
    put_be32(pb, 0); // version & flags
    entryPos = url_ftell(pb);

    if( track->hasKeyframes )
    {
        put_be32(pb, track->entry); // entry count

        if ( track->enc->codec_id != CODEC_ID_H264Z )
        {
            for (i=0; i<track->entry; i++) 
            {
                int cl = i / MOV_INDEX_CLUSTER_SIZE;
                int id = i % MOV_INDEX_CLUSTER_SIZE;

                if(chk_gop)
                {
                    if(i % track->enc->gop_size == 0)
                    {
                        have_key = 0;
                    }
                }

                if(track->cluster[cl][id].key_frame == 1)
                {
                    if( track->enc->codec_id == CODEC_ID_MJPEG ||
                        ( chk_gop && !have_key ) )
                    {
                        put_be32(pb, i+1);
                        index++;
                        have_key = 1;
                    }
                }
            }
        }
        else
        {
            /* in case of H.264 use more understandable logic for writing stss*/
            for (i=0; i<track->sampleCount; i++) 
            {
                int cl = i / MOV_INDEX_CLUSTER_SIZE;
                int id = i % MOV_INDEX_CLUSTER_SIZE;

                if(track->cluster[cl][id].key_frame == 1)
                {
                    put_be32(pb, i+1);
                    index++;
                }
            }
        }
    }
    else
    {
        put_be32(pb, 0); // entry count
    }
    curpos = url_ftell(pb);
    Avf_url_fseek(pb, entryPos, SEEK_SET);
    put_be32(pb, index); // rewrite size
    Avf_url_fseek(pb, curpos, SEEK_SET);

    return updateSize (pb, pos);
}

/* Composition Shift Least GreatestAtom */
static int mov_write_cslg_tag(ByteIOContext *pb, MOVTrack* track)
{
    offset_t pos = url_ftell(pb);

    put_be32(pb, 0); // size
    put_tag(pb, "cslg");
    put_be32(pb, 0); // version & flags

    put_be32(pb, track->sampleDuration);
    put_be32(pb, -track->sampleDuration);
    put_be32(pb, track->sampleDuration);

    put_be32(pb, 0);
    put_be32(pb, track->trackDuration);

    return updateSize (pb, pos);
}

/* Composition Offset Atom*/
static int mov_write_ctts_tag(ByteIOContext *pb, MOVTrack* track, int bMp4Format)
{
    int i, j, index = 0;
    int b_num; // max number of consecutive b-frames
    int comp_offset; // composition offset
    int d_offset = 0; // default offset (0 for .MOV 1 frame duration for .MP4)
    int entry, duration;
    offset_t entryPos, curpos, pos = url_ftell(pb);

    put_be32(pb, 0); // size
    put_tag(pb, "ctts");
    put_be32(pb, 0); // version & flags

    b_num = track->enc->max_b_frames;
    duration = track->sampleDuration; // duration of a frame

    if(b_num == 0){
        // all entries have the same values in this case:
        put_be32(pb, 1); // entry count

        put_be32(pb, track->entry); // sample count
        put_be32(pb, duration);

        return updateSize (pb, pos);
    }

    // contain B_Frames, so the values will be different:

    entryPos = url_ftell(pb);

    put_be32(pb, track->entry); // entry count

    for (i=0; i<track->entry; i++)
    {
        int cl = i / MOV_INDEX_CLUSTER_SIZE;
        int id = i % MOV_INDEX_CLUSTER_SIZE;

        comp_offset = 0;

        if ( bMp4Format )
        {
            d_offset = duration;
        }

        if(track->cluster[cl][id].dependency == 0x08)//b
        {
            // B-Frames are always display 1 frame before their decode order.
            comp_offset = -duration;
        }
        else // I and P
        {
            // For each B frame follow this frame before a non B frame
            // the composition time is delayed of 1 frame duration.
            for( j = 1; (j <= b_num) && (j + i < track->entry); j++)
            {
                int f_cl = (i + j) / MOV_INDEX_CLUSTER_SIZE;
                int f_id = (i + j) % MOV_INDEX_CLUSTER_SIZE;

                if(track->cluster[f_cl][f_id].dependency == 0x08)
                {
                    comp_offset += duration;
                }
                else
                {
                    break;
                }
            }
        }

        entry = d_offset + comp_offset;

        //Write the value in the table
        put_be32(pb, 1);
        put_be32(pb, entry);
        index++;
    }

    curpos = url_ftell(pb);
    Avf_url_fseek(pb, entryPos, SEEK_SET);
    put_be32(pb, index); // rewrite size
    Avf_url_fseek(pb, curpos, SEEK_SET);

    return updateSize (pb, pos);
}

/* Sample Dependency Atom */
static int mov_write_sdtp_tag(ByteIOContext *pb, MOVTrack* track)
{
    int i;
    offset_t pos = url_ftell(pb);

    put_be32(pb, 0); // size
    put_tag(pb, "sdtp");
    put_be32(pb, 0); // version & flags

    for (i=0; i<track->entry; i++) {
        int cl = i / MOV_INDEX_CLUSTER_SIZE;
        int id = i % MOV_INDEX_CLUSTER_SIZE;

        put_byte(pb, track->cluster[cl][id].dependency);
    }

    return updateSize (pb, pos);
}

#if 0 /* unused */
static int mov_write_damr_tag(ByteIOContext *pb)
{
    put_be32(pb, 0x11); /* size */
    put_tag(pb, "damr");
    put_tag(pb, "FFMP");
    put_byte(pb, 0);

    put_be16(pb, 0x80); /* Mode set (all modes for AMR_NB) */
    put_be16(pb, 0xa); /* Mode change period (no restriction) */
    //put_be16(pb, 0x81ff); /* Mode set (all modes for AMR_NB) */
    //put_be16(pb, 1); /* Mode change period (no restriction) */
    return 0x11;
}
#endif /* 0 */

static int mov_write_audio_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);

    // ZZZ: uncompressed sample is 16 bits, unless there is S8 or U8 format
    int uncompressed_smpl_size = 16;
    int bMp4Format = 1;
    put_be32(pb, 0); /* size */

    if ( strcmp("mp4", s->oformat->name) != 0 )
    {
        bMp4Format = 0;
    }

    if(track->enc->codec_id == CODEC_ID_PCM_MULAW)
      put_tag(pb, "ulaw");
    else if(track->enc->codec_id == CODEC_ID_PCM_S16LE)
      put_tag(pb, "sowt");
    else if(track->enc->codec_id == CODEC_ID_PCM_S16BE)
      put_tag(pb, "twos");
    else if(track->enc->codec_id == CODEC_ID_PCM_S8)
    {
      put_tag(pb, "twos");
      uncompressed_smpl_size = 8;
    }
    else if(track->enc->codec_id == CODEC_ID_PCM_U8)
    {
      put_tag(pb, "raw ");
      uncompressed_smpl_size = 8;
    }
    else if(track->enc->codec_id == CODEC_ID_PCM_ALAW)
      put_tag(pb, "alaw");
    else if(track->enc->codec_id == CODEC_ID_ADPCM_IMA_QT)
      put_tag(pb, "ima4");
    else if(track->enc->codec_id == CODEC_ID_MACE3)
      put_tag(pb, "MAC3");
    else if(track->enc->codec_id == CODEC_ID_MACE6)
      put_tag(pb, "MAC6");
    else if(track->enc->codec_id == CODEC_ID_AAC)
      put_tag(pb, "mp4a");
    else if(track->enc->codec_id == CODEC_ID_AACZ)
      put_tag(pb, "mp4a");
    else if(track->enc->codec_id == CODEC_ID_AMR_NB)
      put_tag(pb, "samr");
    else if (track->enc->codec_id == CODEC_ID_ADPCM_MS)
    {
      put_byte(pb, 'm'); put_byte(pb, 's'); put_byte(pb, 0); put_byte(pb, 2);
    }
    else if (track->enc->codec_id == CODEC_ID_ADPCM_IMA_WAV)
    {
      put_byte(pb, 'm'); put_byte(pb, 's'); put_byte(pb, 0); put_byte(pb, 0x11);
    }
    else
      put_tag(pb, "    ");

    put_be32(pb, 0);                    /* Reserved */
    put_be16(pb, 0);                    /* Reserved */
    put_be16(pb, 1);                    /* Data-reference index, XXX  == 1 */

    /* SoundDescription */
    if ( bMp4Format &&
        ((track->enc->codec_id == CODEC_ID_AACZ) ||
          (track->enc->codec_id == CODEC_ID_AAC) ))
    {
        put_be16(pb, 0);                    /* Version */
    }
    else
    {
        put_be16(pb, 1);                    /* Version */
    }
    put_be16(pb, 0);                    /* Revision level */
    put_be32(pb, 0);                    /* Reserved */
    put_be16(pb, track->enc->channels); /* Number of channels */
    put_be16(pb, uncompressed_smpl_size);  /* uncompressed sample size */

    if ( (!bMp4Format) &&
         ((track->enc->codec_id == CODEC_ID_AACZ) ||
          (track->enc->codec_id == CODEC_ID_AAC) ))
    {
        put_be16(pb, -2);                    /* compression ID (= 0) */
    }
    else
    {
        put_be16(pb, 0);                    /* compression ID (= 0) */
    }

    put_be16(pb, 0);                    /* packet size (= 0) */
    put_be16(pb, track->timescale);     /* Time scale */
    put_be16(pb, 0);                    /* Reserved */

    // Extended header (version 1). Add 4 longs:

    if ((track->enc->codec_id == CODEC_ID_ADPCM_MS) ||
        (track->enc->codec_id == CODEC_ID_ADPCM_IMA_WAV))
    {
        put_be32(pb, track->enc->frame_size);
        put_be32(pb, track->enc->block_align / track->enc->channels);
        put_be32(pb, track->enc->block_align);
        put_be32(pb, 2);
    }
    else if (track->enc->codec_id == CODEC_ID_MP2)
    {
        put_be32(pb, track->enc->frame_size);
        // In movdec, the bitrate calculation is based on frame_size and number of samples per block per channel
        // Although block_align is not updated by mp2 codec, adjust it here according to the bitrate
        put_be32(pb, track->enc->bit_rate
                     * track->enc->frame_size
                     / track->enc->sample_rate
                     / track->enc->channels
                     / 8);
        put_be32(pb, track->enc->block_align);
        put_be32(pb, 2);
    }
    else if ( (track->enc->codec_id == CODEC_ID_AACZ) ||
              (track->enc->codec_id == CODEC_ID_AAC) )
    {
        if (bMp4Format)
        {
            // Do not use extended header, instead use esds
            mov_write_esds_tag( pb, track, s );
        }
        else
        {
            offset_t wave_pos;
            // MOV file:

            // Extended header (version 1):
            put_be32(pb, track->enc->frame_size); //samples Per Packet
            put_be32(pb, 2 / track->enc->channels); // bytes Per Packet
            put_be32(pb, 2); // bytes Per Frame;
            put_be32(pb, 2); // bytesPerSample

            wave_pos = url_ftell(pb);

            //wave tag size
            put_be32(pb, 0);
            put_tag (pb, "wave");
            put_be32(pb, 12);
            put_tag (pb, "frma");
            put_tag (pb, "mp4a");
            put_be32(pb, 12);
            put_tag (pb, "mp4a");
            put_be32(pb, 0);

            mov_write_esds_tag( pb, track, s );

            //finishing wave tag
            put_be32(pb, 0x8); // size
            put_be32(pb, 0x0); // null tag

            updateSize (pb, wave_pos);
        }
    }
    else
    {
        // write the extended header for version 1 as 0, so playback reader will ignore these items
        put_be32(pb, 0);
        put_be32(pb, 0);
        put_be32(pb, 0);
        put_be32(pb, 0);
    }


    if( (track->enc->codec_id == CODEC_ID_ADPCM_MS) ||
        (track->enc->codec_id == CODEC_ID_ADPCM_IMA_WAV) )
    {
        // WAV header (new). See: http://www.bulba.eu.org/informatyka/formaty/muzyka/new_wave_types.html
        offset_t wave_pos, ms_pos = 0;

        int ms_codec_id = (track->enc->codec_id == CODEC_ID_ADPCM_MS ? 2 : 17);

        wave_pos = url_ftell(pb);

        put_be32(pb, 0);
        put_tag (pb, "wave");

        put_be32(pb, 12);
        put_tag (pb, "frma");
        put_byte(pb, 'm'); put_byte(pb, 's'); put_byte(pb, 0); put_byte(pb, ms_codec_id);

        ms_pos = url_ftell(pb);

        put_be32(pb, 0);
        put_byte(pb, 'm'); put_byte(pb, 's'); put_byte(pb, 0); put_byte(pb, ms_codec_id);

        put_wav_header(pb, track->enc);

        updateSize (pb, ms_pos);
        updateSize (pb, wave_pos);
    }

    return updateSize (pb, pos);
}

static int mov_write_d263_tag(ByteIOContext *pb)
{
    put_be32(pb, 0xf); /* size */
    put_tag(pb, "d263");
    put_tag(pb, "FFMP");
    put_be16(pb, 0x0a);
    put_byte(pb, 0);
    return 0xf;
}

/* TODO: No idea about these values */
static int mov_write_svq3_tag(ByteIOContext *pb)
{
    put_be32(pb, 0x15);
    put_tag(pb, "SMI ");
    put_tag(pb, "SEQH");
    put_be32(pb, 0x5);
    put_be32(pb, 0xe2c0211d);
    put_be32(pb, 0xc0000000);
    put_byte(pb, 0);
    return 0x15;
}

/* Get/Put unsigned int from memory buffer: */
static inline void mov_enc_put32(uint8_t *buf, int val)
{
    buf[0] = (val >> 24) & 0xFF;
    buf[1] = (val >> 16) & 0xFF;
    buf[2] = (val >> 8 ) & 0xFF;
    buf[3] =  val & 0xFF;
}

static inline unsigned int mov_enc_get32(uint8_t *buf)
{
    return ((unsigned int)buf[0] << 24) |
            ((unsigned int)buf[1] << 16) |
            ((unsigned int)buf[2] << 8)  |
            (unsigned int)buf[3];
}

/* Looking for long h264 NAL start code (4-bytes length)*/
static uint8_t *h264_find_startcode( uint8_t *p, uint8_t *end )
{
    while(( p[0] != 0 || p[1] != 0 || p[2] != 0 || p[3] != 1) && (p < end - 3)){
        p++;
    }

    if(p >= end - 3){
        return end + 3;
    }

    return p;
}

static int mov_write_avcC_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    INT	data_size;
    INT avcc_atom_size = 0;
    uint8_t* data_ptr = track->vosData;
    offset_t pos = url_ftell(pb);

    if (track->vosLen != 0)
    {
        // track->vosData should contains SPS and PPS H264 headers in the separate NAL units;
        // we should write them to avcC atom:

        //SPS:
        // check headers and calculate size:
        data_size = mov_enc_get32(data_ptr);
        data_ptr += data_size + 4;

        if ( data_size + 4 > track->vosLen)
            return 0x0; // some error

        //SPS:
        // check headers and calculate size:
        data_size += mov_enc_get32(data_ptr);
        data_ptr = track->vosData;

        pos = url_ftell(pb);

        put_be32(pb, data_size + 19);
        put_tag(pb, "avcC");
        put_byte(pb, 1);            // version
        put_byte(pb, data_ptr[5]);  // profile
        put_byte(pb, data_ptr[6]);  // profile compatibility
        put_byte(pb, data_ptr[7]);  // level
        put_byte(pb, 0xff);         // 6 bits reserved (111111) + 2 bits nal size length - 1 (11)
        put_byte(pb, 0xe1);         // 3 bits reserved (111) + 5 bits number of sps (00001)

        avcc_atom_size = 14;

        // Put SPS:
        data_size = mov_enc_get32(data_ptr);
        put_be16(pb, data_size);    //SPS Size
        data_ptr += 4;
        put_buffer(pb, data_ptr, data_size);

        put_byte(pb, 1);            // pps num

        avcc_atom_size += 2 + data_size + 1;

        // Put PPS:
        data_ptr += data_size;
        data_size = mov_enc_get32(data_ptr);
        put_be16(pb, data_size);    //PPS Size
        data_ptr += 4;
        put_buffer(pb, data_ptr, data_size);

        avcc_atom_size += 2 + data_size;

        pos = url_ftell(pb) - pos;
    }
    else if (s->sMovStreamsInfo.uiVosLen != 0) {
        pos = url_ftell(pb);

        put_be32(pb, s->sMovStreamsInfo.uiVosLen + 8);
        put_tag(pb, "avcC");
        put_buffer(pb, s->sMovStreamsInfo.pbyVosData, s->sMovStreamsInfo.uiVosLen);

        avcc_atom_size = pos - url_ftell(pb);
    }

    return avcc_atom_size;
}

static int mov_write_colr_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);

    put_be32( pb, 0 );            /* size */
    put_tag( pb, "colr" );      /* name */
    put_tag( pb, "nclc" );      /* color parameter type */
    put_be16( pb, s->puwColrAtom[0] );      /* primaries index */
    put_be16( pb, s->puwColrAtom[1] );      /* transfer function index */
    put_be16( pb, s->puwColrAtom[2] );      /* matrix index */

    return updateSize( pb, pos );
}

static unsigned int descrLength(unsigned int len)
{
    if (len < 0x00000080)
        return 2 + len;
    else if (len < 0x00004000)
        return 3 + len;
    else if(len < 0x00200000)
        return 4 + len;
    else
        return 5 + len;
}

static void putDescr(ByteIOContext *pb, int tag, int size)
{
    uint32_t len;
    uint8_t  vals[4];

    len = size;
    vals[3] = (uint8_t)(len & 0x7f);
    len >>= 7;
    vals[2] = (uint8_t)((len & 0x7f) | 0x80);
    len >>= 7;
    vals[1] = (uint8_t)((len & 0x7f) | 0x80);
    len >>= 7;
    vals[0] = (uint8_t)((len & 0x7f) | 0x80);

    put_byte(pb, tag); // DescriptorTag

    if (size < 0x00000080)
    {
        put_byte(pb, vals[3]);
    }
    else if (size < 0x00004000)
    {
        put_byte(pb, vals[2]);
        put_byte(pb, vals[3]);
    }
    else if (size < 0x00200000)
    {
        put_byte(pb, vals[1]);
        put_byte(pb, vals[2]);
        put_byte(pb, vals[3]);
    }
    else if (size < 0x10000000)
    {
        put_byte(pb, vals[0]);
        put_byte(pb, vals[1]);
        put_byte(pb, vals[2]);
        put_byte(pb, vals[3]);
    }
}

static int mov_write_esds_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s) // Basic
{
    int decoderSpecificInfoLen = track->vosLen ? descrLength(track->vosLen):0;
    offset_t pos = url_ftell(pb);

    put_be32(pb, 0);               // size
    put_tag(pb, "esds");
    put_be32(pb, 0);               // Version

    // ES descriptor
    putDescr(pb, 0x03, 3 + descrLength(13 + decoderSpecificInfoLen) +
             descrLength(1));

    if( track->enc->codec_type == CODEC_TYPE_AUDIO ||
       ( strcmp("mp4", s->oformat->name) == 0 ) )
    {
        put_be16(pb, 0x0000);          // ID (= 0)
    }
    //Do not write this in case of MP4
    else if ( strcmp("mp4", s->oformat->name) != 0 )
    {
        put_be16(pb, 0x0001);          // ID (= 1)
    }

    put_byte(pb, 0x00);            // flags (= no flags)

    // DecoderConfig descriptor
    putDescr(pb, 0x04, 13 + decoderSpecificInfoLen);

    if ((track->enc->codec_id == CODEC_ID_AAC) ||
        (track->enc->codec_id == CODEC_ID_AACZ) )
    {
        put_byte(pb, 0x40);        // Object type indication
    }
    else if(track->enc->codec_id == CODEC_ID_MPEG4Z)
    {
        put_byte(pb, 0x20);        // Object type indication (Visual 14496-2)
    }
    
    if(track->enc->codec_type == CODEC_TYPE_AUDIO)
    {
        put_byte(pb, 0x15);            // flags (= Audiostream)

        put_byte(pb, 0x00);                     // Buffersize DB (24 bits)

        if ( strcmp("mp4", s->oformat->name) == 0 )
        {
            // TODO: Why this hard coded value of 768?
            put_be16(pb, 768 * track->enc->channels);
        }
        else
        {
            if ((track->enc->codec_id == CODEC_ID_AAC) ||
                (track->enc->codec_id == CODEC_ID_AACZ) )
            {
                put_be16(pb, 768 * 2 * 2 * track->enc->channels);
            }
            else
            {
                put_be16(pb, track->enc->frame_size);
            }
        }

        put_be32(pb, track->enc->bit_rate);     // maxbitrate
        put_be32(pb, track->enc->bit_rate);     // avg bitrate
    }
    else
    {
        put_byte(pb, 0x11);            // flags (= Visualstream)

        put_byte(pb, 0x0);             // Buffersize DB (24 bits)
        put_be16(pb, 0x0dd2);          // Buffersize DB

        if (track->enc->codec_type == CODEC_TYPE_VIDEO)
        {
            put_be32(pb, track->enc->bit_rate);     // maxbitrate
            put_be32(pb, track->enc->bit_rate);     // avg bitrate
        }
        else
        {
            // TODO: find real values for these
            put_be32(pb, 0x0002e918);     // maxbitrate
            put_be32(pb, 0x00017e6b);     // avg bitrate
        }
    }

    if (track->vosLen)
    {
        // DecoderSpecific info descriptor
        putDescr(pb, 0x05, track->vosLen);
        put_buffer(pb, track->vosData, track->vosLen);
    }

    // SL descriptor
    //putDescr(pb, 0x06, descrLength(1));
    putDescr(pb, 0x06, 1);
    put_byte(pb, 0x02);
    
    return updateSize (pb, pos);
}

static int mov_write_video_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);
    char* str;

    put_be32(pb, 0); /* size */
    if(track->enc->codec_id == CODEC_ID_SVQ1)
      put_tag(pb, "SVQ1");
    else if(track->enc->codec_id == CODEC_ID_SVQ3)
      put_tag(pb, "SVQ3");
    else if(track->enc->codec_id == CODEC_ID_MPEG4Z)
      put_tag(pb, "mp4v");
    else if(track->enc->codec_id == CODEC_ID_H263)
      put_tag(pb, "s263");
    else if(track->enc->codec_id == CODEC_ID_MJPEG)
    {
        if( s->bMjpegASupport == TRUE )
        {
          put_tag(pb, "mjpa");
        }
        else
        {
            put_tag(pb, "jpeg");
        }
    }
    else if(track->enc->codec_id == CODEC_ID_H264Z)
      put_tag(pb, "avc1");
    else
      put_tag(pb, "    "); /* Unknown tag */

    put_be32(pb, 0); /* Reserved */
    put_be16(pb, 0); /* Reserved */
    put_be16(pb, 1); /* Data-reference index */

    put_le16(pb, s->uwModel); /* Version */
    put_le16(pb, s->uwRevision); /* Revision */
    put_le32(pb, s->ulVendorName); /* 4 char vendor id*/

    if ( strcmp("mp4", s->oformat->name) == 0 )
    {
        put_be32(pb, 0); /* Reserved */
        put_be32(pb, 0); /* Reserved (400)*/
    }
    else
    {
        int temporal_quality = (track->enc->is_intra_only) ? 0 : 512;
        put_be32(pb, temporal_quality);
        put_be32(pb, 512); /* spatial quality */
    }

    put_be16(pb, track->enc->logical_width); /* Video width */
    put_be16(pb, track->enc->logical_height); /* Video height */
    put_be32(pb, 0x00480000); /* Reserved */
    put_be32(pb, 0x00480000); /* Reserved */
    put_be32(pb, 0); /* Reserved */
    put_be16(pb, 1); /* frame count */

    if( track->enc->codec_id == CODEC_ID_MPEG4Z)
    {
        if ( strcmp("mp4", s->oformat->name) == 0 )
        {
            str = "";
        }
        else
        {
            str = "MPEG - 4";
        }
    }
    else if ( track->enc->codec_id == CODEC_ID_H264Z)
    {
        str = "AVC Coding";
    }
    else
    {
        if( s->bMjpegASupport == TRUE )
        {
            str = "Motion-JPEG A";
        }
        else
        {
            str = "Photo - JPEG";
        }
    }

    put_pascal_str( pb, str, 32 );

    put_be16(pb, 0x18); /* Reserved */
    put_be16(pb, 0xffff); /* Reserved */
    if(track->enc->codec_id == CODEC_ID_MPEG4Z)
        mov_write_esds_tag(pb, track, s);
    else if(track->enc->codec_id == CODEC_ID_H263)
        mov_write_d263_tag(pb);
    else if(track->enc->codec_id == CODEC_ID_SVQ3)
        mov_write_svq3_tag(pb);
    else if(track->enc->codec_id == CODEC_ID_H264Z)
    {
            mov_write_avcC_tag(pb, track, s);
                if( s->puwColrAtom != NULL )
                {
                    mov_write_colr_tag(pb, track, s);
                }
            }

    return updateSize (pb, pos);
}

static int mov_write_text_tag(ByteIOContext *pb, MOVTrack* track)
{
    offset_t pos = url_ftell(pb);

    put_be32(pb, 0); /* size */
    put_tag(pb, "text");
    put_be32(pb, 0); /* Reserved */
    put_be16(pb, 0); /* Reserved */
    put_be16(pb, 1); /* Data-reference index */

    /* text description */

    put_be32(pb, 0);  /* display flag */
    put_be32(pb, 1);  /* text alignment. 0 - left
                                         1 - center
                                        -1 - right */

    //Background color
    put_be16(pb, 0x0);                    /* red */
    put_be16(pb, 0x0);                    /* green */
    put_be16(pb, 0x0);                    /* blue */

    //Default text box
    put_be16(pb, 0x0);                    /*top*/
    put_be16(pb, 0x0);                    /*left*/
    put_be16(pb, 0x0);                    /*bottom*/
    put_be16(pb, 0x0);                    /*right*/

    put_be64(pb, 0);                     /*reserved, must be zero*/

    put_be16(pb, 0x3); /*Font number,A 16-bit value that must be set to 0.*/
    /*
    A 16-bit integer that indicates the fonts style. Set this field to 0 for 
    normal text. You can enable other style options by using one or more of the 
    following bit masks:
    0x0001 Bold
    0x0002 Italic
    0x0004 Underline
    0x0008 Outline
    0x0010 Shadow
    0x0020 Condense
    0x0040 Extend
    */
    put_be16(pb, 0x0000);

//    put_byte(pb, 0);                  /*reserved, must be zero*/
    put_be16(pb, 0);                    /*reserved, must be zero*/

    /* foreground color */
    put_be16(pb, 0xFFFF); // R
    put_be16(pb, 0xFFFF); // G
    put_be16(pb, 0xFFFF); // B

    put_pascal_str(pb, "Geneva", 6); // font

    return updateSize (pb, pos);
}

static int mov_write_stsd_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);
    put_be32(pb, 0); /* size */
    put_tag(pb, "stsd");
    put_be32(pb, 0); /* version & flags */
    put_be32(pb, 1); /* entry count */
    if( track->enc->codec_type == CODEC_TYPE_VIDEO )
    {
        mov_write_video_tag(pb, track, s);
    }
    else if( track->enc->codec_type == CODEC_TYPE_AUDIO )
    {
        mov_write_audio_tag(pb, track, s);
    }
    else if( track->enc->codec_type == CODEC_TYPE_TEXT )
    {
        mov_write_text_tag(pb, track);
    }
    return updateSize(pb, pos);
}

static int mov_convert_text_stream_duration( MOVTrack* track )
{
    int i;
    unsigned int cur=0, prev=0;
    unsigned int duration, total=0;
    int cl=0, id=0, cl_p=0, id_p=0;

    for( i = 0; i < track->entry; i++ )
    {
        cl = i / MOV_INDEX_CLUSTER_SIZE;
        id = i % MOV_INDEX_CLUSTER_SIZE;

        cur = track->cluster[cl][id].duration;

        if( i == 0 )
        {
            prev = cur;
            cl_p = cl;
            id_p = id;
            continue;
        }

        duration = cur - prev;
        track->cluster[cl_p][id_p].duration = duration * track->sampleDuration;
        total += track->cluster[cl_p][id_p].duration;
        cl_p = cl;
        id_p = id;
        prev = cur;
    }

    // Update last entry
    track->cluster[cl][id].duration = track->trackDuration - total;

    return 1;
}

/* TODO?: Currently all samples/frames seem to have same duration */
/* Time to sample atom */
static int mov_write_stts_tag(ByteIOContext *pb, MOVTrack* track)
{
    offset_t pos = url_ftell(pb);

    put_be32(pb, 0); /* size */

    put_tag(pb, "stts");
    put_be32(pb, 0); /* version & flags */

    if( track->enc->codec_type != CODEC_TYPE_TEXT )
    {
        put_be32(pb, 1); /* entry count */
        put_be32(pb, track->sampleCount); /* sample count */
        put_be32(pb, track->sampleDuration); /* sample duration */
    }
    else // for text stream the algorythm is slightly differents
    {
        int i;
        unsigned int dur=0, prev_dur=0;
        int count = 0, entries = 0;
        offset_t pos1 = url_ftell( pb );
        put_be32(pb, 0); /* entry count */

        // Text stream will have indexes of video frame
        // in this field
        mov_convert_text_stream_duration( track );

        // Write the durations to the file
        for( i = 0; i < track->entry; i++ )
        {
            int cl = i / MOV_INDEX_CLUSTER_SIZE;
            int id = i % MOV_INDEX_CLUSTER_SIZE;

            dur = track->cluster[cl][id].duration;
            // Initialize
            if( i == 0 )
            {
                prev_dur = dur;
            }

            if( dur != prev_dur )
            {
                // update last entry
                put_be32(pb, count); /* sample count */
                put_be32(pb, prev_dur ); /* sample duration */
                count = 0;
                entries++;
            }

            prev_dur = dur;
            count++;
        }

        put_be32(pb, count); /* sample count */
        put_be32(pb, prev_dur ); /* sample duration */
        entries++;

        updateValue(pb, pos1, entries);
    }

    return updateSize(pb, pos);
}

static int mov_write_dref_tag(ByteIOContext *pb, MOVEncodeContext *mov)
{
    offset_t pos = url_ftell(pb);

    put_be32(pb, 0); /* size */
    put_tag(pb, "dref");
    put_be32(pb, 0); /* version & flags */
    put_be32(pb, 1); /* entry count */

    put_be32(pb, 0xc); /* size */

    if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_DREF_URL )
    {
        put_tag(pb, "url ");
    }
    else
    {
        put_tag(pb, "alis");
    }

    put_be32(pb, 1); /* version & flags */

    return updateSize(pb, pos);
}

static void mov_write_stss_tag_ex(ByteIOContext *pb, MOVTrack* track)
{
    MOVAtomContent *tag = &atom[MOV_ATOM_ID_STSS];
    int delInfo = 0;

    if( tag->data != NULL )
    {
        delInfo = *(int *)tag->data;
    }

    // This table must be present in case of video, because it is mandatory
    // for correct playback on Quick Time player.
    if( track->enc->codec_type == CODEC_TYPE_VIDEO )
    {
        // delInfo determine whether the host decide to keep
        // stss atom or not. 1: delete it. 0: keep it.
        // this atom is only allowed to be deleted in intra-only files
        if( delInfo == 0 ||
            ((track->enc->codec_id != CODEC_ID_MJPEG) && (!track->enc->is_intra_only)))
        {
            mov_write_stss_tag(pb, track);
        }
    }
}

static int mov_write_stbl_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);
    MOVEncodeContext *mov = s->priv_data;
    int bMp4Format = 1;

    if ( strcmp("mp4", s->oformat->name) != 0)
    {
        bMp4Format = 0;
    }

    put_be32(pb, 0); /* size */
    put_tag(pb, "stbl");
    mov_write_stsd_tag(pb, track, s);
    mov_write_stts_tag(pb, track);

    if ( (track->enc->codec_id == CODEC_ID_H264Z) && bMp4Format)
    {
        mov_write_ctts_tag(pb, track, bMp4Format);
    }

    if ( mov->ulEncodingFlags & MOV_ATOM_MOVE_STSS_TO_STBL_END )
    {
        mov_write_stsc_tag(pb, track);
        mov_write_stsz_tag(pb, track);
        mov_write_stco_tag(pb, track);

        if ( (track->enc->codec_id == CODEC_ID_H264Z) && (!bMp4Format) )
        {
            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_CSLG)
            {
                mov_write_cslg_tag(pb, track);
            }

            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_CTTS)
            {
                mov_write_ctts_tag(pb, track, bMp4Format);
            }
            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_SDTP)
            {
                mov_write_sdtp_tag(pb, track);
            }
        }

        mov_write_stss_tag_ex(pb, track);
    }
    else
    {
        mov_write_stss_tag_ex(pb, track);

        mov_write_stsc_tag(pb, track);
        mov_write_stsz_tag(pb, track);
        mov_write_stco_tag(pb, track);

        if ( (track->enc->codec_id == CODEC_ID_H264Z) && (!bMp4Format) )
        {
            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_CSLG)
            {
                mov_write_cslg_tag(pb, track);
            }

            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_CTTS)
            {
                mov_write_ctts_tag(pb, track, bMp4Format);
            }

            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_SDTP)
            {
                mov_write_sdtp_tag(pb, track);
            }
        }
    }

    return updateSize(pb, pos);
}

static int mov_write_dinf_tag(ByteIOContext *pb, MOVEncodeContext *mov)
{
    offset_t pos = url_ftell(pb);
    put_be32(pb, 0); /* size */
    put_tag(pb, "dinf");
    mov_write_dref_tag(pb, mov);
    return updateSize(pb, pos);
}

static int mov_write_smhd_tag(ByteIOContext *pb)
{
    put_be32(pb, 16); /* size */
    put_tag(pb, "smhd");
    put_be32(pb, 0); /* version & flags */
    put_be16(pb, 0); /* reserved (balance, normally = 0) */
    put_be16(pb, 0); /* reserved */
    return 16;
}

static int mov_write_vmhd_tag(ByteIOContext *pb, AVFormatContext *s)
{
    put_be32(pb, 0x14); /* size (always 0x14) */
    put_tag(pb, "vmhd");
    put_be32(pb, 0x01); /* version & flags */

    if ( strcmp("mp4", s->oformat->name) == 0 )
    {
        put_be16(pb, 0x0000); /* Graphics Mode */
        put_be16(pb, 0x0000); /* Operation Color (R) */
        put_be16(pb, 0x0000); /* Operation Color (G) */
        put_be16(pb, 0x0000); /* Operation Color (B) */
    }
    else
    {
        put_be16(pb, 0x0040);
        put_be16(pb, 0x8000);
        put_be16(pb, 0x8000);
        put_be16(pb, 0x8000);
    }

    return 0x14;
}

static int mov_write_gmin_tag(ByteIOContext *pb)
{
    offset_t pos = url_ftell(pb);

    put_be32(pb, 0x0); /* size */
    put_tag(pb, "gmin");

    put_be32(pb, 0); /* version=1B & flags=3B */
    put_be16(pb, 64); // Graphics Mode - dither or copy
    put_be16(pb, 0x0080); // Red of 6 bytes Opcolor
    put_be16(pb, 0x0080); // Green of 6 bytes Opcolor
    put_be16(pb, 0x0080); // Blue of 6 bytes Opcolor
    put_be16(pb, 0); //Balance
    put_be16(pb, 0); //Reserved

    return updateSize(pb, pos);
}

static int mov_write_gmhd_tag(ByteIOContext *pb)
{
    offset_t pos = url_ftell(pb);
    offset_t pos1;

    put_be32(pb, 0x0);
    put_tag(pb, "gmhd");
    mov_write_gmin_tag(pb);

    // TODO
    // patch, no description in spec
    pos1 = url_ftell(pb);
    put_be32(pb, 0x0);
    put_tag(pb, "text");
    put_be32(pb, 0x100);
    put_be32(pb, 0x0);
    put_be32(pb, 0x0);
    put_be32(pb, 0x0);
    put_be32(pb, 0x100);
    put_be32(pb, 0x0);
    put_be32(pb, 0x0);
    put_be32(pb, 0x0);
    put_be32(pb, 0x40);
    updateSize(pb, pos1);

    return updateSize(pb, pos);
}

static int mov_write_minf_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);
    MOVEncodeContext *mov = s->priv_data;

    put_be32(pb, 0); /* size */
    put_tag(pb, "minf");
    if(track->enc->codec_type == CODEC_TYPE_VIDEO)
    {
        mov_write_vmhd_tag(pb, s);
    }
    else if( track->enc->codec_type == CODEC_TYPE_AUDIO )
    {
        mov_write_smhd_tag(pb);
    }
    else if( track->enc->codec_type == CODEC_TYPE_TEXT )
    {
        mov_write_gmhd_tag(pb);
    }
    else if ( !strcmp("mp4", s->oformat->name) )
    {
        // We add the 'nmhd' (null media header) tag
        // into MP4 file in case of undefined codec type
        mov_write_nmhd_tag(pb);
    }

    if ( (mov->ulEncodingFlags & MOV_ATOM_REMOVE_HDLR_ALIS) == 0 )
    {
        mov_write_hdlr_alis_tag(pb, track, s);
    }

    if( ( s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_DINF ) ||
        ( mov->ulEncodingFlags == MOV_ATOM_DEFAULT ) )
    {
        mov_write_dinf_tag(pb, mov);
    }
    mov_write_stbl_tag(pb, track, s);
    return updateSize(pb, pos);
}

static int mov_write_hdlr_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    char *str;
    offset_t pos = url_ftell(pb);

    if(track->enc->codec_type == CODEC_TYPE_VIDEO)
    {
        put_be32(pb, 0); /* size */
        put_tag(pb, "hdlr");
        put_be32(pb, 0); /* Version & flags */

        if ( !strcmp("mp4", s->oformat->name) )
        {
            put_be32(pb ,0); /* reserved */
        }
        else
        {
            put_tag(pb, "mhlr"); /* reserved */
        }

        put_tag(pb, "vide"); /* handler type */
        put_le32(pb , s->ulVendorName); /* 4 character vendor id */
        put_be32(pb ,0); /* reserved */
        put_be32(pb ,0); /* reserved */

        if ( !strcmp("mp4", s->oformat->name) )
        {
            /* .mp4: C string */
            str = "Video Media Handler";
            put_buffer(pb, (const unsigned char *)str, strlen(str)+1); /* include terminator */
        }
        else
        {
            /* .mov: PASCAL string */
            str = "";
            put_byte(pb, strlen(str)); /* string counter */
            put_buffer(pb, (const unsigned char *)str, strlen(str));
        }
    }
    else if( track->enc->codec_type == CODEC_TYPE_AUDIO )
    {
        put_be32(pb, 0); /* size */
        put_tag(pb, "hdlr");
        put_be32(pb, 0); /* Version & flags */

        if ( !strcmp("mp4", s->oformat->name) )
        {
            put_be32(pb ,0); /* reserved */
        }
        else
        {
            put_tag(pb, "mhlr"); /* reserved */
        }

        put_tag(pb, "soun"); /* handler type */
        put_le32(pb, s->ulVendorName); /* handler type */
        put_be32(pb ,0); /* reserved */
        put_be32(pb ,0); /* reserved */

        if ( !strcmp("mp4", s->oformat->name) )
        {
            /* .mp4: C string */
            str = "Sound Media Handler";
            put_buffer(pb, (const unsigned char *)str, strlen(str)+1); /* include terminator */
        }
        else
        {
            /* .mov: PASCAL string */
            str = "";
            put_byte(pb, strlen(str)); /* string counter */
            put_buffer(pb, (const unsigned char *)str, strlen(str));
        }
    }
    else if( track->enc->codec_type == CODEC_TYPE_TEXT )
    {
        put_be32(pb, 0); /* size */
        put_tag(pb, "hdlr");
        put_be32(pb, 0); /* Version & flags */

        if ( !strcmp("mp4", s->oformat->name) )
        {
            put_be32(pb ,0); /* reserved */
        }
        else
        {
            put_tag(pb, "mhlr"); /* reserved */
        }

        put_tag(pb, "text"); /* handler type */
        put_le32(pb, s->ulVendorName); /* handler type */
        put_be32(pb ,0); /* reserved */
        put_be32(pb ,0); /* reserved */

        if ( !strcmp("mp4", s->oformat->name) )
        {
            /* .mp4: C string */
            str = "Text Media Handler";
            put_buffer(pb, (const unsigned char *)str, strlen(str)+1); /* include terminator */
        }
        else
        {
            /* .mov: PASCAL string */
            str = "";
            put_byte(pb, strlen(str)); /* string counter */
            put_buffer(pb, (const unsigned char *)str, strlen(str));
        }
    }

    return updateSize(pb, pos);
}

static int mov_write_hdlr_alis_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    char *str;
    offset_t pos = url_ftell(pb);

    put_be32(pb, 0); /* size */
    put_tag(pb, "hdlr");
    put_be32(pb, 0); /* Version & flags */
    put_tag(pb, "dhlr"); /* reserved */
    put_tag(pb, "alis"); /* handler type */
    put_le32(pb ,s->ulVendorName); /* 4 character vendor id */
    put_be32(pb ,0); /* reserved */
    put_be32(pb ,0); /* reserved */

    str = "";
    put_byte(pb, strlen(str)); /* string counter */
    put_buffer(pb, (const unsigned char *)str, strlen(str));
    return updateSize(pb, pos);
}

static int mov_write_mdhd_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    MOVEncodeContext *mov = (MOVEncodeContext *)s->priv_data;
    GenericCallbackFn lang_callback = s->oformat->generic_callback;
    unsigned short lang_code = 0x55C4; /* language = und */
    
    put_be32(pb, 32); /* size */
    put_tag (pb, "mdhd");
    put_be32(pb, 0); /* Version & flags */
    put_be32(pb, track->time); /* creation time */
    put_be32(pb, track->time); /* modification time */
    if( track->enc->codec_type == CODEC_TYPE_VIDEO ||
        track->enc->codec_type == CODEC_TYPE_TEXT )
    {
        if ( track->enc->codec_id == CODEC_ID_H264Z )
        {
            put_be32(pb, track->timescale); /* time scale (sample rate for audio) */
            // Convert from global timescale to audio-track's timescale
            put_be32(pb, track->trackDuration); /*duration*/
        }
        else
        {
            put_be32(pb, mov->timescale); /* time scale (sample rate for audio) */
            put_be32(pb, track->trackDuration ); /*duration*/
        }
    }
    else if( track->enc->codec_type == CODEC_TYPE_AUDIO )
    {
        put_be32(pb, track->timescale); /* time scale (sample rate for audio) */
        // Convert from global timescale to audio-track's timescale
        //put_be32(pb, (int64_t) track->trackDuration * (int64_t) track->timescale / mov->timescale  ); /*duration*/
        put_be32(pb,track->trackDuration); /*duration*/
    }

    if ( lang_callback != NULL )
    {
        lang_callback(AV_MODIFY_LANGUAGE_CODE, (PVOID)&lang_code, sizeof(lang_code));
    }

    put_be16(pb, lang_code);

    put_be16(pb, 0); /* reserved (quality) */
    return 32;
}

static int mov_write_mdia_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);
    put_be32(pb, 0); /* size */
    put_tag(pb, "mdia");
    mov_write_mdhd_tag(pb, track, s);
    mov_write_hdlr_tag(pb, track, s);
    mov_write_minf_tag(pb, track, s);
    return updateSize(pb, pos);
}

static int mov_write_tkhd_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    MOVEncodeContext *mov = s->priv_data;

    put_be32(pb, 0x5c); /* size (always 0x5c) */
    put_tag(pb, "tkhd");
    if( track->enc->codec_type == CODEC_TYPE_TEXT )
    {
        put_be32(pb, 0xe);
    }
    else
    {
        if ( strcmp("mp4", s->oformat->name) == 0 )
        {
            put_be32(pb, 0x7); /* version & flags (track enabled) */
        }
        else
        {
            put_be32(pb, 0xf); /* version & flags (track enabled) */
        }
    }
    put_be32(pb, track->time); /* creation time */
    put_be32(pb, track->time); /* modification time */
    put_be32(pb, track->trackID); /* track-id */
    put_be32(pb, 0); /* reserved */
	
    put_be32(pb, CEIL_DIV( (long)track->trackDuration *  ((int64_t)mov->timescale),(int64_t)track->timescale ) ); /* duration */

    put_be32(pb, 0); /* reserved */
    put_be32(pb, 0); /* reserved */
    if( track->enc->codec_type == CODEC_TYPE_TEXT )
    {
        put_be16(pb, 0xFFFF);
    }
    else
    {
        put_be16(pb, 0);
    }
    put_be16(pb, 0x0); /* Alternate group */
    /* Volume, only for audio */
    if(track->enc->codec_type == CODEC_TYPE_AUDIO)
        put_be16(pb, 0x0100);
    else
        put_be16(pb, 0);
    put_be16(pb, 0); /* reserved */

    /* Matrix structure */
    put_be32(pb, 0x00010000); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x00010000); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x40000000); /* reserved */

    /* Track width and height, for visual only */
    if( track->enc->codec_type == CODEC_TYPE_VIDEO ) {
        put_be32(pb, track->enc->logical_width*0x10000);
        put_be32(pb, track->enc->logical_height*0x10000);
    }
    else if( track->enc->codec_type == CODEC_TYPE_AUDIO ) {
        put_be32(pb, 0);
        put_be32(pb, 0);
    }
    else if( track->enc->codec_type == CODEC_TYPE_TEXT ) {
        put_be32(pb, track->enc->logical_width*0x10000);
        put_be32(pb, track->enc->logical_height*0x10000);
    }

    return 0x5c;
}

static int mov_write_tref_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    offset_t pos;
    int i;
    MOVEncodeContext *mov = s->priv_data;

    // check if text steams exists
    for( i = 0; i < s->nb_streams; i++ )
    {
        if( ( mov->tracks[i].entry > 0 ) &&
            ( mov->tracks[i].enc->codec_type == CODEC_TYPE_TEXT ) )
        {
            break;
        }
    }

    // if no text stream, return
    if( i >= s->nb_streams )
    {
        return 0;
    }

    pos = url_ftell(pb);
    put_be32(pb, 0); /* size */
    put_tag(pb, "tref");
    put_be32(pb, 12); // size
    put_tag(pb, "chap");
    put_be32(pb, i+1);

    return updateSize(pb, pos);
}

static int mov_write_trak_tag(ByteIOContext *pb, MOVTrack* track, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);
    MOVEncodeContext *mov = s->priv_data;

    put_be32(pb, 0); /* size */
    put_tag(pb, "trak");
    mov_write_tkhd_tag(pb, track, s);

    if( track->enc->codec_type == CODEC_TYPE_VIDEO )
    {
        if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_TREF_SYNC ||
             mov->ulEncodingFlags & MOV_ATOM_INCLUDE_TREF_MPOD )
        {
            mov_write_tref_tag2( pb, mov, 0, 0 );
        }
        else
        {
            mov_write_tref_tag(pb, track, s);
        }
    }

    if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_EDTS )
    {
        mov_write_edts_tag(pb, track, s);
    }

    mov_write_mdia_tag(pb, track, s);

    if ( (mov->ulEncodingFlags & MOV_ATOM_INCLUDE_UUID_USMT_MTDT) ||
         (mov->ulEncodingFlags & MOV_ATOM_INCLUDE_UUID_USMT_MTDF) )
    {
        mov_write_uuid_usmt(pb, mov);
    }

    return updateSize(pb, pos);
}

/* TODO: Not sorted out, but not necessary either */
static int mov_write_iods_tag(ByteIOContext *pb, MOVEncodeContext *mov)
{
    offset_t pos = url_ftell(pb);

    put_be32( pb, 0 ); /* size */
    put_tag ( pb, "iods" );
    put_be32( pb, 0x00000000 ); /* version & flags */
    put_be16( pb, 0x1007 );
    put_byte( pb, 0x00 );
    put_be16( pb, 0x4FFF );
    put_be16( pb, 0xFFFE );
    put_be16( pb, 0x01FF );

    return updateSize(pb, pos);
}

static int mov_write_mvhd_tag(ByteIOContext *pb, MOVEncodeContext *mov, AVFormatContext *s)
{
    int maxTrackID = 1, minTrackLen = 0, maxTrackLen = 0, i;
    int64_t minTrackLenTemp, maxTrackLenTemp;
    ULONG ulCreateTime = mov->time;

    put_be32(pb, 0x6c); /* size (always 0x6c) */
    put_tag(pb, "mvhd");
    put_be32(pb, 0); /* version & flags */

    if ( s->oformat->generic_callback != NULL )
    {
        s->oformat->generic_callback(
            AV_MODIFY_CREATION_TIME,
            &ulCreateTime,
            sizeof(ulCreateTime));
    }

    put_be32(pb, ulCreateTime); /* creation time */

    put_be32(pb, mov->time); /* modification time */
    put_be32(pb, mov->timescale); /* timescale */
    for (i=0; i<MAX_STREAMS; i++) {
        if(mov->tracks[i].entry > 0) {
            // Find the minimum track length ! Otherwise, one track will end
            // and another will continue. This looks bad when the audio is longer
            //minTrackLenTemp = mov->tracks[i].trackDuration
            // convert to mov->timescale:
            minTrackLenTemp = mov->tracks[i].trackDuration * ((int64_t)mov->timescale)/(int64_t)mov->tracks[i].timescale;
            //maxTrackLenTemp = mov->tracks[i].trackDuration;
            maxTrackLenTemp = mov->tracks[i].trackDuration * ((int64_t)mov->timescale)/(int64_t)mov->tracks[i].timescale;

            //first non zero track, or smaller than minimum
            if ( minTrackLenTemp != 0 )
            {
                if ((minTrackLen == 0)||(minTrackLen > minTrackLenTemp))
                {
                    minTrackLen = minTrackLenTemp;
                }
            }

            //first non zero track, or larger than maximum
            if ( maxTrackLenTemp != 0 )
            {
                if ( maxTrackLen < maxTrackLenTemp )
                {
                    maxTrackLen = maxTrackLenTemp;
                }
            }

            if(maxTrackID < mov->tracks[i].trackID)
                maxTrackID = mov->tracks[i].trackID;
        }
    }

    if ( strcmp("mp4", s->oformat->name) != 0  )
    {
        put_be32(pb, minTrackLen); /* duration of shortest track */

        // Update every track's duration to be equal to the duration of the movie
        // In case of MP4 let tracks retain their own exact duration
        for (i=0; i<MAX_STREAMS; i++)
        {
            if(mov->tracks[i].entry > 0)
            {
                // Save the track duration in its own Time Scale
                mov->tracks[i].trackDuration = minTrackLen *
                    ((int64_t)mov->tracks[i].timescale) /
                    (int64_t)mov->timescale;
            }
        }
    }
    else
    {
        // In case of MP4
        put_be32(pb, maxTrackLen); /* duration of longest track */
    }

    put_be32(pb, 0x00010000); /* reserved (preferred rate) 1.0 = normal */
    put_be16(pb, 0x0100); /* reserved (preferred volume) 1.0 = normal */

#ifdef MOV_INCLUDE_DEBUG_INFO
    {
        /* put current CL number into reserved area in movie container */        
        extern const char* AppGetPerforceChange(void);
        
        const char* version_st = AppGetPerforceChange();
        int version_len = strlen(version_st);

        /* length of reserved area is 10 bytes */
        for(i=0;i<10;i++){
            char symbol = 'Z';
            if (i<version_len){
                symbol = version_st[i];
            }
            put_byte(pb, symbol);
        }
    } 
#else
    {
        put_be16(pb, 0); /* reserved */
        put_be32(pb, 0); /* reserved */
        put_be32(pb, 0); /* reserved */
    }
#endif // MOV_INCLUDE_DEBUG_INFO

    /* Matrix structure */
    put_be32(pb, 0x00010000); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x00010000); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x0); /* reserved */
    put_be32(pb, 0x40000000); /* reserved */

    put_be32(pb, 0); /* reserved (preview time) */
    put_be32(pb, 0); /* reserved (preview duration) */
    put_be32(pb, 0); /* reserved (poster time) */
    put_be32(pb, 0); /* reserved (selection time) */
    put_be32(pb, 0); /* reserved (selection duration) */
    put_be32(pb, 0); /* reserved (current time) */
    put_be32(pb, maxTrackID+1); /* Next track id */
    return 0x6c;
}

static int mov_write_moov_tag(ByteIOContext *pb, MOVEncodeContext *mov, AVFormatContext *s)
{
    offset_t pos;
    int i;
    int video_stream = 0;
    uint64_t target_duration = 0;
    BOOL bMp4Format = ( strcmp("mp4", s->oformat->name) == 0);

    pos = url_ftell(pb);
    put_be32(pb, 0); /* size placeholder*/
    put_tag(pb, "moov");
    if ( bMp4Format )
    {
         mov->timescale = MP4_GLOBAL_TIME_SCALE;
    }
    else
    {
         mov->timescale = gl_globalTimescale;
         for( i = 0; i < s->nb_streams; i++ )
         {
            if( mov->tracks[i].enc->codec_id == CODEC_ID_H264Z )
            {
                mov->timescale = mov->tracks[i].enc->frame_rate;
            }
         }
    }

    for( i = 0; i < s->nb_streams; i++ )
    {
        int cl = 0;
        int id = 0;

        if( mov->tracks[i].entry <= 0 )
        {
            continue;
        }

        if( mov->tracks[i].enc->codec_type == CODEC_TYPE_VIDEO )
        {
            video_stream = i;
            //mov->tracks[i].timescale = mov->tracks[i].enc->frame_rate;

            if ( ( mov->tracks[i].enc->codec_id == CODEC_ID_H264Z ) )
            {
                mov->tracks[i].timescale = mov->tracks[i].enc->frame_rate;

                if ( mov->timescale < mov->tracks[i].timescale )
                {
                    mov->timescale = mov->tracks[i].timescale;
                }
            }
            else
            {
                mov->tracks[i].timescale = mov->timescale;
            }

			// Callback bit rate
			if ( s->oformat->generic_callback != NULL )
			{	
				UINT uiVideoBitRate = mov->tracks[i].enc->bit_rate;
				s->oformat->generic_callback( AV_VIDEO_BIT_RATE, &uiVideoBitRate, sizeof( uiVideoBitRate ) );
			} 

            printf("ENC FRAMERATE = %.02f\n", (float)mov->tracks[i].enc->frame_rate / mov->tracks[i].enc->frame_rate_base);
            //mov->tracks[i].sampleDuration = mov->tracks[i].enc->frame_rate_base * mov->timescale / mov->tracks[i].enc->frame_rate;
            mov->tracks[i].sampleDuration = mov->tracks[i].enc->frame_rate_base * mov->tracks[i].timescale / mov->tracks[i].enc->frame_rate;
        }
        else if(mov->tracks[i].enc->codec_type == CODEC_TYPE_AUDIO)
        {
            /* If AMR, track timescale = 8000, AMR_WB = 16000 */
            if(mov->tracks[i].enc->codec_id == CODEC_ID_AMR_NB)
            {
                mov->tracks[i].sampleDuration = 160;  // Bytes per chunk
                mov->tracks[i].timescale = 8000;
            }
            else if (( mov->tracks[i].enc->codec_id == CODEC_ID_ADPCM_MS )
                   ||( mov->tracks[i].enc->codec_id == CODEC_ID_ADPCM_IMA_WAV) )
            {
                mov->tracks[i].timescale = mov->tracks[i].enc->sample_rate;
                mov->tracks[i].sampleDuration = 1;
            }
            else if((mov->tracks[i].enc->codec_id == CODEC_ID_AACZ) ||
                    (mov->tracks[i].enc->codec_id == CODEC_ID_AAC))
            {
                mov->tracks[i].timescale = mov->tracks[i].enc->sample_rate;
                mov->tracks[i].sampleDuration = 1024;
            }
            else
            {
                mov->tracks[i].timescale = mov->tracks[i].enc->sample_rate;
                mov->tracks[i].sampleDuration = mov->tracks[i].enc->frame_size;
            }

            // special case of modifying existing file
            if( s->bModifyMode == TRUE )
            {
                if( mov->tracks[i].enc->codec_id == CODEC_ID_PCM_MULAW &&
                    mov->tracks[i].sampleDuration == 0 )
                {
                    mov->tracks[i].sampleDuration = 1;
                }
            }
        }
        else if( mov->tracks[i].enc->codec_type == CODEC_TYPE_TEXT )
        {
            // text duration will be measured in time scale units of video
            mov->tracks[i].timescale = mov->timescale;

            //
            mov->tracks[i].sampleDuration = mov->tracks[video_stream].sampleDuration;

            // track duration similar to one of video stream
            // assumption: video stream index is earlier one
            mov->tracks[i].trackDuration = mov->tracks[video_stream].trackDuration;

            // set text encoder window
            mov->tracks[i].enc->physical_width = mov->tracks[video_stream].enc->physical_width/2;
            mov->tracks[i].enc->physical_height = mov->tracks[video_stream].enc->physical_height/10;

            mov->tracks[i].enc->logical_width = mov->tracks[video_stream].enc->logical_width/2;
            mov->tracks[i].enc->logical_height = mov->tracks[video_stream].enc->logical_height/10;
        }

        if( mov->tracks[i].enc->codec_type != CODEC_TYPE_TEXT )
        {
            mov->tracks[i].trackDuration =  (int64_t)mov->tracks[i].sampleCount *
                                            (int64_t)mov->tracks[i].sampleDuration;
            // rescale to global timescale
            if( mov->tracks[i].enc->codec_type == CODEC_TYPE_VIDEO )
            {
                //target_duration = mov->tracks[i].trackDuration
                target_duration = mov->tracks[i].trackDuration * 
                                    (int64_t)mov->timescale/(int64_t)mov->tracks[i].timescale;
            }
        }

        cl = (mov->tracks[i].entry -1) / MOV_INDEX_CLUSTER_SIZE;
        id = (mov->tracks[i].entry -1) % MOV_INDEX_CLUSTER_SIZE;

        // PATCH: fixes "last frame white (empty) in QT" bug
        if( (mov->tracks[i].enc->codec_type == CODEC_TYPE_VIDEO) &&
            (mov->tracks[i].cluster[cl][id].size == 0 ) )
        {
            mov->tracks[i].trackDuration -= mov->timescale / mov->tracks[i].enc->frame_rate;
        }

        // In case of MP4, truncate audio duration to be shorter than Video duration
        if (( mov->tracks[i].enc->codec_type == CODEC_TYPE_AUDIO ) &&
            ( (mov->tracks[i].trackDuration* (int64_t)mov->timescale/(int64_t)mov->tracks[i].timescale) > target_duration ) &&
            ( target_duration != 0 ))
        {
            int truncated_entry = 0;
        
            // rescale to track timescale
            target_duration = (int64_t) target_duration *
                              (int64_t) mov->tracks[i].timescale /
                              mov->timescale;

            // resize sampleCount
            mov->tracks[i].sampleCount = (int64_t) target_duration /
                                         (int64_t) mov->tracks[i].sampleDuration;

            truncated_entry = mov->tracks[i].sampleCount / mov->tracks[i].cluster[0][0].samplesInChunk;

            if ( truncated_entry > mov->tracks[i].entry )
            {
                OsSysHalt( "Invalid truncating of MP4 audio stream" );
            }

            mov->tracks[i].entry = truncated_entry;

            // recalc duration
            mov->tracks[i].trackDuration =  (int64_t)mov->tracks[i].sampleCount *
                                            (int64_t)mov->tracks[i].sampleDuration;

            // do not rescale to global duration since it has been done elsewhere
            //mov->tracks[i].trackDuration = (int64_t) mov->tracks[i].trackDuration *
            //                               mov->timescale /
            //                               (int64_t) mov->tracks[i].timescale;
        }

        mov->tracks[i].time = mov->time;
        mov->tracks[i].trackID = i+1;
    }

    mov_write_mvhd_tag(pb, mov, s);

    if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_IODS )
    {
        mov_write_iods_tag(pb, mov);
    }

    for( i = 0; i < s->nb_streams; i++ )
    {
        if( mov->tracks[i].entry > 0 )
        {
            mov_write_trak_tag(pb, &(mov->tracks[i]), s);
        }
    }

    if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_MVEX )
    {
        mov_write_mvex_tag(pb);
    }

    // Callback Frame Count  
    if ( s->oformat->generic_callback != NULL )
    {   
        UINT uiframeCount = mov->tracks[video_stream].sampleCount;
        s->oformat->generic_callback( AV_VIDEO_FRAME_COUNT, &uiframeCount, sizeof( uiframeCount ) );
    } 

    // If user-data atom is present
    if (s->pbyUserAtom != NULL)
    {
        mov_write_udta_tag(pb, s->pbyUserAtom);
    }
    else
    {
        GenericCallbackFn udta_callback = s->oformat->generic_callback;
        
        uint8_t* udta_buffer = NULL;
        uint32_t udta_size = 0; 

        if ( udta_callback != NULL && 
             udta_callback(AV_MODIFY_UDTA_ATOM,
                           (void*)&udta_buffer,
                           (unsigned int)&udta_size) != 0 )
        {
            offset_t pos_udta = url_ftell(pb);
            
            put_be32( pb, 0 );
            put_tag( pb, "udta" );
            
            put_buffer( pb, udta_buffer, udta_size );
            avf_free( udta_buffer );

            updateSize(pb, pos_udta);
        }
    }

    return updateSize(pb, pos);
}

int mov_write_mdat_tag(ByteIOContext *pb, AVFormatContext *s)
{
    MOVEncodeContext* mov = s->priv_data;

    if ( !(mov->ulEncodingFlags & MOV_ATOM_REMOVE_EX_SIZE ) )       
    {  
    put_be32(pb, 8);    // placeholder for extended size field (64 bit)
    put_tag(pb, ( strcmp("mp4", s->oformat->name) == 0 ) ? "free" : "wide");
    }

    mov->mdat_pos = url_ftell(pb);
    put_be32(pb, 0); /* size placeholder*/
    put_tag(pb, "mdat");
    return 0;
}

int mov_write_ftyp_tag(ByteIOContext *pb, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);
    MOVEncodeContext* mov = s->priv_data;
    GenericCallbackFn ftyp_callback = s->oformat->generic_callback;

    uint8_t* ftyp_buffer = NULL;
    uint32_t ftyp_size = 0; 

    put_be32( pb, 0 );
    put_tag( pb, "ftyp" );

    if ( ftyp_callback != NULL && 
         ftyp_callback(AV_MODIFY_FTYP_ATOM,
                       (void*)&ftyp_buffer,
                       (unsigned int)&ftyp_size) != 0 )
    {
        put_buffer_cached( pb, ftyp_buffer, ftyp_size );
        avf_free( ftyp_buffer );
    }
    else
    {
        if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_EXTENDED_FTYP )
        {
            // If we here then we use extended .MP4
            put_tag ( pb, "MSNV" );     /* major brand */

            if ( strcmp("mp4", s->oformat->name) == 0 )
            {
                put_be32( pb, 0x01290046 ); //format_version, owner_code
            }
            else
            {
                put_be32( pb, 0x0100002B ); /* minor version */
            }
            put_tag ( pb, "MSNV" );     /* Sonys private brand */
            put_tag ( pb, "mp42" );
            put_tag ( pb, "isom" );
        }
        else
        {
            if ( !strcmp("3gp", s->oformat->name) )
                put_tag( pb, "3gp4" );
            else
                put_tag( pb, "isom" );

            put_be32( pb, 0x200 );

            if ( !strcmp("3gp", s->oformat->name) )
                put_tag( pb, "3gp4" );
            else
                put_tag( pb, "mp41" );
        }
    }

    return updateSize(pb, pos);
}

/*
    mov_fill_info
*/
static int mov_fill_info(ByteIOContext *pb, AVFormatContext *s, PAVMOVIEINFO pMovInfo)
{
    MOVEncodeContext *mov = s->priv_data;
    int i;

    for (i=0; i<MAX_STREAMS; i++)
    {
        if(mov->tracks[i].ents_allocated > 0)
        {
            if( mov->tracks[i].enc->codec_type == CODEC_TYPE_VIDEO )
            {
                pMovInfo->ulFirstSampleSize = mov->tracks[i].cluster[0][0].size;
                pMovInfo->ulFirstSampleOffset = mov->tracks[i].cluster[0][0].pos;
                pMovInfo->uwWidth = mov->tracks[i].enc->physical_width ;
                pMovInfo->uwHeight = mov->tracks[i].enc->physical_height ;
                pMovInfo->ulSampleCount = mov->tracks[i].sampleCount ;
            }

            if( mov->tracks[i].enc->codec_type == CODEC_TYPE_AUDIO )
            {
                pMovInfo->ulCodecID            = mov->tracks[i].enc->codec_id ;
                pMovInfo->ulAudioNumOfChannels = mov->tracks[i].enc->channels;
                pMovInfo->ulAudioSampleRate    = mov->tracks[i].enc->sample_rate;
                pMovInfo->ulAudioChunkCount    = mov->tracks[i].sampleCount;
            }

        }
    }

    return 0;
}

static int mov_update_skip_tag(ByteIOContext *pb, AVFormatContext *s)
{
    PBYTE pbySkipAtom = s->pbySkipAtom;
    AVMOVIEINFO sMovInfo;
    int   atomSize    = (int)SWAP_WORDS(((PULONG)pbySkipAtom )[0]);

    mov_fill_info( pb, s, &sMovInfo );

    if ( s->uiStream == AV_SECONDARY_STREAM )
    {
        sMovInfo.uiModifyAtom = AV_MODIFY_UUID_PROF_SECONDARY;
    }
    else
    {
        sMovInfo.uiModifyAtom = AV_MODIFY_UUID_PROF;
    }

    if( fnMovSkipAtomCb != NULL )
    {
        if( fnMovSkipAtomCb(&sMovInfo, pbySkipAtom, atomSize) )
        {
            if( atomSize !=  (int)SWAP_WORDS(((PULONG)pbySkipAtom )[0]) )
            {
                printf(" skip atom size is not correspond !\n");
                return 0;
            }

            put_buffer(pb, pbySkipAtom, atomSize);
        }
    }

    return 0;
}

static int mov_update_custom_tag(ByteIOContext *pb, AVFormatContext *s)
{
    PBYTE pbyCustomAtom = s->pbyCustomAtom;
    AVMOVIEINFO sMovInfo;
    int   atomSize    = (int)SWAP_WORDS(((PULONG)pbyCustomAtom )[0]);

    mov_fill_info( pb, s, &sMovInfo );

    if ( s->uiStream == AV_SECONDARY_STREAM )
    {
        sMovInfo.uiModifyAtom = AV_MODIFY_UUID_PROF_SECONDARY;
    }
    else
    {
        sMovInfo.uiModifyAtom = AV_MODIFY_UUID_PROF;
    }

    if( fnMovCustomAtomCb != NULL )
    {
        if( fnMovCustomAtomCb(&sMovInfo, pbyCustomAtom, atomSize) )
        {
            if( atomSize !=  (int)SWAP_WORDS(((PULONG)pbyCustomAtom )[0]) )
            {
                printf(" atom size is not correspond !\n");
                return 0;
            }

            put_buffer(pb, pbyCustomAtom, atomSize);
        }
    }

    return 0;
}

static int mov_write_header(AVFormatContext *s)
{
    ByteIOContext *pb = &s->pb;
    int i;
    MOVEncodeContext *mov = s->priv_data;

    // Init stream static params
    s->previous_stream_index = 0;
    s->first_frame_got = 0;

    for(i = 0; i < s->nb_streams; i++)
    {
        if(s->streams[i]->codec.codec_type == CODEC_TYPE_VIDEO)
        {
            s->ulSyntaxElementsFlag = s->streams[i]->codec.syntax_elements_flag;
        }
    }

    if(s->oformat != NULL)
    {
        mov->ulEncodingFlags = MOV_ATOM_DEFAULT;

        if( !strcmp("3gp", s->oformat->name) || !strcmp("mp4", s->oformat->name) )
        {
            if ( !strcmp("mp4", s->oformat->name) )
            {
                mov->ulEncodingFlags =   MOV_ATOM_INCLUDE_EXTENDED_FTYP
                                       | MOV_ATOM_INCLUDE_UUID_PROF
                                       | MOV_ATOM_INCLUDE_EDTS
                                       | MOV_ATOM_REMOVE_HDLR_ALIS
                                       | MOV_ATOM_INCLUDE_DREF_URL
                                       | MOV_ATOM_INCLUDE_UUID_USMT_MTDT
                                       | MOV_ATOM_MOVE_STSS_TO_STBL_END

                                   //  | MOV_ATOM_INCLUDE_UUID_USMT_MTDF
                                   //  | MOV_ATOM_INCLUDE_IODS
                                   //  | MOV_ATOM_INCLUDE_FREE
                                   //  | MOV_ATOM_INCLUDE_TREF_SYNC
                                   //  | MOV_ATOM_INCLUDE_TREF_MPOD
                                   //  | MOV_ATOM_INCLUDE_MVEX
                                   //  | MOV_ATOM_INCLUDE_MOOF

                                       ;

            }

            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_PROF)
            {
                mov->ulEncodingFlags |= MOV_ATOM_INCLUDE_UUID_PROF;
            }
            else
            {
                mov->ulEncodingFlags &= ~((ULONG)MOV_ATOM_INCLUDE_UUID_PROF);
            }

            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_EDTS){
            	mov->ulEncodingFlags |= MOV_ATOM_INCLUDE_EDTS;
            }
            else
            {
                mov->ulEncodingFlags &= ~((ULONG)MOV_ATOM_INCLUDE_EDTS);
            }

            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_FREE){
                mov->ulEncodingFlags |= MOV_ATOM_INCLUDE_FREE;
            }
            else
            {
                mov->ulEncodingFlags &= ~((ULONG)MOV_ATOM_INCLUDE_FREE);
            }

            if ( s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_EX_SIZE_REMOVE )
            {
                mov->ulEncodingFlags |= MOV_ATOM_REMOVE_EX_SIZE;
            }
            else
            {
                mov->ulEncodingFlags &= ~((ULONG)MOV_ATOM_REMOVE_EX_SIZE);
            }
            
             /* Updating of 'ftyp' atom is not allowed in Modify mode because the 
                      atom is situated before 'moov' and during updating of 'ftyp' the 
                      'moov' atom can be corrupted */
            if( ( s->flags & AVFMT_MODIFY_APPEND ) == 0 )
            {
                mov_write_ftyp_tag(pb,s);
            }

            if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_UUID_PROF )
            {
                if( ( s->flags & AVFMT_MODIFY_APPEND ) == 0 )
                {
                    mov_write_uuid_prof_tag(pb, s);
                }
            }
        }
        else if (!strcmp("mov", s->oformat->name))
        {
            mov->ulEncodingFlags |= MOV_ATOM_INCLUDE_UUID_USMT_MTDT;
            
            if (s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_FTYP){
                if( ( s->flags & AVFMT_MODIFY_APPEND ) == 0 )
                {
                    mov_write_ftyp_tag(pb,s);
                }
            }

            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_PROF)
            {
                mov->ulEncodingFlags |= MOV_ATOM_INCLUDE_UUID_PROF;
            }
            else
            {
                mov->ulEncodingFlags &= ~((ULONG)MOV_ATOM_INCLUDE_UUID_PROF);
            }

            if (s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_FREE){
                mov->ulEncodingFlags |= MOV_ATOM_INCLUDE_FREE;   
            }

            if(s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_EDTS)
            {                
                mov->ulEncodingFlags |= MOV_ATOM_INCLUDE_EDTS;
            }
            else
            {
                mov->ulEncodingFlags &= ~((ULONG)MOV_ATOM_INCLUDE_EDTS);
            }        

            if ( s->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_EX_SIZE_REMOVE )
            {  
                mov->ulEncodingFlags |= MOV_ATOM_REMOVE_EX_SIZE;
            }
            else
            {
                mov->ulEncodingFlags &= ~((ULONG)MOV_ATOM_REMOVE_EX_SIZE);
            }
            
            if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_UUID_PROF )
            {
                if( ( s->flags & AVFMT_MODIFY_APPEND ) == 0 )
                {
                    mov_write_uuid_prof_tag(pb, s);
                }
            }
        }
    }

    put_flush_packet(pb);

    if ( ( s->flags & AVFMT_MODIFY_APPEND ) == AVFMT_MODIFY_APPEND )
    {
        s->ulUpdatedFileSize = url_filesize(url_fileno(pb));

        printf("Append to %s at offset %#llx\n", s->filename, s->ulUpdatedFileSize );
    }
    mov->bSkipAtomSpaceCreated = FALSE;
    mov->bCustomAtomSpaceCreated = FALSE;

    return 0;
}

static unsigned int
timestamp()
{
#ifdef ZAVF
#ifdef BUILD_VERILOG
	ULONG ulRtcTime  = 0;
	ULONG ulBaseYear = 0;
#else
    // Read time from RTC component.
    ULONG ulRtcTime  = RtcGetUserTime();
    ULONG ulBaseYear = RtcGetBaseYear();
#endif
    ULONG ulNumBissextileYears = (ulBaseYear+
                                  MOV_TIMESTAMP_GMTIME_REL_YEARS -
                                  MOV_TIMESTAMP_MAC_FORMAT_BASE_YEAR)/4 + 1;

    ulRtcTime += ( ( ulBaseYear+
                     MOV_TIMESTAMP_GMTIME_REL_YEARS-
                     MOV_TIMESTAMP_MAC_FORMAT_BASE_YEAR ) *
                     MOV_TIMESTAMP_DAYS_IN_NON_BISSEXT_YEAR +
                     ulNumBissextileYears ) *
                     MOV_TIMESTAMP_SEC_IN_DAY;

    return ulRtcTime;
#else
    return 0;
#endif
}

static int mov_write_get_h264_dependancy(uint8_t *buf, int size, int* pbKeyFrame, int bMp4Format, int *skip_bytes)
{
    unsigned char extra_zeroes = 0;
    uint8_t* buf_start = buf;

    *pbKeyFrame = 0;
    *skip_bytes = 0;

    while(size > 4){
        unsigned char NALUnitType;

        uint8_t* pStartCodePrefix = h264_find_startcode(buf, buf + size);

        NALUnitType = pStartCodePrefix[4] & 0x1F;

        if(NALUnitType < 6){ // data
            GetBitContext gb;
            int slice_type;

            // modyfy NAL start code
            if (bMp4Format) {
                mov_enc_put32(buf, size - 4);
            } else {
                int parity = (buf - buf_start + size) & 1;
                mov_enc_put32(buf, size + parity - 4);
            }

            buf += 4;

            init_get_bits(&gb, buf + 1, size - 5);

            get_ue_golomb(&gb);

            slice_type = get_ue_golomb(&gb);

            if((slice_type == 0) || (slice_type == 5)){ // P-type
                return 0x40;
            }

            if((slice_type == 1) || (slice_type == 6)){ // B-type
                return 0x08;
            }

            if((slice_type == 2) || (slice_type == 7)){ // I-type
                *pbKeyFrame = 1;
                return 0;
            }
        }
        else
        {
            uint8_t* pNextStartCodePrefix = h264_find_startcode(pStartCodePrefix + 4, buf + size);

            if(pNextStartCodePrefix != NULL){

                extra_zeroes = 0;
                while ( *(pNextStartCodePrefix - extra_zeroes - 1) == 0 )
                {
                    extra_zeroes++;
                }

                if (extra_zeroes > 0)
                {
                    uint8_t* src = pNextStartCodePrefix - extra_zeroes;
                    uint8_t* dst = pNextStartCodePrefix;

                    while (src >= buf_start)
                        *dst-- = *src--;

                    buf_start += extra_zeroes;
                    *skip_bytes += extra_zeroes;

                    buf += extra_zeroes;
                    size -= extra_zeroes;
                }

                mov_enc_put32(buf, pNextStartCodePrefix - buf - 4);

                // check for recovering point SEI as start of GOP
                if(NALUnitType == 6) {   
                    //SEI:
                    int payload_type = 0;
                    int offset = 5;
                    uint8_t tmp_byte;

                    payload_type = 0;
                    tmp_byte = buf[offset++];
                    
                    while (tmp_byte == 0xFF) {
                        payload_type += 255;
                        tmp_byte = buf[offset++];
                    }
 
                    payload_type += tmp_byte;

                    if(payload_type == 6) {   
                        //recovery point
                        *pbKeyFrame = 1;
                    }
                }

                size -= pNextStartCodePrefix - buf;
                buf = pNextStartCodePrefix;
            }
            else
            {
                // last NAL in packet
                if (bMp4Format) {
                    mov_enc_put32(buf, size - 4);
                } else {
                    int parity = (buf - buf_start + size) & 1;
                    mov_enc_put32(buf, size + parity - 4);
                }

                return 0;// mark it as key frame?
            }
        }
    }

    return 0;
}

static int mov_write_packet(AVFormatContext *s, int stream_index,
                            uint8_t *buf, int size, int64_t pts)
{
    MOVEncodeContext *mov = s->priv_data;
    ByteIOContext *pb = &s->pb;
    AVCodecContext *enc = &s->streams[stream_index]->codec;
    MOVTrack* trk = &mov->tracks[stream_index];
    int cl, id;
    unsigned int samplesInChunk = 0;
    unsigned char dependancy = 0;
    int bKeyFrame = s->key_frame;
    int bMp4Format = 1;
    int iSkipBytes = 0;
    BOOL bMOValign = FALSE;

    if ( strcmp("mp4", s->oformat->name) != 0 )
    {
        bMp4Format = 0;
    }

    if ( (!mov->bSkipAtomSpaceCreated) && ( s->pbySkipAtom != NULL ) )
    {
        PBYTE pbyBuffer = s->pbySkipAtom;
        offset_t atomSize = SWAP_WORDS(((PULONG)pbyBuffer)[0]);

        mov->skip_atom_offset = url_ftell(pb);
        put_buffer(pb, pbyBuffer, atomSize);

        put_flush_packet(pb);

        mov->bSkipAtomSpaceCreated = TRUE;
    }
    if ( (!mov->bCustomAtomSpaceCreated) && ( s->pbyCustomAtom != NULL ) )
    {
        PBYTE pbyBuffer = s->pbyCustomAtom;
        offset_t atomSize = SWAP_WORDS(((PULONG)pbyBuffer)[0]);

        mov->custom_atom_offset = url_ftell(pb);
        put_buffer(pb, pbyBuffer, atomSize);

        put_flush_packet(pb);
        mov->bCustomAtomSpaceCreated = TRUE;
    }

    // If there is a thumbnail image to be written, and no space was
    // left for it (has to be a first frame) - leave the space for thumbnail.
    if ((s->ulThumbnailImageSize != 0)&&(!mov->bThumbSpaceCreated))
    {
        int uiThumbSpace;
        char* emptyBuffer;

        // The needed space is: 0xFA (constant fields length)
        // + 6 (endpic marker) + actual thumbnail size
        uiThumbSpace = 0xFA + 6 + s->ulThumbnailImageSize;

        // Add the padding (aligned to long)
        while (uiThumbSpace % 4 != 0)
            uiThumbSpace++;

        emptyBuffer = avf_malloc(uiThumbSpace);

        put_buffer(pb, (const unsigned char *)emptyBuffer, uiThumbSpace);

        avf_free(emptyBuffer);

        mov->bThumbSpaceCreated = TRUE;
    }

    if (url_is_streamed(&s->pb))
    {
        return 0; /* Can't handle that */
    }

    //mjpeg allows empty frames
    //if (!size) return 0;

    if (enc->codec_type == CODEC_TYPE_VIDEO )
    {
        samplesInChunk = 1;
    }
    else if (enc->codec_type == CODEC_TYPE_AUDIO )
    {
        if( enc->codec_id == CODEC_ID_AMR_NB)
        {
            /* We must find out how many AMR blocks there are in one packet */
            static uint16_t packed_size[16] =
            {13, 14, 16, 18, 20, 21, 27, 32, 6, 0, 0, 0, 0, 0, 0, 0};
            int len = 0;

            while (len < size && samplesInChunk < 100)
            {
                len += packed_size[(buf[len] >> 3) & 0x0F];
                samplesInChunk++;
            }
        }
        else if(enc->codec_id == CODEC_ID_PCM_ALAW ||
                enc->codec_id == CODEC_ID_PCM_MULAW )
        {
            samplesInChunk = size/enc->channels;
        }
        else if(( enc->codec_id == CODEC_ID_PCM_S16LE)  ||
                ( enc->codec_id == CODEC_ID_PCM_S16BE ) ||
                ( enc->codec_id == CODEC_ID_PCM_U16LE ) ||
                ( enc->codec_id == CODEC_ID_PCM_U16BE ))
        {
            samplesInChunk = size / 2 / enc->channels;
        }
        else if( (enc->codec_id == CODEC_ID_ADPCM_MS) ||
                 (enc->codec_id == CODEC_ID_ADPCM_IMA_WAV) )
        {
            samplesInChunk = enc->frame_size;
        }
        else if(( enc->codec_id == CODEC_ID_AACZ ) ||
                ( enc->codec_id == CODEC_ID_AAC )  ||
                ( enc->codec_id == CODEC_ID_MP2 ))
        {
            samplesInChunk = 1;
        }
        else
        {
            samplesInChunk = size;
        }
    }
    else if( enc->codec_type == CODEC_TYPE_TEXT )
    {
        samplesInChunk = 1;
    }

    if ( ( enc->codec_id == CODEC_ID_MPEG4Z || enc->codec_id == CODEC_ID_AAC || enc->codec_id == CODEC_ID_AACZ )
            && trk->vosLen == 0)
    {
        trk->vosLen = enc->extradata_size;
        if (enc->extradata_size != 0)
        {
            trk->vosData = avf_malloc(trk->vosLen);
            memcpy(trk->vosData, enc->extradata, trk->vosLen);
        }
    }

    if ( enc->codec_id == CODEC_ID_H264Z )
    {

        if ( size > 0 )
        {
            if ( trk->vosLen == 0 )
            {
                // handle first H264 frame:
                // Get SPS and PPS headers from first frame into trk->vosData;

                uint8_t* hrd_ptr = h264_find_startcode(buf, buf + size);

                if(hrd_ptr == buf)
                {
                    //SPS should start from the beginning:
                    uint8_t* pps_hrd_ptr = hrd_ptr;

                    hrd_ptr = h264_find_startcode(hrd_ptr + 4, buf + size);

                    if(hrd_ptr < buf + size)
                    {
                        // PPS:

                        mov_enc_put32(pps_hrd_ptr, hrd_ptr - pps_hrd_ptr - 4);

                        pps_hrd_ptr = hrd_ptr;

                        hrd_ptr = h264_find_startcode(hrd_ptr + 4, buf + size);

                        if(hrd_ptr < buf + size){

                            mov_enc_put32(pps_hrd_ptr, hrd_ptr - pps_hrd_ptr - 4);

                            trk->vosLen = hrd_ptr - buf;

                            if (trk->vosLen != 0)
                            {
                                trk->vosData = avf_malloc(trk->vosLen);
                                memcpy(trk->vosData, buf, trk->vosLen);
                            }

                            // should use here memory moving instead of changing code
                            //      (because of avf_realloc used below):
                            memcpy(buf, hrd_ptr, size - trk->vosLen);

                            size -= trk->vosLen;
                        }
                    }
                }

                s->first_frame_got = 0;
            }

            dependancy = mov_write_get_h264_dependancy(buf, size, &bKeyFrame, bMp4Format, &iSkipBytes);

            buf += iSkipBytes;
            size -= iSkipBytes;
        }
    }

    cl = trk->entry / MOV_INDEX_CLUSTER_SIZE;
    id = trk->entry % MOV_INDEX_CLUSTER_SIZE;

    if( trk->ents_allocated <= trk->entry )
    {
        trk->cluster = avf_realloc(trk->cluster, (cl+1)*sizeof(void*));
        if (!trk->cluster)
        {
            return -1;
        }
        trk->cluster[cl] = avf_mallocz(MOV_INDEX_CLUSTER_SIZE*sizeof(MOVIentry));
        if (!trk->cluster[cl])
        {
            return -1;
        }

        // Why do we need to memset a data buffer before use??!
        // AlexF.
        //memset( trk->cluster[cl], 0, MOV_INDEX_CLUSTER_SIZE*sizeof(MOVIentry));

        trk->ents_allocated += MOV_INDEX_CLUSTER_SIZE;
    }

    if( mov->mdat_written == 0 )
    {
        mov_write_mdat_tag(pb, s);
        mov->mdat_written = 1;
        mov->time = timestamp();
    }

        // NOT NEEDED FOR MP4
    if ( !bMp4Format &&
        // With MOV + H264 + AAC, if we add this byte to both AAC and H264
        // packets, as was done before, COACH chokes on playback. If we don't
        // add it at all mplayer chokes. If don't add it to AAC it plays back
        // fine on both mplayer and COACH.
        !((enc->codec_type == CODEC_TYPE_AUDIO) &&
           (( enc->codec_id == CODEC_ID_AACZ ) ||
            ( enc->codec_id == CODEC_ID_AAC ))) )
    {
        if( size & 1 )
        {
            size++;
            bMOValign = TRUE;
        }
    }

    if( s->eMetaData == META_DATA_QIK ) 
    {
        SQIKDATA sQik;
        UINT uiFrameNumber = s->streams[stream_index]->frame_number;

        sQik.uiFrameSize = SWAP_WORDS(size);
        sQik.uiTimeStamp = SWAP_WORDS(s->streams[stream_index]->time_length);

        if( enc->codec_type == CODEC_TYPE_AUDIO ) 
        {
            sQik.byPacketType = 0;
        }
        else
        {
            if( bKeyFrame == TRUE) 
            {
                sQik.byPacketType = 0xC0;
            }
            else
            {
                sQik.byPacketType = 0x80;
            }
        }

        sQik.byLiveCounter[0] = (uiFrameNumber & 0xFF0000) >> 16;
        sQik.byLiveCounter[1] = (uiFrameNumber & 0x00FF00) >> 8;
        sQik.byLiveCounter[2] = uiFrameNumber & 0x0000FF;

        put_buffer( pb, (PBYTE)&sQik, sizeof(SQIKDATA) );

        trk->mdat_size += sizeof(SQIKDATA);           
    }

    if( s->bModifyMode == FALSE )
    {
        trk->cluster[cl][id].pos = url_ftell(pb) - mov->movi_list;
    }
    else
    {
        offset_t next_pos;

        // in case we're not going to write first frame
        if( trk->entry > 0 )
        {
            int prev_cl = (trk->entry - 1) / MOV_INDEX_CLUSTER_SIZE;
            int prev_id = (trk->entry - 1) % MOV_INDEX_CLUSTER_SIZE;
            next_pos = trk->cluster[prev_cl][prev_id].pos +
                trk->cluster[prev_cl][prev_id].size;
            Avf_url_fseek( pb, next_pos, SEEK_SET );
        }
        else
        {
            mov_seek_last_chunk_end_pos( s );
            next_pos = s->ulUpdatedFileSize;
        }

        trk->cluster[cl][id].pos = next_pos;
    }

    trk->cluster[cl][id].samplesInChunk = samplesInChunk;
    trk->cluster[cl][id].size = size;


    if( enc->codec_type == CODEC_TYPE_VIDEO )
    {
        trk->cluster[cl][id].key_frame = bKeyFrame; //enc->coded_frame->key_frame;

        trk->cluster[cl][id].dependency = dependancy;

        // It's good to store the correct durations here:
        //trk->cluster[cl][id].duration = pts;

        if(bKeyFrame)
        {
            trk->hasKeyframes = 1;
        }
    }
    else if( enc->codec_type == CODEC_TYPE_TEXT )
    {
        // pts is index of video frame in case of
        trk->cluster[cl][id].duration = pts;
    }

    trk->enc = enc;
    trk->entry++;
    trk->sampleCount += samplesInChunk;
    trk->mdat_size += size;

    s->streams[stream_index]->stream_total_nb_frames = trk->sampleCount;

    if ( s->first_frame_got && trk->entry > 1) // chunks support code:
    {
        if ((s->previous_stream_index != stream_index) || (trk->SamplesInChunk >= trk->maxSamplesPerChunk))
        {
            // new chunk:
            trk->chunksCount ++;
            trk->SamplesInChunk = 0;
        }
    }
    else
    {
        // first frame, init data:
        // do it once for each track
        trk->chunksCount = 1;
        s->first_frame_got = 1;

        if(s->nb_streams == 1)
        {
            // single stream, chunks splitting by packets number;
            if(trk->enc->codec_type == CODEC_TYPE_VIDEO)
            {
                trk->maxSamplesPerChunk = trk->enc->gop_size > 1 ? trk->enc->gop_size : 1;
            }
            else if(trk->enc->codec_type == CODEC_TYPE_AUDIO)
            {
                if( (trk->enc->codec_id == CODEC_ID_ADPCM_MS) ||
                    (trk->enc->codec_id == CODEC_ID_ADPCM_IMA_WAV) ||
                    (trk->enc->codec_id == CODEC_ID_PCM_ALAW) ||
                    (trk->enc->codec_id == CODEC_ID_PCM_MULAW) )
                {
                    trk->maxSamplesPerChunk = samplesInChunk;
                }
                else
                {
                    trk->maxSamplesPerChunk = trk->enc->sample_rate / MOV_MAX_CHUNK_SIZE;
                }
            }
            else
            {
                trk->maxSamplesPerChunk = 1;
            }
        }
        else
        {
            // several streams, chunks splitting by writing interleaving;
            int audio_exist = 0;

            if (trk->enc->codec_type == CODEC_TYPE_TEXT)
            {
                trk->maxSamplesPerChunk = 1;
            }
            else
            {
                if (trk->enc->codec_type == CODEC_TYPE_AUDIO)
                {
                    audio_exist = 1;
                }

                if( (trk->enc->codec_id == CODEC_ID_ADPCM_MS) ||
                    (trk->enc->codec_id == CODEC_ID_ADPCM_IMA_WAV) ||
                    (trk->enc->codec_id == CODEC_ID_PCM_ALAW) ||
                    (trk->enc->codec_id == CODEC_ID_PCM_MULAW) )
                {
                    trk->maxSamplesPerChunk = samplesInChunk;
                }
                else
                {
                    trk->maxSamplesPerChunk = MOV_MAX_CHUNK_SIZE;
                }
            }

            // if audio doesn't exist we suppose that movie contains 2 streams:
            // video and text, in that case these streams should be split
            // by packets number
            if (!audio_exist)
            {
                // video stream
                mov->tracks[0].maxSamplesPerChunk = trk->enc->gop_size > 1 ?
                                                    trk->enc->gop_size : 1;
            }
            else
            {
                mov->tracks[0].maxSamplesPerChunk = MOV_MAX_CHUNK_SIZE;
            }
        }
    }

    s->previous_stream_index = stream_index;
    trk->SamplesInChunk += samplesInChunk;

    if ( bMOValign == TRUE )
    {
        put_buffer(pb, buf, size - 1);
        put_byte(pb, 0);
    }
    else
    {
    put_buffer(pb, buf, size);
    }

    put_flush_packet(pb);
    return 0;
}

static int mov_release_trailer_resources(AVFormatContext *s)
{
    int i = 0, j=0;
    MOVEncodeContext *mov = s->priv_data;

    if( mov != NULL )
    {
        for (i=0; i<MAX_STREAMS; i++)
        {
            for (j=0; j<mov->tracks[i].ents_allocated/MOV_INDEX_CLUSTER_SIZE; j++)
            {
                avf_free(mov->tracks[i].cluster[j]);
            }

            avf_free(mov->tracks[i].cluster);

            if( mov->tracks[i].vosLen )
            {
                avf_free( mov->tracks[i].vosData );
            }

            mov->tracks[i].cluster = NULL;
            mov->tracks[i].ents_allocated = mov->tracks[i].entry = 0;
        }
    }

    // :ZZZ:
    for(i=0;i<s->nb_streams;i++)
    {
        AVStream *st = s->streams[i];
        avf_free(st->codec.extradata);
    }

    return 0;
}

static int mov_write_trailer(AVFormatContext *s)
{
    MOVEncodeContext *mov = s->priv_data;
    ByteIOContext *pb = &s->pb;

    if( (pb->opaque == NULL) || (mov == NULL) )
    {
        mov_release_trailer_resources( s );
        return 0;
    }

    //Check if silence data is needed
    if ( s->ulMuteEndAudioBytes != 0 )
    {
        mov_write_audio_mute_end( pb,
                                  mov,
                                  (int)s->ulMuteEndAudioBytes,
                                  s->iAudioSilencePattern );
    }

    // :ZZZ: check if pb still exists
    if ( pb->write_packet != NULL )
    {
        int i = 0;
        uint64_t j = 0;
        offset_t file_size;

        put_flush_packet_end(pb);

        /* Write size of mdat tag */
        j = url_ftell(pb) - mov->mdat_pos - 8;

        pb->must_flush = 1;
        Avf_url_fseek(pb, 0, SEEK_SET);

        // If we need to update skip atom.
        if( s->pbySkipAtom != NULL )
        {
            offset_t pos = url_ftell(pb);
            Avf_url_fseek(pb, mov->skip_atom_offset, SEEK_SET);
            mov_update_skip_tag(pb, s);
            Avf_url_fseek(pb, pos, SEEK_SET);
        }

        if( s->pbyCustomAtom != NULL )
        {
            offset_t pos = url_ftell(pb);
            Avf_url_fseek(pb, mov->custom_atom_offset, SEEK_SET);
            mov_update_custom_tag(pb, s);
            Avf_url_fseek(pb, pos, SEEK_SET);
        }

        // Write preview atoms (thumbnail)
        if (s->ulThumbnailImageSize != 0)
        {
            put_flush_packet_end(pb);
            pb->pos = url_ftell(pb);

            mov_write_pnot_tag(pb, mov);
            mov_write_pict_tag(pb, s);
            
            put_flush_packet_end(pb);
            pb->must_flush = 1;
        }

        // Recover from file reopen (from burst to regular mode)
        file_size = url_filesize(url_fileno(pb));

        if (j + 8 <= UINT32_MAX) {
            Avf_url_fseek(pb, mov->mdat_pos, SEEK_SET);

            if (( s->flags & AVFMT_MODIFY_APPEND ) == AVFMT_MODIFY_APPEND ) {
                int a = mov_get_last_chunk_end(s, NULL);

                put_be32(pb, a - mov->mdat_pos);
                mov_seek_last_chunk_end_pos(s);
            } else {
                put_be32(pb, j + 8);
                Avf_url_fseek(pb, file_size, SEEK_SET);
            }
            
        } else {
            /* overwrite 'wide' placeholder atom */
            Avf_url_fseek(pb, mov->mdat_pos - 8, SEEK_SET);

            put_be32(pb, 1);
            put_tag(pb, "mdat");
            put_be64(pb, j + 16);
        
            Avf_url_fseek(pb, file_size, SEEK_SET);

            if( s->bModifyMode )
            {
                mov_seek_last_chunk_end_pos(s);
            }
        }

        put_flush_packet_end(pb);
        pb->pos = url_ftell(pb);

        mov_write_moov_tag(pb, mov, s);

        put_flush_packet(pb);

        if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_MOOF )
        {
            MOV_tfhd_t tfhd = { 0 };
            MOV_trun_t trun = { 0 };

#ifdef MOV_DEBUG
            // This part of code just for testing
            {
                UINT i = 0;
                tfhd.flags = TFHD_FLAG_BASE_DATA_OFFSET         |
                             TFHD_FLAG_SAMPLE_DESCRIPTION_INDEX |
                             TFHD_FLAG_DEFAULT_SAMPLE_DURATION  |
                             TFHD_FLAG_DEFAULT_SAMPLE_SIZE;
                tfhd.track_id = 1;
                tfhd.base_data_offset = 2;
                tfhd.sample_description_index = 3;
                tfhd.default_sample_duration = 4;
                tfhd.default_sample_size = 5;
                tfhd.default_sample_flags = 6;

                trun.flags = TRUN_FLAG_DATA_OFFSET          |
                             TRUN_FLAG_FIRST_SAMPLE_FLAGS   |
                             TRUN_FLAG_SAMPLE_DURATION      |
                             TRUN_FLAG_SAMPLE_SIZE          |
                             TRUN_FLAG_SAMPLE_FLAGS         |
                             TRUN_FLAG_SAMPLE_COMPOSITION_TIME_OFFSETS;
                trun.sample_count = 2;
                trun.data_offset = 3;
                trun.first_sample_flags = 4;
                trun.samples = avf_malloc( sizeof(MOV_trun_sample_t) * trun.sample_count );
                for ( i = 0; i < trun.sample_count; i++ )
                {
                    trun.samples[i].sample_duration = (i*10) + 1;
                    trun.samples[i].sample_size = (i*10) + 2;
                    trun.samples[i].sample_flags = (i*10) + 3;
                    trun.samples[i].sample_composition_time_offset = (i*10) + 4;
                }
            }
            // End test
#endif // MOV_DEBUG

            mov_write_moof_tag( pb, tfhd, trun );

            if ( trun.samples != NULL )
                avf_free( trun.samples );
        }

        if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_FREE )
        {
            //CL#101283
            GenericCallbackFn free_callback = s->oformat->generic_callback;
            
            uint8_t* free_buffer = NULL;
            uint32_t free_size = 0; 
            
            if ( free_callback != NULL && 
                 free_callback(AV_MODIFY_FREE_ATOM,
                               (void*)&free_buffer,
                               (unsigned int)&free_size) != 0 )
            {
                
                offset_t pos_udta = url_ftell(pb);
                
                put_be32( pb, 0 );
                put_tag( pb, "free" );
                
                put_buffer( pb, free_buffer, free_size );
                avf_free( free_buffer );
                
                updateSize(pb, pos_udta);      
            }
        }

        put_flush_packet_end(pb);
    }

    s->ulUpdatedFileSize = url_ftell( &s->pb);
    mov_release_trailer_resources( s );

    return 0;
}

int mov_estimate_trailer_size(AVFormatContext *s)
{
    MOVEncodeContext *mov = s->priv_data;
    int needed_space = 0;
    int i;
    int bMp4Format = 1;
    MOVIentry        dummyEntry = {0};

    if ( strcmp("mp4", s->oformat->name) != 0 )
    {
        bMp4Format = 0;
    }

    // All the constant-size headers in mvhd
    needed_space += 100;

	if (mov == NULL)
    {
		return needed_space;
    }
    // Add variable size headers
    for (i=0; i<MAX_STREAMS; i++)
    {
        int entries = mov->tracks[i].entry;

        if (entries <=0)
        {
            continue;
        }

        // estimation value of constant-size headers in trak
        needed_space += 500;

        if (mov->tracks[i].enc->codec_type == CODEC_TYPE_VIDEO)
        {
            // Count "stco" entries
            needed_space += sizeof(dummyEntry.pos)*entries;

            // Count "stsz" entries
            needed_space += sizeof(LONG)*entries;

            // Count "stss" entries.
            // TODO: Actually, "stts" keeps an entry per key frames only.
            // Meanwhile - assume that every frame is a key frame
            needed_space += sizeof(LONG)*entries;

            if ( (mov->tracks[i].enc->codec_id == CODEC_ID_H264Z) && bMp4Format)
            {
                // Count "ctts" entries
                needed_space += sizeof(LONG)*entries;
            }

            // Count "stsc" entries
            needed_space += sizeof(LONG)*entries;
        }
        else if (mov->tracks[i].enc->codec_type == CODEC_TYPE_TEXT)
        {
            // stts
            needed_space += 2*sizeof(LONG)*entries;

            // stsz
            needed_space += sizeof(LONG)*entries;

            // stco
            needed_space += sizeof(offset_t)*entries;
        }
        else
        {
            // Count "stco" entry
            needed_space += sizeof(dummyEntry.pos)*entries;

            // Count "stsz" entries
            needed_space += sizeof(LONG)*entries;

            // Count "stsc" entries
            needed_space += sizeof(LONG)*entries;
            
            // "stss" is not present for audio
        }
    }

    if (s->pbyUserAtom != NULL)
    {
        int atomSize = SWAP_WORDS(((PULONG)s->pbyUserAtom)[0]);

        needed_space += atomSize;
    }

    if ( s->oformat->generic_callback != NULL )
    {
    	UINT uiAtomSize = 0;
    	if ( s->oformat->generic_callback( AV_THUMBNAILS_SIZE, NULL, (UINT)&uiAtomSize ) != 0 )
    	{
    		needed_space += uiAtomSize;
    	}
    }

    return needed_space;
}

static int mov_set_parameters(AVFormatContext *s, AVFormatParameters *ap, void *pval)
{
    MOVAtomContent *tag = NULL;

    if( pval != NULL )
    {
        tag = (MOVAtomContent*)pval;

        if( tag->tag_id == MOV_ATOM_ID_SKIP )
        {
            if( tag->op == MOV_ATOM_REGISTER_CALLBACK )
            {
                fnMovSkipAtomCb = (FMOVSKIPATOMCB)tag->data;
                return 0;
            }
            else // Data buffer ptr
            {
                s->pbySkipAtom = (PBYTE)tag->data;
            }
        }
        else if( tag->tag_id == MOV_ATOM_ID_CUSTOM )
        {
            if( tag->op == MOV_ATOM_REGISTER_CALLBACK )
            {
                fnMovCustomAtomCb = (FMOVSKIPATOMCB)tag->data;
                return 0;
            }
            else // Data buffer ptr
            {
                s->pbyCustomAtom = (PBYTE)tag->data;
            }
        }

        memcpy( &atom[tag->tag_id], pval, sizeof(MOVAtomContent) );
    }

    return 0;
}

static offset_t mov_get_last_chunk_end( AVFormatContext *s, offset_t* mdat_pos )
{
    MOVEncodeContext *mov = s->priv_data;
    offset_t chunkPos=0;
    int chunkSize=0;
    int i;
    offset_t mdat_end = 0;
    int flag = 0;
    for( i = 0; i < mov->nb_streams; i++ )
    {
        MOVTrack* trk = &mov->tracks[i];
        int cl, id;

        if( trk->entry <= 0 )
        {
            continue;
        }

        cl = (trk->entry-1) / MOV_INDEX_CLUSTER_SIZE;
        id = (trk->entry-1) % MOV_INDEX_CLUSTER_SIZE;

        chunkPos = trk->cluster[cl][id].pos;
        chunkSize = trk->cluster[cl][id].size;


        if( flag == 0 )
        {
            mdat_end = chunkPos + chunkSize;

            flag = 1;
            continue;
        }

        if( mdat_end < (ULONG)(chunkPos + chunkSize) )
        {
            mdat_end = chunkPos + chunkSize;
        }

        if( mdat_pos != NULL && trk->cluster[0][0].pos < mov->mdat_pos )
        {
            *mdat_pos = trk->cluster[0][0].pos - 8;
        }
    }

    return mdat_end;
}

int mov_seek_last_chunk_end_pos(AVFormatContext *s)
{
    ByteIOContext *pb = &s->pb;
    MOVEncodeContext *mov = s->priv_data;

    put_flush_packet_end(pb);

    Avf_url_fseek(pb, 0, SEEK_SET);

    pb->must_flush = 1;

    s->ulUpdatedFileSize = mov_get_last_chunk_end( s, &mov->mdat_pos );

    mov->mdat_written = 1;

    Avf_url_fseek(pb, s->ulUpdatedFileSize, SEEK_SET);

    return 0;
}

static BOOL mov_write_audio_mute_is_supported( int codec_id )
{
    BOOL ret = FALSE;

    switch (codec_id)
    {
        case CODEC_ID_PCM_S8:
        case CODEC_ID_PCM_U8:
        case CODEC_ID_PCM_MULAW:
        {
            ret = TRUE;
            break;
        }
        default:
        {
            ret = FALSE;
            break;
        }
    }

    return ret;
}

static void	mov_write_audio_mute_frames( ByteIOContext *pb,
                                         MOVTrack *track,
                                         int chunkCount,
                                         int bytePos,
                                         int mutePattern )
{
    int samplePos = 0;
    int writeByte = 0;

    int j = 0;

    int cl = 0;
    int id = 0;

    for( j=0; j<chunkCount; j++ )
    {
        int i = 0;

        cl = ( track->entry-1-j ) / MOV_INDEX_CLUSTER_SIZE;
        id = ( track->entry-1-j ) % MOV_INDEX_CLUSTER_SIZE;

        // Is it last chunk
        if( chunkCount-1 == j )
        {
            samplePos = bytePos;
            writeByte = track->cluster[cl][id].samplesInChunk - bytePos;
        }
        else
        {
            // Not the last chunk
            samplePos = 0;
            writeByte = track->cluster[cl][id].samplesInChunk;
        }

        mov_seek_chunk_pos(pb, &track->cluster[cl][id], samplePos);

        for(i=0; i < writeByte; i++)
        {
            put_byte(pb, mutePattern);
        }
    }
}

static int mov_write_audio_mute_end(ByteIOContext *pb, MOVEncodeContext *mov, int size, int mutePattern)
{
    MOVTrack *track = NULL;

    int chunkCount = 0;
    int bytePos    = 0;

    int i = 0;

    //found the audio track
    for (i=0; i<MAX_STREAMS; i++)
    {
        track = &(mov->tracks[i]);

        if( track->enc->codec_type == CODEC_TYPE_AUDIO )
        {
            break;
        }
    }

    if ( i == MAX_STREAMS)
    {
        return 0;
    }

    if ( mov_write_audio_mute_is_supported( track->enc->codec_id ) == FALSE )
    {
        OsSysHalt("FATAL ERR: Muting AUDIO in MOV isn't supported for this codec" );
    }

    // Query the samples in chunk from the end
    mov_query_audio_sample_pos(track, size, &chunkCount, &bytePos);

    mov_write_audio_mute_frames( pb, track, chunkCount, bytePos, mutePattern );

    return 0;
}

static int mov_query_audio_sample_pos(MOVTrack* track, int sampleSize, int* chunkCount, int* byte)
{
    int sampleCount = 0;
    int i = 0;

    *chunkCount = 0;
    *byte = 0;

    //Counting from the last chunk
    for ( i=( track->entry-1 ); i>=0; i--)
    {
        int cl = i / MOV_INDEX_CLUSTER_SIZE;
        int id = i % MOV_INDEX_CLUSTER_SIZE;

        sampleCount += track->cluster[cl][id].samplesInChunk;
        (*chunkCount)++;

        if( sampleCount >= sampleSize )
        {
            *byte = sampleCount - sampleSize;
            break;
        }
    }

    return 0;
}

static int mov_seek_chunk_pos(ByteIOContext *pb, MOVIentry* enc, int byte)
{
    put_flush_packet_end(pb);

    pb->must_flush = 1;

    Avf_url_fseek(pb, 0, SEEK_SET);

    Avf_url_fseek(pb, (enc->pos+byte), SEEK_SET);

    return 0;
}

static int mov_parse_track_context( MOVTrack *trk,
                                    AVStream *st_out,
                                    int index )
{
    int cl=0, id =0;
    int i;
    unsigned int uiSampleCount   = 0;
    unsigned int uiNextSamplePos = 0;
    int entry = 0;
    int stsc_pos = 0;
    int ctts_pos = 0;
    int ctts_count = 0;
    int stss_pos = 0;
    int chunk = 0;
    int sample = 0;

    MOVStreamContext *sc = (MOVStreamContext *)st_out->priv_data;
    AVCodecContext *enco = &st_out->codec;

    int iSttsIndex = 0;
    long durTime  = sc->sample_duration[iSttsIndex++];

    int iStssIndex = 0;
    long nextKeyFrame = -1;//sc->sync_samples[iStssIndex];

    int iStscIndex = 0;
    long samplesInChunk = sc->sample_to_chunk[0].count;
    long nextFirst = -1;


    if( sc->sample_to_chunk_sz > 1 )
    {
        nextFirst = sc->sample_to_chunk[1].first;
    }
////////////////////////////////////////////////////////////////////////////////////
	if ( (trk != NULL) && (trk->enc != NULL) )
	{
    	// in case of ADPSM the smaple real size in defined inside the
    	// codec structure. See parsing of STSD atom
    	if( trk->enc->codec_id == CODEC_ID_ADPCM_MS )
    	{
        	printf("ERROR: See comments\n");
    	}
	}
	
    //There is 'stss' parsing
    if ( sc->sync_samples_count > 0 )
    {
        nextKeyFrame = sc->sync_samples[iStssIndex];
    }

    if( enco->codec_id == CODEC_ID_MPEG4Z || 
        enco->codec_id == CODEC_ID_AAC || 
        enco->codec_id == CODEC_ID_AACZ || 
        enco->codec_id == CODEC_ID_H264Z )
    {
        trk->vosLen = enco->extradata_size;
        if (enco->extradata_size != 0)
        {
            trk->vosData = avf_malloc(trk->vosLen);
            memcpy(trk->vosData, enco->extradata, trk->vosLen);
            avf_free( enco->extradata );
        }
    }
    else
    {
        trk->vosLen         = 0;
        trk->vosData        = 0;
    }

    // Get pic_init_qp from PPS
    if ( ( enco->codec_id == CODEC_ID_H264Z ) && 
         ( trk->vosData != NULL  ) )
    {
        GetBitContext gb;
        uint8_t *ptr = trk->vosData;
        unsigned int size;

        // SPS
        ptr += mov_enc_get32(trk->vosData);
        ptr += 4;

        // PPS
        size = mov_enc_get32(ptr);
        // size
        ptr += 4;
        // NAL type
        ptr += 1;

        init_get_bits(&gb, ptr, size * 8);

        get_ue_golomb(&gb);
        get_ue_golomb(&gb);
        get_bits1(&gb);
        get_bits1(&gb);
        get_ue_golomb(&gb);
        get_ue_golomb(&gb);
        get_ue_golomb(&gb);
        get_bits1(&gb);
        get_bits(&gb, 2);
        st_out->codec.pic_init_qp = get_se_golomb(&gb) + 26;
    }
    
    trk->entry          = sc->sample_count;
    trk->chunksCount    = sc->chunk_count;
    trk->mdat_size      = 0;  // updated when filling index table
    trk->ents_allocated = 0;  // index table
    trk->timescale      = 0;  // updated when writing trailer
    trk->time           = 0;
    trk->trackDuration  = 0;
    trk->sampleCount    = sc->sample_count;
    trk->sampleDuration = 0;  // todo
    trk->hasKeyframes = ( sc->sync_samples_count > 0 ) ? 1 : 0;
    trk->trackID        = index;
    trk->enc            = enco;
    trk->cluster        = NULL;

////////////////////////////////////////////////////////////////////////////////////

    {
        cl = ( trk->entry / MOV_INDEX_CLUSTER_SIZE ) + 1;
        trk->cluster = avf_mallocz(cl * sizeof(void*));
        for( i = 0; i < cl; i++ )
        {
            trk->cluster[i] = avf_mallocz(MOV_INDEX_CLUSTER_SIZE*sizeof(MOVIentry));
        }
        trk->ents_allocated = cl * MOV_INDEX_CLUSTER_SIZE;
    }

    if ( sc->ctts_entry_num > 0 )
    {
        ctts_count = sc->ctts_table[0].sample_count;
    }

    for(chunk = 0; chunk < trk->chunksCount; chunk++)
    {
        cl = uiSampleCount/ MOV_INDEX_CLUSTER_SIZE;
        id = uiSampleCount % MOV_INDEX_CLUSTER_SIZE;

        uiNextSamplePos = sc->chunk_offsets[chunk];

        if( chunk == (nextFirst - 1) )
        {
            if( ++iStscIndex < sc->sample_to_chunk_sz )
            {
                samplesInChunk = sc->sample_to_chunk[iStscIndex].count;
            }

            if( (iStscIndex + 1) < sc->sample_to_chunk_sz )
            {
                nextFirst = sc->sample_to_chunk[iStscIndex + 1].first;
            }
            else
            {
                nextFirst = -1;
            }
        }

        if( enco->codec_type == CODEC_TYPE_AUDIO && 
            ( enco->codec_id == CODEC_ID_AAC  || 
              enco->codec_id == CODEC_ID_AACZ || 
              enco->codec_id == CODEC_ID_MP2 ) )
        {
            //samples in chunk always 1
            trk->cluster[cl][id].samplesInChunk = 1;
        }
        else 
        {
            trk->cluster[cl][id].samplesInChunk = samplesInChunk;
        }

        // update duration entry
        trk->cluster[cl][id].duration = durTime;

        if( iSttsIndex < sc->sample_duration_num )
        {
            durTime  = sc->sample_duration[iSttsIndex++];
        }

        if( enco->codec_type == CODEC_TYPE_AUDIO && 
            ( enco->codec_id == CODEC_ID_AAC  || 
              enco->codec_id == CODEC_ID_AACZ || 
              enco->codec_id == CODEC_ID_MP2 ) )
        {
            for( i = 0; i < samplesInChunk; i++ )
            {
                cl = uiSampleCount/ MOV_INDEX_CLUSTER_SIZE;
                id = uiSampleCount % MOV_INDEX_CLUSTER_SIZE;

                trk->cluster[cl][id].samplesInChunk = 1;

                trk->cluster[cl][id].pos = uiNextSamplePos;

                // read sample sizes from index table
                trk->cluster[cl][id].size = ( sc->sample_size == 0 ) ? 
                                                sc->sample_sizes[uiSampleCount] : 
                                                sc->sample_size;

                //calculate next sample position if chunk contain more than 1 sample
                uiNextSamplePos = uiNextSamplePos + trk->cluster[cl][id].size;

                // update track data size
                trk->mdat_size += trk->cluster[cl][id].size;

                // update key frame entry
                if( sc->sync_samples_count > 0 && chunk == (nextKeyFrame - 1) )
                {
                    // update key frame entry
                    trk->cluster[cl][id].key_frame = 1;

                    if( ++iStssIndex < sc->sync_samples_count )
                    {
                        nextKeyFrame = sc->sync_samples[iStssIndex];
                    }
                    else
                    {
                        nextKeyFrame = -1;
                    }
                }
                else
                {
                    trk->cluster[cl][id].key_frame = 0;
                }

                uiSampleCount++;
            }
        }
        else 
        {
            offset_t chunk_offset = sc->chunk_offsets[chunk];
            offset_t sample_offset = chunk_offset;

            if ( (stsc_pos < sc->sample_to_chunk_sz - 1) &&
                 (chunk == sc->sample_to_chunk[stsc_pos + 1].first - 1) )
            {
                stsc_pos++;
            }

            entry = uiSampleCount;

            for ( sample = 0; sample < samplesInChunk; sample++ )
            {
                cl = entry / MOV_INDEX_CLUSTER_SIZE;
                id = entry % MOV_INDEX_CLUSTER_SIZE;

                if ( sc->ctts_entry_num > 0 )
                {
                    if ( entry > ctts_count - 1 )
                    {
                        ctts_pos++;
                        ctts_count += sc->ctts_table[ctts_pos].sample_count;
                    }
                    if ( ( sc->ctts_entry_num > 1 ) && 
                         ( sc->ctts_table[ctts_pos].sample_offset == 0 ) )
                    {
                        trk->cluster[cl][id].dependency = 0x0;
                    }
                    else
                    {
                        trk->cluster[cl][id].dependency = 0;
                    }
                }

                trk->cluster[cl][id].samplesInChunk = 1;  //?
                trk->cluster[cl][id].size =
                    ( sc->sample_size == 0 ) ? sc->sample_sizes[entry] : sc->sample_size;
                trk->cluster[cl][id].pos = sample_offset;
                sample_offset += trk->cluster[cl][id].size;

                // TODO: parse stts correctly
                trk->cluster[cl][id].duration = sc->sample_duration[0];
                if ( trk->hasKeyframes == 1 &&
                     entry == sc->sync_samples[stss_pos] - 1 )
                {
                    trk->cluster[cl][id].key_frame = 1;
                    stss_pos++;
                }
                else
                {
                    trk->cluster[cl][id].key_frame = 0;
                }

                trk->mdat_size += trk->cluster[cl][id].size;
                entry++;
                uiSampleCount++;
            }
        }
    }

    return 0;
}

int mov_parse_private_context( struct AVFormatContext *s_out, struct AVFormatContext *s_in )
{
    MOVDecodeContext *movIn  = (MOVDecodeContext *)s_in->priv_data;
    MOVEncodeContext *movOut = (MOVEncodeContext *)s_out->priv_data;
    UINT i;


    //Fill in the needed data content for encoder
    movOut->time                = movIn->fc->ulCreTimeStamp;
    movOut->nb_streams          = movIn->total_streams;
    movOut->mdat_pos            = movIn->mdat_offset - 8;
    movOut->mdat_written        = 1;
    movOut->timescale           = 0;
    movOut->movi_list           = 0;
    movOut->bThumbSpaceCreated  = 1;
    movOut->skip_atom_offset    = 0;
    movOut->custom_atom_offset  = 0;
    movOut->bCustomAtomSpaceCreated = FALSE;
    movOut->bSkipAtomSpaceCreated = FALSE;

    //write stream data
    for( i = 0; i < (UINT)movOut->nb_streams; i++ )
    {
        MOVTrack *trk     = &movOut->tracks[i];
        AVStream *st_out   = s_out->streams[i];

        mov_parse_track_context( trk, st_out, i );

        if ( trk->enc->codec_type == CODEC_TYPE_TEXT )
        {
            trk->maxSamplesPerChunk = 1;
        }
        else if ( ( trk->enc->codec_type == CODEC_TYPE_VIDEO ) && 
                  ( trk->enc->codec_id == CODEC_ID_H264Z ) )
        {
            if ( movIn->dependency != NULL ) {
                int j;
                for ( j = 0 ; j < trk->entry ; ++j ) 
                {
                    int cl = j / MOV_INDEX_CLUSTER_SIZE;
                    int id = j % MOV_INDEX_CLUSTER_SIZE;

                    trk->cluster[cl][id].dependency = movIn->dependency[ j ];
                }
            }
        }
        else
        {
            trk->maxSamplesPerChunk = MOV_MAX_CHUNK_SIZE;
        }
    }

    // single stream, chunks splitting by packets number;
    if(movOut->nb_streams == 1)
    {
        MOVTrack *trk     = &movOut->tracks[0];

        if(trk->enc->codec_type == CODEC_TYPE_VIDEO)
        {
            trk->maxSamplesPerChunk = trk->enc->gop_size > 1 ?
                                      trk->enc->gop_size : 1;
        }
        else if(trk->enc->codec_type == CODEC_TYPE_AUDIO)
        {
            trk->maxSamplesPerChunk = trk->enc->sample_rate / MOV_MAX_CHUNK_SIZE;
        }
        else
        {
            trk->maxSamplesPerChunk = 1;
        }
    }

    s_out->sMovStreamsInfo = s_in->sMovStreamsInfo;
    s_out->ulSyntaxElementsFlag = s_in->ulSyntaxElementsFlag;

    if (s_out->ulSyntaxElementsFlag & MPEG4_SYNTAX_ELEMENT_EDTS)
    {
        movOut->ulEncodingFlags |= MOV_ATOM_INCLUDE_EDTS;
    }

    return 0;
}

/*
  % func

    mov_delete_last_packet

  %desc

    delete last packet from the movie file

*/

static int mov_delete_last_packet( struct AVFormatContext *s,
                                   unsigned long flags )
{
    MOVEncodeContext *mov = (MOVEncodeContext *)s->priv_data;
    int stream_index =- 1;
    int i;
    offset_t max_offset = 0;
    MOVTrack* trk;
    int cl, id;

    // find stream containing last packet
    for( i = 0; i < mov->nb_streams; i++ )
    {
        trk = &mov->tracks[i];

        if( trk->entry <= 0 )
        {
            continue;
        }

        cl = ( trk->entry-1 ) / MOV_INDEX_CLUSTER_SIZE;
        id = ( trk->entry-1 ) % MOV_INDEX_CLUSTER_SIZE;

        if( trk->cluster[cl][id].pos > max_offset )
        {
            max_offset = trk->cluster[cl][id].pos;
            stream_index = i;
        }
    }

    if( stream_index == -1 )
    {
        return 1;
    }

    trk = &mov->tracks[stream_index];
    cl = ( trk->entry - 1 ) / MOV_INDEX_CLUSTER_SIZE;
    id = ( trk->entry - 1 ) % MOV_INDEX_CLUSTER_SIZE;

    trk->mdat_size -= trk->cluster[cl][id].size;

    trk->entry -= 1;
    trk->sampleCount -= 1;

    return 0;
}

/*
  % func

    mov_delete_packet

  %desc

    delete packet from the movie file

*/

int mov_delete_packet( struct AVFormatContext *s, unsigned long flags)
{
    int res = 0;

    if( flags & AV_CMD_DELETE_PACKET_LAST )
    {
        res = mov_delete_last_packet( s, flags );
    }

    return res;
}

int mov_add_stream(struct AVFormatContext *s, int id)
{
    MOVEncodeContext *mov = (MOVEncodeContext *)s->priv_data;
    AVStream *st = s->streams[id];
    MOVTrack *trk     = &mov->tracks[id];

    mov->nb_streams++;
    if( mov->nb_streams != s->nb_streams ||
        mov->nb_streams != (id+1) )
    {
        printf("Streams number inconsistency!!!\n");
    }

    memset( trk, 0, sizeof(MOVTrack) );

    trk->enc = &st->codec;

    return 0;
}

static int mov_video_to_text(struct AVFormatContext *s, int video, int text_stream)
{
    MOVEncodeContext *mov = (MOVEncodeContext *)s->priv_data;
    MOVTrack *trk     = &mov->tracks[text_stream];
    int i;

    for( i = 0; i < trk->entry; i++ )
    {
        int cl = i / MOV_INDEX_CLUSTER_SIZE;
        int id = i % MOV_INDEX_CLUSTER_SIZE;

        if( trk->cluster[cl][id].duration >= (unsigned int)video )
        {
            return i;
        }
    }

    return i;
}

static int mov_read_text_stream( struct AVFormatContext *s,
                                 int first,
                                 int total,
                                 char** text,
                                 int* indexes,
                                 int stream_id )
{
    int i;
    MOVEncodeContext *mov = (MOVEncodeContext *)s->priv_data;
    MOVTrack *trk = &mov->tracks[stream_id];
    offset_t pos;
    int size;
    unsigned char buf[256] = {0};
    int str_size;

    for( i = 0; i < total; i++ )
    {
        int entry = first+i;
        int cl = entry / MOV_INDEX_CLUSTER_SIZE;
        int id = entry % MOV_INDEX_CLUSTER_SIZE;

        // Get offset
        pos = trk->cluster[cl][id].pos;
        Avf_url_fseek(&s->pb, pos, SEEK_SET);

        // Get size
        size = trk->cluster[cl][id].size;

        // Read packet
        get_buffer( &s->pb, buf, size );

        // Copy string
        str_size = (int)(buf[0]<<8) + (int)(buf[1]<<0) + 2;
        text[i] = avf_mallocz( str_size + 1 );
        memcpy( text[i], buf, str_size );

        // Get index
        indexes[i] = trk->cluster[cl][id].duration;
    }

    return 0;

}

/**

%func

    mov_write_uuid_prof_tag

%desc

    Writes user extension atom with file profile data

*/

static int mov_write_uuid_prof_tag(ByteIOContext *pb, AVFormatContext *s)
{
    offset_t pos = url_ftell(pb);
    GenericCallbackFn uuid_prof_callback = s->oformat->generic_callback;
    
    uint8_t* uuid_buffer = NULL;
    uint32_t uuid_size = 0; 

    UINT uiModifyAtom = AV_MODIFY_UUID_PROF;

    if ( s->uiStream == AV_SECONDARY_STREAM )
    {
       uiModifyAtom = AV_MODIFY_UUID_PROF_SECONDARY;
    }

    if ( uuid_prof_callback != NULL && 
         uuid_prof_callback((unsigned int)uiModifyAtom,
                            (void*)&uuid_buffer,
                            (unsigned int)&uuid_size) != 0 )
    {
        put_be32( pb, 0 );              /* size */
        put_tag ( pb, "uuid" );
        
        put_buffer_cached( pb, uuid_buffer, uuid_size );
        avf_free( uuid_buffer );
    }
    else
    {
        int audioRate = 0, 
            frameRate = 0, 
            audio_kbitrate = 0, 
            video_kbitrate = 0, 
            video_max_kbitrate = 0, 
            i, 
            profiles_num = 1;
        AVCodecContext *pVideoCodec, *pAudioCodec;
    
        if ( s == NULL || pb == NULL ||
             s->streams[0] == NULL ||
             s->nb_streams <= 0 )
        {
            printf("ERROR: mov_write_uuid_prof_tag: Invalid input data!\n");
            return 0;
        }

        pVideoCodec = NULL;
        pAudioCodec = NULL;

        for ( i = 0; i < s->nb_streams; i++ )
        {
            if( s->streams[i] == NULL )
            {
                printf("ERROR: mov_write_uuid_prof_tag: Invalid input data!\n");
                return 0;
            }

            if( s->streams[i]->codec.codec_type == CODEC_TYPE_VIDEO)
            {
                pVideoCodec = &s->streams[i]->codec;
            }
            else if( s->streams[i]->codec.codec_type == CODEC_TYPE_AUDIO)
            {
                pAudioCodec = &s->streams[i]->codec;
            }
        }

        if( pAudioCodec != NULL)
        {
            audioRate = pAudioCodec->sample_rate;
            audio_kbitrate = pAudioCodec->bit_rate / 1000;
            profiles_num ++;
        }

        if( pVideoCodec != NULL)
        {
            frameRate = mov_get_vprf_frame_rate(pVideoCodec);
            video_kbitrate = FFMIN( pVideoCodec->bit_rate / 1000, 800 - audio_kbitrate );
            video_max_kbitrate = video_kbitrate;
            profiles_num ++;
        }

        put_be32( pb, 0 );              /* size */
        put_tag ( pb, "uuid" );

        /* 96 bits UUID: {50524F46-21D2-4FCE-BB88-695CFAC9C740} */
        put_tag ( pb, "PROF");          /* 0x50524F46 */
        put_be32( pb, 0x21D24FCE );
        put_be32( pb, 0xBB88695C );
        put_be32( pb, 0xFAC9C740 );

        put_be32( pb, 0x00000000 );     /* 1 byte - version; 3 bytes - flags */
        put_be32( pb, profiles_num );     /* profile entry count - up to 3: FPRF, APRF, VPRF */

        put_be32( pb, 0x00000014 );     /* size */
        put_tag ( pb, "FPRF");
        put_be32( pb, 0x00000000 );     /* flags */
        put_be32( pb, 0x00000000 );     /* reserved */
        put_be32( pb, 0x00000000 );     /* reserved? */

        if( pAudioCodec != NULL)
        {
            put_be32( pb, 0x0000002C );     /* size */
            put_tag ( pb, "APRF");          /* audio profile */
            put_be32( pb, 0x00000000 );     /* version? */
            put_be32( pb, 0x00000002 );     /* track id */
            put_tag ( pb, "mp4a" );         /* codec type */
            put_be32( pb, 0x00000229 );     /* codec specific information */
            put_be32( pb, 0x00000000 );     /* audio attribute flags */
            put_be32( pb, audio_kbitrate ); /* average bitrate */
            put_be32( pb, audio_kbitrate ); /* max bitrate */
            put_be32( pb, audioRate );      /* sampling rate */
            put_be32( pb, pAudioCodec->channels );  /* audio channel number */
        }

        if( pVideoCodec != NULL)
        {
            put_be32( pb, 0x00000034 );     /* size */
            put_tag ( pb, "VPRF" );         /* video profile */
            put_be32( pb, 0x00000000 );     /* version? */
            put_be32( pb, 0x00000001 );     /* track id */

            if ( pVideoCodec->codec_id == CODEC_ID_H264Z )
            {
                put_tag ( pb, "avc1" );     /* codec type */
                //put_be32( pb, 0x014D0015 ); /* codec specific information */
                if(pVideoCodec->physical_width <= 640)
                {
                    put_be32( pb, 0x014D401E ); /* codec specific information */

                    video_max_kbitrate = 0x2710;
                }
                else
                {
                    put_be32( pb, 0x014D0028 ); /* codec specific information */

                    video_max_kbitrate = 0x3E80;
                }

                if (pVideoCodec->bit_rate > 1000)
                {
                    video_kbitrate = pVideoCodec->bit_rate/1000;
                }
                else
                {
                    video_kbitrate = 1;
                }

                put_be32( pb, 0x00020002 );     /* video attribute flags */
            }
            else if ( strcmp("mp4", s->oformat->name) == 0 )
            {
                put_tag(pb, "mp4v");        /* codec type */
                /* codec specific information */
                if (( pVideoCodec->physical_width <= 320 ) && ( pVideoCodec->physical_height <= 240 ))
                {
                    if ((pVideoCodec->frame_rate /pVideoCodec->frame_rate_base) <= 15)
                    {
                        // Low resolution setting, use lower level
                        put_be32(pb, 0x00000103 );   /* profile and level */
                    }
                    else
                    {
                        // Low resolution setting, use lower level
                        put_be32(pb, 0x000001F3 );   /* profile and level */
                    }
                }
                else
                {
                    put_be32(pb, 0x000001F5 );   /* profile and level */
                }

                put_be32( pb, 0x00000000 );     /* video attribute flags */
            }else{
            put_be32( pb, 0x00000000 );     /* video attribute flags */
            }

            //put_be32( pb, 0x00000000 );     /* video attribute flags */

            put_be32( pb, video_kbitrate ); /* average bitrate */
            put_be32( pb, video_max_kbitrate ); /* max bitrate */
            put_be32( pb, frameRate );      /* average frame rate */
            put_be32( pb, frameRate );      /* max frame rate */
            /* visual_size */
            put_be16(pb, pVideoCodec->logical_width);
            put_be16(pb, pVideoCodec->logical_height);
            put_be32(pb, 0x00010001);       /* pixel aspect ratio */
        }
    }

    return updateSize(pb, pos);
}

/**

%func

    mov_write_edts_tag

%desc

    Writes edit list container
*/

static int mov_write_edts_tag(ByteIOContext *pb, MOVTrack *track, AVFormatContext *s)
{
    offset_t pos_edts = url_ftell(pb);
    offset_t pos_elst = 0;
    MOVEncodeContext *mov = (MOVEncodeContext *)s->priv_data;

    put_be32( pb, 0 );          /* size  */
    put_tag ( pb, "edts" );

    pos_elst = url_ftell(pb);
    put_be32( pb, 0 );          /* size  */
    put_tag ( pb, "elst" );
    put_be32( pb, 0x00000000 ); /* 1 byte - version; 3 bytes - flags */
    put_be32( pb, 0x00000001 ); /* entry count */

    // Duration is already calculated in mov_write_mvhd_tag
    //put_be32( pb, track->trackDuration ); /* segment duration */
    // in movie timescale (90000)
    put_be32( pb, track->trackDuration * ((int64_t)mov->timescale)/(int64_t)track->timescale); /* segment duration */

    /* media time */
    if ( track->enc->codec_id == CODEC_ID_H264Z && strcmp("mp4", s->oformat->name) == 0 )
    {
        put_be32( pb, track->sampleDuration ); //put_be32(pb, track->cluster[0].cts);
    }
    else
    {
        put_be32( pb, 0x00000000 ); //put_be32(pb, track->cluster[0].cts);
    }

    put_be16( pb, 0x0001 );     /* media rate */
    put_be16( pb, 0x0000 );     /* reserved */

    updateSize( pb, pos_elst );

    return updateSize( pb, pos_edts );
}

/**

%func

    mov_write_uuid_usmt

%desc

     Writes user extension atom with user specific meta data
*/

static int mov_write_uuid_usmt(ByteIOContext *pb, MOVEncodeContext *mov)
{
    offset_t pos_uuid = url_ftell(pb);

    put_be32( pb, 0 );
    put_tag ( pb, "uuid" );

    /* 96 bits UUID: {55534D54-21D2-4FCE-BB88-695CFAC9C740} */
    put_tag ( pb, "USMT" );         /* 0x55534D54*/
    put_be32( pb, 0x21D24FCE );
    put_be32( pb, 0xBB88695C );
    put_be32( pb, 0xFAC9C740 );

    if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_UUID_USMT_MTDF )
    {
        offset_t pos_mtdf = url_ftell(pb);

        put_be32( pb, 0 );          /* size */

        put_tag ( pb, "MTDF" );
        put_be16( pb, 0x0001 );     /* number of unit */
        put_be16( pb, 0x0012 );     /* record size */
        put_be32( pb, 0x0000000A ); /* data type ID */
        put_be32( pb, 0x55C40000);  /* ? */
        put_be16( pb, 0x0000 );     /* language */
        put_be16( pb, 0x0001 );     /* encoding type */
        put_be32( pb, 0x00000000 ); /* description */

        updateSize( pb, pos_mtdf );
    }

    if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_UUID_USMT_MTDT )
    {
        offset_t pos_mtdt = url_ftell(pb);

        put_be32( pb, 0 );          /* size */

        put_tag ( pb, "MTDT" );
        put_be16( pb, 0x0001 );     /* number of unit */
        put_be16( pb, 0x0012 );     /* data unit size */
        put_be32( pb, 0x0000000A ); /* data type ID */
        put_be16( pb, 0x55C4);  /* read_only_flag(1MSB) | language(15LSB) */
        put_be16( pb, 0x0000);  /* encoding_type = non_character_data */
        /* mata_data */
        put_be32( pb, 0x00000001 );     /* presentation_type = an original or main track */
        put_be32( pb, 0x00000000 ); /* meta data */

        updateSize( pb, pos_mtdt );
    }

    return updateSize( pb, pos_uuid );
}

/**

%func

    mov_write_tref_tag2

%desc

     Writes track reference container atom
*/

static int mov_write_tref_tag2(ByteIOContext *pb,
                               MOVEncodeContext *mov,
                               int syncID,
                               int odID)
{
    offset_t pos = url_ftell(pb);

    put_be32( pb, 0x00000000 );     /* size */
    put_tag ( pb, "tref" );

    if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_TREF_SYNC )
    {
        put_be32( pb, 0x0000000C ); /* size */
        put_tag ( pb, "sync" );
        put_be32( pb, syncID );     /* sync track ID */
    }

    if ( mov->ulEncodingFlags & MOV_ATOM_INCLUDE_TREF_MPOD )
    {
        put_be32( pb, 0x0000000C ); /* size */
        put_tag ( pb, "mpod" );
        put_be32( pb, odID );       /* OD track ID */
    }

    return updateSize( pb, pos );
}

/**

%func

    mov_write_nmhd_tag

%desc

     Writes null media header atom
*/

static int mov_write_nmhd_tag(ByteIOContext *pb)
{
    offset_t pos = url_ftell(pb);

    put_be32( pb, 0 );          /* size */
    put_tag ( pb, "nmhd" );
    put_be32( pb, 0x00000000 ); /* version & flags */

    return updateSize( pb, pos );
}

/**

%func

    mov_write_mvex_tag

%desc

     Writes movie extends box
*/

static int mov_write_mvex_tag(ByteIOContext *pb)
{
    offset_t pos = url_ftell( pb );

    put_be32( pb, 0x00000000 ); /* size */
    put_tag ( pb, "mvex" );

    put_be32( pb, 0x00000010 ); /* size */
    put_tag ( pb, "mehd" );
    put_be32( pb, 0x00000000 ); /* version & flags */
    put_be32( pb, 0x00000000 ); /* fragment duration */

    put_be32( pb, 0x00000020 ); /* size */
    put_tag ( pb, "trex" );
    put_be32( pb, 0x00000000 ); /* version & flags */
    put_be32( pb, 0x00000000 ); /* track ID */
    put_be32( pb, 0x00000000 ); /* default sample description index */
    put_be32( pb, 0x00000000 ); /* default sample duration */
    put_be32( pb, 0x00000000 ); /* default sample size */
    put_be32( pb, 0x00000000 ); /* default sample flags */

    return updateSize( pb, pos );
}

/**

%func

    mov_write_moof_tag

%desc

     Writes movie fragment atom
*/

static int mov_write_moof_tag(ByteIOContext *pb,
                              MOV_tfhd_t tfhd_data,
                              MOV_trun_t trun_data)
{
    offset_t pos_trun, pos_tfhd, pos_traf, pos_moof = url_ftell( pb );
    unsigned int i;

    put_be32( pb, 0x00000000 ); /* size */
    put_tag ( pb, "moof" );

    put_be32( pb, 0x00000010 ); /* size */
    put_tag ( pb, "mfhd" );
    put_be32( pb, 0x00000000 ); /* version & flags */
    put_be32( pb, 0x00000001 ); /* sequence number */

    pos_traf = url_ftell( pb );
    put_be32( pb, 0x00000000 ); /* size */
    put_tag ( pb, "traf" );

    pos_tfhd = url_ftell( pb );
    put_be32( pb, 0x00000000 ); /* size */
    put_tag ( pb, "tfhd" );
    put_be16( pb, 0x0000 );     /* version */
    put_be16( pb, tfhd_data.flags );
    put_be32( pb, tfhd_data.track_id );
    if ( tfhd_data.flags & TFHD_FLAG_BASE_DATA_OFFSET  )
        put_be64( pb, tfhd_data.base_data_offset );
    if ( tfhd_data.flags & TFHD_FLAG_SAMPLE_DESCRIPTION_INDEX  )
        put_be32( pb, tfhd_data.sample_description_index );
    if ( tfhd_data.flags & TFHD_FLAG_DEFAULT_SAMPLE_DURATION  )
        put_be32( pb, tfhd_data.default_sample_duration );
    if ( tfhd_data.flags & TFHD_FLAG_DEFAULT_SAMPLE_SIZE  )
        put_be32( pb, tfhd_data.default_sample_size );
    updateSize( pb, pos_tfhd );

    pos_trun = url_ftell( pb );
    put_be32( pb, 0x00000000 ); /* size */
    put_tag ( pb, "trun" );
    put_be16( pb, 0x0000 );     /* version */
    put_be16( pb, trun_data.flags );
    put_be32( pb, trun_data.sample_count );
    if ( trun_data.flags & TRUN_FLAG_DATA_OFFSET )
        put_be32( pb, trun_data.data_offset );
    if ( trun_data.flags & TRUN_FLAG_FIRST_SAMPLE_FLAGS )
        put_be32( pb, trun_data.first_sample_flags );
    for ( i = 0; i < trun_data.sample_count; i++ )
    {
        if ( trun_data.flags & TRUN_FLAG_SAMPLE_DURATION )
            put_be32( pb, trun_data.samples[i].sample_duration );
        if ( trun_data.flags & TRUN_FLAG_SAMPLE_SIZE )
            put_be32( pb, trun_data.samples[i].sample_size );
        if ( trun_data.flags & TRUN_FLAG_SAMPLE_FLAGS )
            put_be32( pb, trun_data.samples[i].sample_flags );
        if ( trun_data.flags & TRUN_FLAG_SAMPLE_COMPOSITION_TIME_OFFSETS )
            put_be32( pb, trun_data.samples[i].sample_composition_time_offset );
    }
    updateSize( pb, pos_trun );

    updateSize( pb, pos_traf );

    return updateSize( pb, pos_moof );
}

AVOutputFormat mov_oformat = {
    "mov",
    "mov format",
    NULL,
    "mov",
    sizeof(MOVEncodeContext),
    CODEC_ID_PCM_ALAW,
    CODEC_ID_MPEG4Z,
    mov_write_header,
    mov_write_packet,
    mov_write_trailer,
    0,
    mov_set_parameters,
    0,
    mov_estimate_trailer_size,
    mov_parse_private_context,
    mov_delete_packet,
    mov_add_stream,
    mov_read_text_stream,
    mov_video_to_text,
};

AVOutputFormat _3gp_oformat = {
    "3gp",
    "3gp format",
    NULL,
    "3gp",
    sizeof(MOVEncodeContext),
    CODEC_ID_AMR_NB,
    CODEC_ID_H263,
    mov_write_header,
    mov_write_packet,
    mov_write_trailer,
};

AVOutputFormat mp4_oformat = {
    "mp4",
    "mp4 format",
    "application/mp4",
    "mp4,m4a",
    sizeof(MOVEncodeContext),
    CODEC_ID_AAC,
    CODEC_ID_MPEG4Z,
    mov_write_header,
    mov_write_packet,
    mov_write_trailer,
    0,
    mov_set_parameters,
    0,
    mov_estimate_trailer_size,
    mov_parse_private_context,
};

int Avf_movenc_init(void)
{
    av_register_output_format(&mov_oformat);
    av_register_output_format(&_3gp_oformat);
    av_register_output_format(&mp4_oformat);
    return 0;
}

